-------------------------------------------------------------------------------
-- BusInt.vhd
-- ------------
-- Bus interface for the processor. Handles sequencing IM and DM request to 
-- only one memory instance. IM has higher priority than DM and is served 
-- first. The serving is performed through a Mealy-type state machine (to allow
-- single cycle access for IM, which saves a wait state).
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973431 Rune M. Barnkob
--             :  c973816 Emil Heinze
--
-- Last Update :  2001/01/28 - made dual-port interface and moved tri-state to
--                             external module
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- BusInt
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity BusInt is
  port(
    -- global signals
    clk       : in  bit_t;
    reset     : in  bit_t;
    
    -- interface towards the cpu core
    im_addr   : in  halfword_t;
    im_datain : in  halfword_t;
    im_dataout: out halfword_t;
    im_rw     : in  bit_t;
    im_req    : in  bit_t;
    im_hold   : out bit_t;
    dm_addr   : in  halfword_t;
    dm_datain : in  halfword_t;
    dm_dataout: out halfword_t;
    dm_rw     : in  bit_t;
    dm_req    : in  bit_t;
    dm_hold   : out bit_t;
   
    -- interface towards memory
    mem_addr    : out halfword_t;
    mem_datain  : in  halfword_t;
    mem_dataout : out halfword_t;
    mem_rw      : out bit_t;
    mem_req     : out bit_t;
    mem_hold    : in  bit_t
  );
end entity BusInt;

architecture behaviour of BusInt is
  type states is (pass_im, serve_im, init_dm, serve_dm);
  signal curr_st: states;      -- the current state of the interface
  signal nextstate: states;    -- the next state (after the clock)
  signal im_en : bit_t;        -- enable signal to catch holding data from an im-read
  signal gl_hold: bit_t;       -- global hold signal
  signal source: bit_t;        -- signal that represents which source is using the memory now ('0': IM, '1': DM)
  signal im_intdo: halfword_t; -- signal for register

begin
  -- purpose: set all signals to internal and external systems
  -- type   : combinatorial
  -- input  : curr_st, mem_hold, im_req, dm_req
  -- output : gl_hold, nextstate, source, im_en
  NEXTSTATE_LOGIC: process(curr_st, mem_hold, im_req, dm_req)
  begin
    case curr_st is
      when pass_im =>
        gl_hold <= mem_hold or dm_req;
        if mem_hold='0' or im_req='0' then 
          nextstate <= init_dm;
        else
          nextstate <= serve_im;
        end if;
        source <= '0'; -- mux im_addr
        im_en <= '1';
        
      when serve_im =>
        gl_hold <= mem_hold or dm_req;
        if mem_hold='0' and dm_req='1' then
          nextstate <= init_dm;
        elsif mem_hold='1' then
          nextstate <= serve_im;
        else
          nextstate <= pass_im;
        end if;
        source <= '0';
        im_en <= '1';
        
      when init_dm =>
        gl_hold <= mem_hold;
        if mem_hold='1' then 
          nextstate <= serve_dm;
        else
          nextstate <= pass_im;
        end if;
        source <= '1';
        im_en <= '0';
      
      when serve_dm => 
        gl_hold <= mem_hold;
        if mem_hold='1' then
          nextstate <= serve_dm;
        else
          nextstate <= pass_im;
        end if;
        source <= '1';
        im_en <= '0';
      
      when others =>
        gl_hold <='0';        
        nextstate <= pass_im;
        source <= '0';
        im_en <= '0';
        
    end case;
  end process NEXTSTATE_LOGIC;
  
  -- purpose: state register for the Mealy machine
  -- type   : sequential
  -- input  : reset, clk
  -- output : curr_st
  STATE_REG: process(clk, reset)
  begin
    if reset/='1' then
      curr_st <= pass_im;
    elsif clk'event and clk='1' then
      curr_st <= nextstate;
    end if;
  end process STATE_REG;

  -- purpose: register for catching data read from memory to IM (to fix bug)
  -- type   : sequential
  -- input  : clk, reset
  -- output : im_intdo
  CATCH_IM: process(clk, reset)
  begin
    if reset/='1' then
      im_intdo <= halfword_zero;
    elsif clk'event and clk='1' then
      if im_en='1' then
        im_intdo <= mem_datain;
      end if;
    end if;
  end process CATCH_IM;

  -- bypass the register during the IM access cycle
  IM_MUX: im_dataout <= mem_datain when im_en='1' else im_intdo;  
  
  -- pass the data read from memory directly to the dm_din
  MEM_TO_DATA: dm_dataout <= mem_datain;
  -- pass dm_datain to the internal temporary signal to force on pins
  DATA_TO_MEM: mem_dataout <= dm_datain;
  
  -- same global hold signal on both IM and DM
  SET_IMHOLD: im_hold <= gl_hold;
  SET_DMHOLD: dm_hold <= gl_hold;
  -- mux'es to memory
  MUX_ADDR: mem_addr <= dm_addr when source='1' else im_addr;
  MUX_REQ: mem_req <= dm_req when source='1' else im_req;
  MUX_RW: mem_rw <= dm_rw when source='1' else im_rw;
end architecture behaviour;
