-------------------------------------------------------------------------------
-- MiniMIPS.vhd
-- ------------
-- This is the final implementation version of the CPU and the bus interface.
-- It has a single port memory interface using tri-state buffers on the data
-- and address pins.
--
-- Authors : Rune M. Barnkob (c973431)
--         : Mikael Lund Jepsen (c973811)
--         : Emil Heinze (c973816)
--
-- Version : 2001/01/23 - initial version
--           2001/01/28 - all bus drivers moved to MiniMIPS.vhd
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use WORK.types.all;

entity MiniMIPS is port (
  clk   : in    bit_t;       -- global clock
  reset : in    bit_t;       -- global reset
  
  req   : out   bit_t;       -- request to memory
  rw    : out   bit_t;       -- read-not-write signal to memory
  hold  : in    bit_t;       -- hold from memory
  addr  : out   halfword_t;  -- address towards memory
  data  : inout halfword_t   -- data to send/read to/from memory
  );
end entity MiniMIPS;

architecture structure of MiniMIPS is
  component cpu is port (
      clk :    in  bit_t;            -- The clock.
      reset:   in  bit_t;            -- The reset signal.
      im_addr: out halfword_t;       -- Address bus for instruction.
      im_din:  in  halfword_t;       -- The instruction bus.
      im_dout: out halfword_t;       -- 
      im_req:  out bit_t;            -- Request signal for instruction.
      im_rw:   out bit_t;            -- Read/Write signal for instruction.
      im_hold: in  bit_t;            -- Hold signal for instruction.
      dm_addr: out halfword_t;       -- Address bus for data.
      dm_din:  in  halfword_t;       -- The data in bus
      dm_dout: out halfword_t;       -- Data bus for output
      dm_req:  out bit_t;        -- Request signal for data.
      dm_rw:   out bit_t;       -- Read/Write signal for data.
      dm_hold: in  bit_t        -- Hold signal for data.
    );
  end component cpu;
  component BusInt is port(
    -- global signals
    clk       : in  bit_t;
    reset     : in  bit_t;
    
    -- interface towards the cpu core
    im_addr   : in  halfword_t;
    im_datain : in  halfword_t;
    im_dataout: out halfword_t;
    im_rw     : in  bit_t;
    im_req    : in  bit_t;
    im_hold   : out bit_t;
    dm_addr   : in  halfword_t;
    dm_datain : in  halfword_t;
    dm_dataout: out halfword_t;
    dm_rw     : in  bit_t;
    dm_req    : in  bit_t;
    dm_hold   : out bit_t;
   
    -- interface towards memory
    mem_addr    : out halfword_t;
    mem_datain  : in  halfword_t;
    mem_dataout : out halfword_t;
    mem_rw      : out bit_t;
    mem_req     : out bit_t;
    mem_hold    : in  bit_t
  );
  end component BusInt;
  component im_cache is port(
    reset:    in  bit_t;
    clk:      in  bit_t;
    -- Interface towards CPU core
    imcpu_req:   in  bit_t; -- Request signal from CPU
    imcpu_hold:  out bit_t; -- Hold signal to CPU while accessing memory
    imcpu_rw:    in  bit_t; -- rw (expected constant =1)
    imcpu_addr:  in  std_logic_vector(15 downto 0); -- Address from CPU
    imcpu_dout:  out halfword_t; -- The data read either from cache or memory
    imcpu_din:   in  halfword_t; -- Dummy port
  
    -- The interface against the bus interface
    imbus_req:   out bit_t; -- request to send to bus interface
    imbus_hold:  in  bit_t; -- hold from bus interface
    imbus_rw:    out bit_t; -- rw signal to bus interface
    imbus_addr:  out halfword_t; -- address to bus interface
    imbus_dout:  out halfword_t; -- data send to businterface (dummy)
    imbus_din:   in  halfword_t  -- data received from memory
  );
  end component im_cache;
  
  -- signals from I-cache to bus interface
  signal im_addr: halfword_t;
  signal im_din: halfword_t;
  signal im_dout: halfword_t;
  signal im_req:  bit_t;
  signal im_rw:   bit_t;
  signal im_hold: bit_t;
  -- signals from CPU to I-cache
  signal ic_addr: halfword_t;
  signal ic_din: halfword_t;
  signal ic_dout: halfword_t;
  signal ic_req:  bit_t;
  signal ic_rw:   bit_t;
  signal ic_hold: bit_t;
  -- signals from CPU to bus interface
  signal dm_addr: halfword_t;
  signal dm_din: halfword_t;
  signal dm_dout: halfword_t;
  signal dm_req:  bit_t;
  signal dm_rw:   bit_t;
  signal dm_hold: bit_t;
  -- signals from bus interface to main memory
  signal mem_addr: halfword_t;
  signal tomemdata: halfword_t;
  signal mem_req: bit_t;
  signal mem_rw: bit_t;
  signal drive_db: bit_t;
  
begin
  -- instantiates the CPU core
  processor: cpu port map (
    clk => clk,
    reset => reset,
    im_addr => ic_addr,
    im_din => ic_dout,
    im_dout => ic_din,
    im_req => ic_req,
    im_rw => ic_rw,
    im_hold => ic_hold,
    dm_addr => dm_addr,
    dm_din => dm_din,
    dm_dout => dm_dout,
    dm_req => dm_req,
    dm_rw => dm_rw,
    dm_hold => dm_hold
  );
  -- instantiates the cache memory
--  icache: im_cache port map (
--    reset => reset,
--    clk => clk,
--    imcpu_req => ic_req,
--    imcpu_hold => ic_hold,
--    imcpu_rw => ic_rw,
--    imcpu_addr => ic_addr,
--    imcpu_dout => ic_dout,
--    imcpu_din => ic_din,
--    imbus_req => im_req,
--    imbus_hold => im_hold,
--    imbus_rw => im_rw,
--    imbus_addr => im_addr,
--    imbus_dout => im_dout,
--    imbus_din => im_din
--  );
  -- instantiates the bus interface from the cpu
  bussys: busint port map (
    clk => clk,
    reset => reset,
--    im_addr => im_addr,
--    im_datain => im_dout,
--    im_dataout => im_din,
--    im_req => im_req,
--    im_rw => im_rw,
--    im_hold => im_hold,
    im_addr => ic_addr,
    im_datain => ic_din,
    im_dataout => ic_dout,
    im_req => ic_req,
    im_rw => ic_rw,
    im_hold => ic_hold,
    dm_addr => dm_addr,
    dm_datain => dm_dout,
    dm_dataout => dm_din,
    dm_req => dm_req,
    dm_rw => dm_rw,
    dm_hold => dm_hold,
    mem_rw => mem_rw,
    mem_req => mem_req,
    mem_hold => hold,
    mem_addr => addr,
--    mem_addr => mem_addr,
    mem_datain => data,
    mem_dataout => tomemdata
  );
  
  drive_db <= (not mem_rw) and mem_req and ((not clk) or hold);
  req <= mem_req;
  rw <= mem_rw;
--  req <= '1' when mem_req='1' else 'Z';
--  rw <= mem_rw when mem_req='1' else 'Z';
--  addr(00) <= mem_addr(00) when mem_req='1' else 'Z';
--  addr(01) <= mem_addr(01) when mem_req='1' else 'Z';
--  addr(02) <= mem_addr(02) when mem_req='1' else 'Z';
--  addr(03) <= mem_addr(03) when mem_req='1' else 'Z';
--  addr(04) <= mem_addr(04) when mem_req='1' else 'Z';
--  addr(05) <= mem_addr(05) when mem_req='1' else 'Z';
--  addr(06) <= mem_addr(06) when mem_req='1' else 'Z';
--  addr(07) <= mem_addr(07) when mem_req='1' else 'Z';
--  addr(08) <= mem_addr(08) when mem_req='1' else 'Z';
--  addr(09) <= mem_addr(09) when mem_req='1' else 'Z';
--  addr(10) <= mem_addr(10) when mem_req='1' else 'Z';
--  addr(11) <= mem_addr(11) when mem_req='1' else 'Z';
--  addr(12) <= mem_addr(12) when mem_req='1' else 'Z';
--  addr(13) <= mem_addr(13) when mem_req='1' else 'Z';
--  addr(14) <= mem_addr(14) when mem_req='1' else 'Z';
--  addr(15) <= mem_addr(15) when mem_req='1' else 'Z';
  data(00) <= tomemdata(00) when drive_db='1' else 'Z';
  data(01) <= tomemdata(01) when drive_db='1' else 'Z';
  data(02) <= tomemdata(02) when drive_db='1' else 'Z';
  data(03) <= tomemdata(03) when drive_db='1' else 'Z';
  data(04) <= tomemdata(04) when drive_db='1' else 'Z';
  data(05) <= tomemdata(05) when drive_db='1' else 'Z';
  data(06) <= tomemdata(06) when drive_db='1' else 'Z';
  data(07) <= tomemdata(07) when drive_db='1' else 'Z';
  data(08) <= tomemdata(08) when drive_db='1' else 'Z';
  data(09) <= tomemdata(09) when drive_db='1' else 'Z';
  data(10) <= tomemdata(10) when drive_db='1' else 'Z';
  data(11) <= tomemdata(11) when drive_db='1' else 'Z';
  data(12) <= tomemdata(12) when drive_db='1' else 'Z';
  data(13) <= tomemdata(13) when drive_db='1' else 'Z';
  data(14) <= tomemdata(14) when drive_db='1' else 'Z';
  data(15) <= tomemdata(15) when drive_db='1' else 'Z';
end structure;
