-- -----------------------------------------------------------------------------
--
--  Title      :  Testbench for implementation Mini-MIPS project.
--             :
--  Developers :  Rune M. Barnkob (c973431)
--             :  Mikael Lund Jepsen (c973811)
--             :  Emil Heinze (c973816)
--             :
--  Purpose    :  This is a new test-bench using dual-port memory.
--             :
--  Revision   :  1.0 2001/01/18     Initial version
--             :  1.1 2001/01/24     Configurations added
--              
-- -----------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use WORK.types.all;

entity mips is
end mips;

architecture structure of mips is
  component clock 
    generic (period:        time := 50 ns);
    port (clk:              out std_logic := '0');
  end component clock;

  component MiniMIPS is port (
    clk   : in    bit_t;       -- global clock
    reset : in    bit_t;       -- global reset
    
    req   : out   bit_t;       -- request to memory
    rw    : out   bit_t;       -- read-not-write signal to memory
    hold  : in    bit_t;       -- hold from memory
    addr  : out   halfword_t;  -- address towards memory
    data  : inout halfword_t   -- data to send/read to/from memory
    );
  end component MiniMIPS;
  
  component mem2 is
    generic (filename:    string  := "";      -- Name of memory load file
             n:           natural := 0;       -- Number of hold cycles(wait states)
             Tclock:      time := 50 ns);     -- Length of clock-period.
    port (clk:        in  std_logic;              -- The clock.
          addr:       in  halfword_t;         -- The address bus.
          data:       inout halfword_t;             -- The data bus.
          rw:         in  std_logic;              -- Read/Write signal.
          req:        in  std_logic;              -- Request memory operation signal.
          hold:       out std_logic);             -- Hold signal.
  end component mem2;

  signal clk:      bit_t;
  signal reset:    bit_t;

  signal mem_addr: halfword_t;
  signal mem_d: halfword_t;
  signal mem_req: bit_t;
  signal mem_rw: bit_t;
  signal mem_hold: bit_t;

begin
  reset <= '0', '1' after 57 ns;
  -- the clock generator
  sysclk: clock port map (clk => clk);
  -- the CPU with bus interface
  processor: MiniMIPS port map (
    clk => clk,
    reset => reset,
    addr => mem_addr,
    data => mem_d,
    rw => mem_rw,
    req => mem_req,
    hold => mem_hold
  );
  -- the memory
  memory: mem2 port map (
    clk => clk,
    addr => mem_addr,
    data => mem_d,
    rw => mem_rw,
    req => mem_req,
    hold => mem_hold
  );
end structure;

-------------------------------------------------------------------------------- 
-- a configuration for the testbench. this configuration will make the cpu load
-- the file "fib2.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration t_fib of mips is
  for structure
      for sysclk: clock
          use entity work.clock(behaviour)
              generic map(period => 500 ns);
      end for;

--      for processor: MiniMIPS
--              use entity work.MiniMIPS(structure);
--      end for;

      for memory: mem2
          use entity work.mem2(behaviour)
              generic map(filename => "progs\fib2.bin",
                          n        => 2);
      end for;
  end for;
end t_fib;


-------------------------------------------------------------------------------- 
-- a configuration for the testbench. this configuration will make the cpu load
-- the file "pipetest.src.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration t_pipe of mips is
  for structure
      for sysclk: clock
          use entity work.clock(behaviour)
              generic map(period => 500 ns);
      end for;

--      for processor: MiniMIPS
--              use entity work.MiniMIPS(structure);
--      end for;

      for memory: mem2
          use entity work.mem2(behaviour)
              generic map(filename => "progs\pipetest.src.bin",
                          n        => 2);
      end for;
  end for;
end t_pipe;

-------------------------------------------------------------------------------- 
-- a configuration for the testbench. this configuration will make the cpu load
-- the file "memtest.src.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration t_mem of mips is
  for structure
      for sysclk: clock
          use entity work.clock(behaviour)
              generic map(period => 500 ns);
      end for;

--      for processor: MiniMIPS
--              use entity work.MiniMIPS(structure);
--      end for;

      for memory: mem2
          use entity work.mem2(behaviour)
              generic map(filename => "progs\memtest.src.bin",
                          n        => 2);
      end for;
  end for;
end t_mem;

-------------------------------------------------------------------------------- 
-- a configuration for the testbench. this configuration will make the cpu load
-- the file "memtest.src.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration t_fwd of mips is
  for structure
      for sysclk: clock
          use entity work.clock(behaviour)
              generic map(period => 500 ns);
      end for;

--      for processor: MiniMIPS
--              use entity work.MiniMIPS(structure);
--      end for;

      for memory: mem2
          use entity work.mem2(behaviour)
              generic map(filename => "progs\fwdtest.src.bin",
                          n        => 2);
      end for;
  end for;
end t_fwd;

-------------------------------------------------------------------------------- 
-- a configuration for the testbench. this configuration will make the cpu load
-- the file "memtest.src.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration t_loop of mips is
  for structure
      for sysclk: clock
          use entity work.clock(behaviour)
              generic map(period => 500 ns);
      end for;

--      for processor: MiniMIPS
--              use entity work.MiniMIPS(structure);
--      end for;

      for memory: mem2
          use entity work.mem2(behaviour)
              generic map(filename => "progs\loop.src.bin",
                          n        => 2);
      end for;
  end for;
end t_loop;

-------------------------------------------------------------------------------- 
-- a configuration for the testbench. this configuration will make the cpu load
-- the file "slt.src.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration t_slt of mips is
  for structure
      for sysclk: clock
          use entity work.clock(behaviour)
              generic map(period => 500 ns);
      end for;

--      for processor: MiniMIPS
--              use entity work.MiniMIPS(structure);
--      end for;

      for memory: mem2
          use entity work.mem2(behaviour)
              generic map(filename => "progs\slt.src.bin",
                          n        => 2);
      end for;
  end for;
end t_slt;