-- -----------------------------------------------------------------------------
--
--  Title      :  Useful types and constants in a nice package.
--             :
--  Developers :  Niels Haandbk -- c958307@student.dtu.dk
--             :
--  Purpose    :  This design contains a package with types and constants. The
--             :  types are std_logic_vector's with names and sizes that 
--             :  correspond to the standard MIPS notation.
--             :
--  Revision   :  1.0    27-8-99     Initial version
--              
-- -----------------------------------------------------------------------------

--------------------------------------------------------------------------------
--    Type name |  MIPS name | size in bits 
--        bit_t |     bit    | 1
--       byte_t |    byte    | 8
--   halfword_t |  halfword  | 16
--       word_t |    word    | 32
-- doubleword_t | doubleword | 64
-- The constants can be used to set all bits in a signal or variable of type 
-- byte_t, halfword_t, word_t and doubleword_t to either '0', '1', 'X' or 'Z'.
--------------------------------------------------------------------------------

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE STD.textio.ALL;

PACKAGE types IS
    SUBTYPE bit_t           IS std_logic;
    SUBTYPE byte_t          IS std_logic_vector(7 DOWNTO 0);
    SUBTYPE halfword_t      IS std_logic_vector(15 DOWNTO 0);
    SUBTYPE word_t          IS std_logic_vector(31 DOWNTO 0);
    SUBTYPE doubleword_t    IS std_logic_vector(63 DOWNTO 0);
    
    CONSTANT byte_zero:     byte_t := "00000000";
    CONSTANT byte_one:      byte_t := "11111111";
    CONSTANT byte_x:        byte_t := "XXXXXXXX";
    CONSTANT byte_z:        byte_t := "ZZZZZZZZ";
    
    CONSTANT halfword_zero: halfword_t := byte_zero & byte_zero;
    CONSTANT halfword_one:  halfword_t := byte_one & byte_one;
    CONSTANT halfword_x:    halfword_t := byte_x & byte_x;
    CONSTANT halfword_z:    halfword_t := byte_z & byte_z;
    
    CONSTANT word_zero:     word_t := halfword_zero & halfword_zero;
    CONSTANT word_one:      word_t := halfword_one & halfword_one;
    CONSTANT word_x:        word_t := halfword_x & halfword_x;
    CONSTANT word_z:        word_t := halfword_z & halfword_z;
    
    CONSTANT doubleword_zero:   doubleword_t := word_zero & word_zero;
    CONSTANT doubleword_one:    doubleword_t := word_one & word_one;
    CONSTANT doubleword_x:      doubleword_t := word_x & word_x;
    CONSTANT doubleword_z:      doubleword_t := word_z & word_z;

    -- Mini-MIPS specific types and constants.
    SUBTYPE OpCode          IS std_logic_vector(2 DOWNTO 0);
    SUBTYPE FuncCode        IS std_logic_vector(3 DOWNTO 0);

    SUBTYPE MemAddr         IS INTEGER RANGE 0 TO 2**16 - 1;
    TYPE MemType            IS ARRAY(MemAddr) OF halfword_t;

    SUBTYPE RegAddr         IS INTEGER RANGE 0 TO 7;
    TYPE RegType            IS ARRAY(RegAddr) OF halfword_t;

    -- These constants enable easier-to-read identification of 
    -- instructions.
    CONSTANT op_Rformat:    OpCode := "000";
    CONSTANT op_addi:       OpCode := "001";
    CONSTANT op_j:          OpCode := "010";
    CONSTANT op_jal:        OpCode := "011";
    CONSTANT op_beq:        OpCode := "100";
    CONSTANT op_bne:        OpCode := "101";
    CONSTANT op_lw:         OpCode := "110";
    CONSTANT op_sw:         OpCode := "111";
    CONSTANT funct_sll:     FuncCode := "0000";
    CONSTANT funct_srl:     FuncCode := "0001";
    CONSTANT funct_sra:     FuncCode := "0010";
    CONSTANT funct_jr:      FuncCode := "0100";
    CONSTANT funct_add:     FuncCode := "1000";
    CONSTANT funct_sub:     FuncCode := "1001";
    CONSTANT funct_and:     FuncCode := "1010";
    CONSTANT funct_or: 	    FuncCode := "1011";
    CONSTANT funct_xor:     FuncCode := "1100";
    CONSTANT funct_slt:     FuncCode := "1110";
    CONSTANT funct_sltu:    FuncCode := "1111";
    
    CONSTANT Tpd: 			TIME := 2 ns;
    
    -- This procedure loads the contents of "filename" into the memory.
    PROCEDURE LoadMemoryFromFile(VARIABLE memory: OUT MemType;
                                 CONSTANT filename: IN STRING);
END types;

--------------------------------------------------------------------------------
-- A body for the types package. It implements the LoadMemoryFromFile 
-- procedure.
--------------------------------------------------------------------------------

PACKAGE body types IS

    PROCEDURE LoadMemoryFromFile(VARIABLE memory: OUT MemType;
                                 CONSTANT filename: IN STRING) IS
        FILE memfile:     TEXT OPEN READ_MODE IS filename;
        VARIABLE memline: LINE;
        VARIABLE tmp:     BIT_VECTOR(15 downto 0);    
    BEGIN
        FOR i IN memory'RANGE LOOP
            EXIT WHEN ENDFILE(memfile);
            READLINE(memfile, memline);
            READ(memline, tmp);
            memory(i) := TO_STDLOGICVECTOR(tmp);
        END LOOP;
    END LoadMemoryFromFile;

END types;    
