/////////////////////////////////////////////////////////////////////////////
// asm.cpp
// -------
//
// Assembler for the MiniMIPS. For more info, see asm.h
//
//
// The grammar is:
// ---------------
// 
// See "scanner.h" for definition of <ident>, <const> and more...
//
// *** EXPRESSIONS ***
// E0    ::= ["-"] <E1> { ("+" | "-") <E1> }
// E1    ::= <E2> { "*" <E2> }
// E2    ::= <value> | "(" <E0> ")"
// value ::= <const> | <ident(type==const)> | "@" <ident(type!=const)>
//
// *** INSTRUCTIONS ***
// <mne_2regtype>    ::= "sll" | "srl" | "sra" 
// <mne_3regtype>    ::= "add" | "sub" | "and" | "or" | "xor" | "slt" | "sltu"
// <mne_immtype>     ::= "addi"
// <mne_memtype>     ::= "lw" | "sw"
// <mne_branchtype>  ::= "beq" | "bne"
// <mne_jumptype>    ::= "j" | "jal"
// <mne_jumpregtype> ::= "jr"
//
// <2regtype>        ::= <mne_2regtype> <reg> "," <reg>
// <3regtype>        ::= <mne_3regtype> <reg> "," <reg> "," <reg>
// <immtype>         ::= <mne_immtype> <reg> "," <reg> "," <E0>
// <memtype>         ::= <mne_memtype> <reg> "," <ident(type==word/ascii/asciiz>"("<reg>")"
// <branchtype>      ::= <mne_branchtype> <reg> "," <reg> "," <ident(type==label)>
// <jumptype>        ::= <mne_jumptype> <ident(type==label)> 
// <jumpregtype>     ::= <mne_jumpregtype> <reg>
//
// <instruction>     ::= <2regtype> | <3regtype> | <immtype> | <memtype> | <branchtype> | <jumptype> | <jumpregtype>
//
//
// *** LABELS ***
// value_list ::= <E0> { "," <E0> }
// storage    ::= "." "word" <value_list> | "." "asciiz" <string> | "." "ascii" <string>
// label      ::= <ident> ":" [ <storage> | <instruction> ]
// constant   ::= <ident> "=" <E0>
//
// *** DIRECTIVE ***
// directive ::= "seg" <ident> | "end" | "code" | "data" | "org" <E0 resolvable>
//
// *** ASSEMBLER LINES ***
//	comment		::= "#"<rest-of-line>
//	line			 ::= [ <label> | <inst> ] [ <comment> ] 
//
#include <string>
#include "asm.h"
#include "error.h"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <strstream>
#include "disasm.h"
#define DEBUG(s) //(clog << s << endl << flush)


#define MISS_SEGMENT_HEADERS
#define VSYSTEM

string lalign(int isize, string i, char fill=' ') {
	string res;
	int lp1;
	for (lp1=0; lp1<isize; lp1++)
		res += (lp1<i.length()?i[lp1]:fill);
	return res;
}


// groups:
// E_ - general
// EL_ - label errors
// ER_ - register-type-instructions errors
// EI_ - immediate-type-instructions errors
// EM_ - memory-type-instructions errors
// EJ_ - jump-"-"- errors
// EB_ - branch-"-" errors
// ED_ - directive errors
#define E_UNKNOWNTOKEN 1
#define E_ADDITEMS 2
#define E_NOTANINST 4
#define E_REGEXPECTED 5
#define E_INVALIDREG 6
#define E_UNDECLARED 7
#define E_CONSTEXPECTED 8
#define E_IDENTEXPECTED 9
#define E_CANTTAKEADDR 10
#define E_VALUEEXPECTED 11
#define E_MISSINGPAR 12

#define EL_DUPLICATEIDENT 100
#define EL_DISPLACEMENTERROR 101
#define EL_COUNTEXPECTED 102
#define EL_VALLISTCOMMASEP 103
#define EL_UNPARSED 104
#define EL_INVALIDSTORAGE 105
#define EL_EXPECTEDSTRING 106

#define ER_INV 200
#define ER_COMMAEXPECTED 201

#define EI_INV 300
#define EI_COMMAEXPECTED 301
#define EI_OUTOFRANGE 302

#define EM_INV 400
#define EM_COMMAEXPECTED 401
#define EM_TYPEERROR 402
#define EM_UNREACHABLE 403
#define EM_MISSINGBASEREG 404

#define EB_INV 500
#define EB_COMMAEXPECTED 501
#define EB_UNREACHABLE 502
#define EB_TARGETNOTLABEL 503
#define EB_INTERSEGMENT 504
#define EB_MISSINGDEST 505

#define EJ_INV 600
#define EJ_UNREACHABLE 601
#define EJ_TARGETNOTLABEL 602
#define EJ_INTERSEGMENT 603
#define EJ_MISSINGDEST 604

#define ED_INV 900
#define ED_SEGNAMEEXPECTED 901
#define ED_SEGNAMEDUPLICATE 902
#define ED_IDENTEXPECTED 903
#define ED_UNKNOWN 904
#define ED_ASSUMEFOLLOWREG 905
#define ED_COLONBETWEENASSUME 906
#define ED_IDNOTSEG 907
#define ED_ASSUMESYNTAX 908
#define ED_ORIGINREQ 909

class ESyntax: public Error {
public:

	ESyntax(int se): Error(se) {}

	ESyntax(int se, string message ): Error(se) {}


	string getMsg() {
		ostrstream msg;
		msg << "STX" << setw(3) << setfill('0') << ecode << ": ";

		switch (ecode) {
			case E_UNKNOWNTOKEN: msg << "Unknown token"; break;
			case E_ADDITEMS: msg << "additional items on line, end-of-line expected"; break;
			case E_NOTANINST: msg << "not an instruction"; break;
			case E_REGEXPECTED: msg << "register expected"; break;
			case E_INVALIDREG: msg << "invalid register!"; break;
			case E_UNDECLARED: msg << "undeclared identifier!"; break;
			case E_CONSTEXPECTED: msg << "identifier must be constant!"; break;
			case E_IDENTEXPECTED: msg << "identifier expected"; break;
			case E_CANTTAKEADDR: msg << "cannot take address of constant"; break;
			case E_VALUEEXPECTED: msg << "value expected"; break;
			case E_MISSINGPAR: msg << "missing )"; break;
			
			case EL_DUPLICATEIDENT: msg << "duplicate identifier"; break;
			case EL_DISPLACEMENTERROR: msg << "label address mismatch in second pass!"; break;
			case EL_COUNTEXPECTED: msg << "* must be followed by a count"; break;
			case EL_VALLISTCOMMASEP: msg << "list of values must be seperated by ,"; break;
			case EL_UNPARSED: msg << "identifier must be followed by either : or ="; break;
			case EL_INVALIDSTORAGE: msg << "storage definition invalid - must be word, ascii or asciiz"; break;
			case EL_EXPECTEDSTRING: msg << ".ascii or .asciiz must be followed by a string"; break;
			
			case ER_INV: msg << "trying to parse as register-type instruction when it is not"; break;
			case ER_COMMAEXPECTED: msg << "expected , between operands"; break;
			
			case EI_INV: msg << "trying to parse as one-reg when it is not"; break;
			case EI_COMMAEXPECTED: msg << "expected , between operands"; break;
			case EI_OUTOFRANGE: msg << "the expression for the immediate is out of range"; break;
			
			case EM_INV: msg << "trying to parse as load/store when it is not"; break;
			case EM_COMMAEXPECTED: msg << "expected , between operands"; break;
			case EM_TYPEERROR: msg << "identifier not of memory cell type"; break;
			case EM_UNREACHABLE: msg << "identifier unreachable (offset > +/-64)"; break;
			case EM_MISSINGBASEREG: msg << "missing baseregister for offset"; break;
			
			case EB_INV: msg << "trying to parse as bitman when it is not"; break;
			case EB_COMMAEXPECTED: msg << "expected , between operands"; break;
			case EB_UNREACHABLE: msg << "branch target unreachable ( > +/- 63)"; break;
			case EB_MISSINGDEST: msg << "no destination supplied for branch"; break;
			case EB_TARGETNOTLABEL: msg << "identifier not defined as label"; break;
			case EB_INTERSEGMENT: msg << "inter-segment branch not supported"; break;
			
			case EJ_INV: msg << "trying to parse as jump when it is not"; break;
			case EJ_UNREACHABLE: msg << "jump target unreachable (outside 8k boundaries)"; break;
			case EJ_TARGETNOTLABEL: msg << "identifier not defined as label"; break;
			case EJ_INTERSEGMENT: msg << "inter-segment jumps not supported"; break;
			case EJ_MISSINGDEST: msg << "no destination supplied for jump"; break;

			case ED_INV: msg << "trying to parse directive that does not start with ."; break;
			case ED_SEGNAMEEXPECTED: msg << "segment name expected after .seg"; break;
			case ED_SEGNAMEDUPLICATE: msg << "segment name must be unique in file"; break;
			case ED_IDENTEXPECTED: msg << "identifier expected"; break;
			case ED_UNKNOWN: msg << "unknown assembler directive"; break;
			case ED_ASSUMEFOLLOWREG: msg << ".assume must be followed by register"; break;
			case ED_ASSUMESYNTAX: msg << ".assume must be followed by <register>:<segname>"; break;
			case ED_COLONBETWEENASSUME: msg << "register and segment name must be separated by :"; break;
			case ED_IDNOTSEG: msg << "identifier not segment"; break;
			case ED_ORIGINREQ: msg << "new origin must be resolvable in first pass"; break;
			
			default: msg << "???";
		}
		if (emsg!="") msg << " - " << emsg;
		msg << ends;
		return msg.str();
	}
};



//////////////////////////////////////////////////////////////////////////////
//
// Identifiers:
//
class Ident {
	public:
		// type of this identifier
		enum IdentType { ID_CONST, ID_LABEL, ID_WORD, ID_ASCII, ID_ASCIIZ };
		IdentType type;
		// name of this identifier
		string name;
		// address of this identifier (or value if constant)
		WORD addr;
		// segment in which this identifier is located
		pSeg segm;

		Ident(): name(""), addr(0), type(ID_CONST), segm(NULL) {}

		Ident( string nname, IdentType ttype, WORD aaddr=0, pSeg ssegm=NULL ):
			name(nname), addr(aaddr), type(ttype), segm(ssegm) {}
};



//////////////////////////////////////////////////////////////////////////////
//
// Segment:
//
// Segments are used to implement the different storage-areas in a file.
// Each segment has a name, a counter and a lot of data.
// The segment is based on WORD's - all indexes are counted in words
// 
class Segment {
	public:
		// create a new segment with the name string
		Segment(string);
		// free the segment after use
		virtual ~Segment();

		// returns the current index in the segment
		WORD getCount() { return count; }
		
		// set the segment counter
		void setCount(WORD newcount);
		
		// returns the largest address used in the segment
		WORD getMaxAddr() { return maxcount; }
		// returns the size of the segment
		WORD getSize() { return size; }

		// returns name of segment
		string getName() { return name; }

		// increases segment size by count. Cannot be done when the segment has
		// been realised!
		WORD inc( WORD=1 );

		// realize segment. Only when the segment is realized can data be added
		// to the segment. If realisation fails, an exception may be thrown.
		void realize();

		// access data-area. Throws exception if not realized.
		void set( WORD, WORD );
		WORD get( WORD addr );
		void put( WORD );

		// Will write segment to stream in format:
		// struct segment {
		//	 CHAR segname[32];
		//	 WORD seglength;
		//	 WORD segdata[seglength];
		// }
		void writeToStream(ostream& out);
		
		// Will write an image suitable for loading into V-Systems (lines of 16-bit numbers)
		void writeImage(ostream& out);

		// use disassembler to disassemble segment
		void view( ostream& o );

	private:
		string name;
		WORD count;
		WORD* data;
		WORD size;
		WORD maxcount;
};

// Creating the segment initialises all variables to 0 (or null)
Segment::Segment(string nname): name(nname), count(0), data(NULL), maxcount(0) {}

// Frees the data (if any)
Segment::~Segment() {
	if (data) delete data;
}

// Increases the segment pointer by size (=1) 
WORD Segment::inc( WORD size=1 ) {
	WORD oldcount = count;
	count += size;
	return oldcount;
}

// Moves the segment counter to the new position (remenbers to update the largest
// pointer used (for later realisation)
void Segment::setCount(WORD newcount) { 
	if (count>maxcount) maxcount=count; 
	count = newcount;
	if (count>maxcount) maxcount=count; 
}

// Allocates the storage required for this segment. Fails if already realised.
void Segment::realize() {
	if (data) throw Error(100,"Segment::realize() - Segment already realized");
	if (count>maxcount) maxcount = count;
	size = maxcount;
	data = new WORD[size+1];
	// reset memory area
	WORD *d=data;
	for (int lp1=0; lp1<size; lp1++, d++)
		*d = 0;
	count = 0;
}

// Modifies a value at an address in the segment
void Segment::set( WORD addr, WORD dat ) {
	if (!data) throw Error(101, "Segment::set() - Segment not realized: "+name);
	if (addr>size) throw Error(102, " Segment::set() - Out of bounds in segment: "+name);
	*(data+addr)=dat;
}

// Gets value at an address in the segment
WORD Segment::get( WORD addr ) {
	if (!data) throw Error(101, "Segment::get() - Segment not realized: "+name);
	if (addr>size) throw Error(102, "Segment::get() - Out of bounds in segment: "+name);
	return *(data+addr);
}

// Adds the value to the segment and moves the segment counter
void Segment::put( WORD w ) {
	if (data) {
		if (count>size) throw Error(102, "Segment::put() - Out of bounds in segment: "+name);
		else *(data+count) = w;
	}
	count++;
}

// Write a segment to a stream. The format is:
// struct segment {
//	 CHAR segname[32];
//	 WORD seglength;
//	 WORD segdata[seglength];
// }
void Segment::writeToStream(ostream& out) {
	SegHeader segh;
	for (int lp1=0; lp1<32; lp1++)
		 segh.name[lp1] = (lp1<name.length()? name[lp1] : 0);
	segh.length = size;
	out.write(&segh,sizeof(segh));
//	out.write(data,segh.length*sizeof(WORD));
	for (int lp1=0; lp1<segh.length; lp1++) {
		char h = (data[lp1]&0xFF00)>>8;
		char l = data[lp1]&0xFF;
		out.write(&h,sizeof(char));
		out.write(&l,sizeof(char));
	}
}

// Instead of writing the data in the segment in a binary format,
// this method will write the segment as an image composed of lines,
// each containing a 16-bit number (this format is supported by
// V-Systems)
void Segment::writeImage(ostream& out) {
	WORD v;
	char s[17];
	s[16] = 0;
	for (int lp1=0; lp1<size; lp1++) {
		v = data[lp1];
		for (int lp2=0; lp2<16; lp2++) {
			s[lp2] = (v&0x8000?'1':'0');
			v <<= 1;
		}
		out << s << endl;
	}
}

// Dumps a disassembly of the 
void Segment::view( ostream& o) {
	if (!data) throw Error(103, "Segment::view() - Cannot view unrealized segment: "+name);
	if (name!="data") {
		for (WORD lp1=0; lp1<size; lp1++)
			o << DisAsm(data[lp1], lp1) << endl;
	} else {
		WORD pos = 0;
		while (pos < size) {
			o << hex << setw(4) << setfill('0') << pos << "h:\t";
			for (WORD lp1=0; lp1<8 && pos<size; lp1++, pos++)
				o << " 0x" << hex << setw(4) << setfill('0') << data[pos];
			o << endl;
		}				
	}
	
}

//////////////////////////////////////////////////////////////////////////////
// ASSEMBLER
//

// The constructor for the assembler will initialize all fields
Assembler::Assembler() {
	segments["code"] = new Segment("code");
	segments["data"] = new Segment("data");
	curseg = segments["code"];
	passCount = psFirst;
}


// clear all segments and idents.
Assembler::~Assembler() {
	map<string, pSeg>::iterator segit = segments.begin();
	while (segit!=segments.end()) {
		delete segit->second;
		segit++;
	}
	segments.clear();
	idents.clear();
}

// Move to the next token in the input
void Assembler::nextToken() {
	curtok = sc.next();
}


// Look-up identifier in identifier database
Ident Assembler::lookup(string id) {
	map<string,Ident>::iterator l = idents.find(id);
	if (l==idents.end())
		throw ESyntax(E_UNDECLARED, id);
	return l->second;
}


// Check for identifier in both segment- and ident-lists!
bool Assembler::idExists(string id) {

	map<string,Ident>::iterator l = idents.find(id);
	if (l!=idents.end())
		return true;

	map<string,pSeg>::iterator s = segments.find(id);
	if (s!=segments.end())
		return true;

	return false;
}


// The main method for the assembler. Opens the "infile", parses it, realizes 
// segments, parses it again, writes data to "outfile" and a map is written to
// "mapfile"
int Assembler::assemble( string infile, string outfile, string mapfile ) {
	char buf[256];
	ifstream in(infile.c_str());
	bool errors;


	// first pass
	cout << "MiniMIPS assembler (2000-11-09) Rune M. Barnkob (c973431@student.dtu.dk)" << endl;
	cout << "------------------------------------------------------------------------" << endl;
	cout << "Source:	 " << infile << endl;
	cout << "Object:	 " << outfile << endl;
	cout << "Map file: " << mapfile << endl;
	cout << endl;

	cout << "First pass..." << endl;

	eof = false;
	errors = false;
	lineno = 0;
	while (!in.eof() && !eof) try {
			in.getline(buf,256);
			lineno++;
			prsLine(buf);
		} catch (ESyntax& e) {
			errors=true;
			cerr << infile << "(" << dec << lineno << "): " << e.getMsg() << endl;
		} catch (Error& e) {
			cerr << infile << "(" << dec << lineno << "): FATAL: " << e.getMsg() << endl;
			return -1;
		}
	if (errors) return -1;

	cout << "No syntax errors detected in first pass!" << endl << endl;

	// realize all segments
	cout << "Realizing segments in memory...";
	for (map<string,pSeg>::iterator p=segments.begin(); p!=segments.end(); p++)
		 p->second->realize();
	cout << "done" << endl << endl;

	// second pass
	ofstream fmapfile;
	bool makemapfile = mapfile!="";
	if (makemapfile) {
		fmapfile.open(mapfile.c_str());
		fmapfile << "FILE:	 " << mapfile << endl;
		fmapfile << "SOURCE: " << infile << endl;
		fmapfile << "OBJECT: " << outfile << endl;
		fmapfile << lalign(80,"",'-') << endl << endl;
		fmapfile << "* Source-listing and segment counter: " << endl << endl;
		fmapfile << lalign(80,"",'-') << endl << endl;
	}

	cout << "Second pass - final assembly..." << endl;
	in.seekg(0);
	in.clear();
	lineno = 0;
	passCount = psFinal;
	curseg = segments["code"];
	eof = false;
	try {
		while (!in.eof() && !eof) {
			WORD pos = curseg->getCount();
			in.getline(buf,256);
			lineno++;
			if (makemapfile) 
				fmapfile << curseg->getName() << ":$" << hex << setw(4) << setfill('0') << pos << ">" << buf << endl;
			prsLine(buf);
			if (makemapfile && pos!=curseg->getCount() && (signed)curseg->getCount()-pos<16 ) {
				fmapfile << "\t";
				for (WORD lp1=pos; lp1<curseg->getCount(); lp1++)
					fmapfile << "0x" << hex << setw(4) << setfill('0') << curseg->get(lp1) << " ";
				fmapfile << endl;
			}
		}
	} catch (ESyntax& e) {
			cerr << infile << "(" << dec << lineno << "): " << e.getMsg() << endl;
		return -1;
	}
	in.close();
	cout << endl;

	cout << "Writing object file...";
	fstream out(outfile.c_str(), ios::out );
	for (map<string,pSeg>::iterator p=segments.begin(); p!=segments.end(); p++)
		 p->second->writeImage(out);
	out.close();
	cout << "done" << endl << endl;

	if (makemapfile) {
		fmapfile << endl;

		fmapfile << lalign(80,"",'-') << endl << endl;
		fmapfile << "* Identifiers: " << endl << endl;
		fmapfile << lalign(80,"",'-') << endl << endl;

		fmapfile << lalign(20,"Name:") << " ";
		fmapfile << lalign(10,"Segment:") << " ";
		fmapfile << lalign(10,"Address:") << " ";
		fmapfile << lalign( 5,"Type:") << endl << endl;

		string idtype[] = { "const","label","word","ascii","asciiz"};
		for (map<string, Ident>::iterator i=idents.begin(); i!=idents.end(); i++) {
			fmapfile << lalign(20,i->second.name) << " ";
			fmapfile << lalign(10,(i->second.segm!=NULL?i->second.segm->getName():string("-"))) << " ";
			fmapfile << setw(4) << hex << setfill('0') << i->second.addr << setfill(' ') << "h	    ";
			fmapfile << lalign(5,idtype[i->second.type]) << endl;
		}
		fmapfile << endl;

		fmapfile << lalign(80,"",'-') << endl << endl;
		fmapfile << "* Segments: " << endl << endl;
		fmapfile << lalign(80,"",'-') << endl << endl;
		for (map<string,pSeg>::iterator p=segments.begin(); p!=segments.end(); p++) {
			fmapfile << "Segment: " << p->second->getName() << "    Size: " << dec << p->second->getSize() << endl << endl;
			p->second->view(fmapfile);
			fmapfile << endl;
		}
		fmapfile << endl << "----==== EOF ====----" << endl;
		fmapfile.close();
	}

	cout << "Done assembling..." << endl;
	return 0;
}


// Start parsing a line. Look for <ident> and handle as a label. Look for "." and handle as directive.
// branches into:
void Assembler::prsLine( string line ) {
	sc.newLine(line);
	nextToken();

	// empty line
	switch (curtok) {
		case TT_EOL: // empty line
			return;
			break;
			
		case TT_DOT: // directive
			prsDirective();
			break;
			
		case TT_ID:  // label
			prsLabel();
			if (curtok!=TT_EOL) prsInst();
			break;
			
		default: // instruction
			prsInst();
	}
			
}


// <directive> ::= "." <dir>
void Assembler::prsDirective() {
	DEBUG("Assembler::prsDirective()");

	if (curtok!=TT_DOT) throw ESyntax(ED_INV);
	nextToken();

	switch (curtok) {
		// "." "seg" <id>
		case TT_SEG: {
			nextToken();
			switch (curtok) {
				case TT_DATA:
					curseg = segments["data"];
					break;
					
				case TT_CODE:
					curseg = segments["code"];
					break;
					
				case TT_ID: {
					string id = sc.getLex();
					if (idents.find(id)!=idents.end()) throw ESyntax(ED_SEGNAMEDUPLICATE, id);
					if (segments[id]==NULL) segments[id] = new Segment(id);
					curseg=segments[id];
					break;
				}
				
				default:
					throw ESyntax(ED_IDENTEXPECTED);
			}
			break;
		}
		
		// .end
		case TT_END: 
			eof = true;
			break;
			
		// .data
		case TT_DATA:
			curseg = segments["data"];
			break;
		
		// .code
		case TT_CODE:
			curseg = segments["code"];
			break;
			
		// .org
		case TT_ORG:
			nextToken();
			try { 
				WORD v = prsValue(true);
				curseg->setCount(v);
			} catch (ESyntax) {
				throw ESyntax(ED_ORIGINREQ);
			}
			break;
			
		// .word / .ascii / .asciiz
		case TT_WORD: case TT_ASCII: case TT_ASCIIZ:
			prsStorage("");
			break;
		
		default:
			throw ESyntax(ED_UNKNOWN);
	}
}




// parses an expanded value-expression
// <E0> ::= <E1> { ("+"|"-") <E1> }
INT Assembler::prsExpr0( bool resolve ) {
	DEBUG("Assembler::prsExpr0()");
	WORD v = 0;
	if (curtok==TT_MINUS) {
		nextToken();
		v = -prsExpr1( resolve );
	} else v = prsExpr1( resolve );
	while (curtok==TT_PLUS || curtok==TT_MINUS) {

		if (curtok==TT_PLUS) {
			nextToken();
			v += prsExpr1( resolve );
		}
		if (curtok==TT_MINUS) {
			nextToken();
			v -= prsExpr1( resolve );
		}
	}
	return v;
}


// parses an expanded value-expression
// <E1> ::= <E2> { "*" <E2> }
INT Assembler::prsExpr1( bool resolve ) {
	DEBUG("Assembler::prsExpr1()");
	WORD v = prsExpr2( resolve );
	while (curtok==TT_MULT) {
		nextToken();
		v *= prsExpr2( resolve );
	}
	return v;
}


// parses an expanded value-expression
// <E2> ::= "(" <E0> ")" | <value>
INT Assembler::prsExpr2( bool resolve ) {
	DEBUG("Assembler::prsExpr2()");
	WORD v;
	if (curtok==TT_LP) {
		nextToken();
		v = prsExpr0( resolve );
		if (curtok!=TT_RP) throw ESyntax(E_MISSINGPAR);
		nextToken();
	} else
		v = prsValue( resolve );
	return v;
}


// This method parses a simple value -
// <value> ::= <number> | <ident(type==const)> | "@" <ident(type!=const)>
INT Assembler::prsValue( bool resolve ) {
	DEBUG("Assembler::prsValue()");
	WORD v;
	switch (curtok) {
		// read constant
		case TT_CONST:
			v = sc.getWord();
			break;
 
		// read identifier (must be constant)
		case TT_ID:
			if (resolve) {
				Ident i = lookup(sc.getLex());
				if (i.type!=Ident::ID_CONST) throw ESyntax(E_CONSTEXPECTED, i.name);
				v = i.addr;
			} else
				v = 0;
			break;
 
		// read absolute address
		case TT_AT:
			nextToken();
			if (curtok!=TT_ID) throw ESyntax(E_IDENTEXPECTED, sc.getLex());
			if (resolve) {
				Ident i = lookup(sc.getLex());
				if (i.type==Ident::ID_CONST) throw ESyntax(E_CANTTAKEADDR, i.name);
				v = i.addr;
			} else
				v = 0;
			break;

		default: throw ESyntax(E_VALUEEXPECTED);
	}
	nextToken();
	return v;
}


// parses a register - returns regno.
int Assembler::prsReg() {
	DEBUG("Assembler::prsReg()");
	if (curtok!=TT_REG) throw ESyntax(E_REGEXPECTED);

	WORD regno=sc.getWord();
	nextToken();
	
	if (regno>7) throw ESyntax(E_INVALIDREG);
	return regno;
}


// <constant> ::= <ident> "=" <E0>
// <label>    ::= <ident> ":" [ <storage> | <instruction> ] | <constant>
void Assembler::prsLabel() {
	string id = sc.getLex();
	DEBUG("Assembler::prsLabel() ");
	
	// avoid duplicate identifiers
	if (passCount==psFirst && idExists(id) ) throw ESyntax(EL_DUPLICATEIDENT, id);
	nextToken();
	
	// constant - must read number after
	if (curtok==TT_EQ) {
		DEBUG(" - found constant: " << id);
		nextToken();
		WORD val = prsExpr0( true ); // must resolve to immediate value!
		// add constant to idents if first-pass.
		idents[id] = Ident(id, Ident::ID_CONST, val);

		if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
		return;
	}
	// label - expects instruction or storagespecifier
	if (curtok==TT_COLON) {
		DEBUG(" - found label: " << id);

		WORD adr = curseg->getCount();

		if (passCount!=psFinal) idents[id] = Ident(id, Ident::ID_LABEL, adr, curseg );
		if (passCount==psFinal) if (adr!=idents[id].addr) throw ESyntax(EL_DISPLACEMENTERROR, id);
		nextToken();
		// catch storage-specifiers
		if (curtok==TT_DOT) {
			DEBUG("	 - found storage");
			nextToken();
			prsStorage(id);
			return ;
		}
		// a storage-specifier did not follow the label, so an instruction may
		// be here
		DEBUG(" - no storage-specifier, crude label");
		return ;
	} 
	throw ESyntax(EL_UNPARSED);
}


// <storage> ::= "word" ["*" <E2> | <value_list>] | "ascii" <string> | "asciiz" <string>
// <value_list> ::= <E0_string> { "," <number_string> }
// <E0_string> ::= <E0> | <string>
void Assembler::prsStorage(string id) {
	DEBUG("Assembler::prsStorage("<<id<<")");
	switch (curtok) {
		case TT_WORD: 
			if (passCount!=psFinal && id!="") idents[id].type=Ident::ID_WORD;
			nextToken();
			// look for "*" <E2 that is resolvable>
			if (curtok==TT_MULT) {
				nextToken();
				WORD count = prsExpr2(true);
				curseg->inc(count);
				nextToken();
				if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
				return;
			}
			// look for value list (more on line)
			if (curtok!=TT_EOL) {
				WORD offset = 0;
				while (curtok!=TT_EOL) {
					if (curtok==TT_STRING) {
						string s = sc.getLex();
						DEBUG("In vallist: found the string: " << s );
						for (int lp1=0; lp1<s.length(); lp1++)
							curseg->put( static_cast<CHAR>(s[lp1]) );
						nextToken();
					} else {
						WORD val = prsExpr0(passCount==psFinal);
						curseg->put( val );
					}
					if (curtok!=TT_COMMA && curtok!=TT_EOL) throw ESyntax(EL_VALLISTCOMMASEP);
					if (curtok==TT_COMMA) nextToken();
				}
			} else
			// if neither value-list or rep-count use default one instance uninitialised
			curseg->inc(1);
			break;
					
		case TT_ASCII: {
			string s;
			if (passCount!=psFinal && id!="") idents[id].type=Ident::ID_WORD;//Ident::ID_ASCII;
			nextToken();
			if (curtok!=TT_STRING) throw ESyntax(EL_EXPECTEDSTRING);
			s = sc.getLex();
			for (int lp1=0; lp1<s.length(); lp1++)
				curseg->put( static_cast<CHAR>(s[lp1]) );
			nextToken();
			break;
		}

		case TT_ASCIIZ: {
			string s;
			if (passCount!=psFinal && id!="") idents[id].type=Ident::ID_WORD;//Ident::ID_ASCIIZ;
			nextToken();
			if (curtok!=TT_STRING) throw ESyntax(EL_EXPECTEDSTRING);
			s = sc.getLex();
			for (int lp1=0; lp1<s.length(); lp1++)
				curseg->put( static_cast<CHAR>(s[lp1]) );
			curseg->put(0);
			nextToken();
			break;
		}

		default: throw ESyntax(EL_INVALIDSTORAGE);
	}

	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}



// <mne_2regtype>    ::= "sll" | "srl" | "sra" 
// <mne_3regtype>    ::= "add" | "sub" | "and" | "or" | "xor" | "slt" | "sltu"
// <mne_immtype>     ::= "addi"
// <mne_memtype>     ::= "lw" | "sw"
// <mne_branchtype>  ::= "beq" | "bne"
// <mne_jumptype>    ::= "j" | "jal"
// <mne_jumpregtype> ::= "jr"
//
// <2regtype>        ::= <mne_2regtype> <reg> "," <reg>
// <3regtype>        ::= <mne_3regtype> <reg> "," <reg> "," <reg>
// <immtype>         ::= <mne_immtype> <reg> "," <reg> "," <value>
// <memtype>         ::= <mne_memtype> <reg> "," <ident(type==word/ascii/asciiz>"("<reg>")"
// <branchtype>      ::= <mne_branchtype> <reg> "," <reg> "," <ident(type==label)>
// <jumptype>        ::= <mne_jumptype> <ident(type==label)> 
// <jumpregtype>     ::= <mne_jumpregtype> <reg>
//
// <instruction>     ::= <2regtype> | <3regtype> | <immtype> | <memtype> | <branchtype> | <jumptype> | <jumpregtype>
void Assembler::prsInst() {
	DEBUG("Assembler::prsInst()");
	switch (curtok) {
		case TT_SLL: case TT_SRL: case TT_SRA:
			prs2RegType();
			break;
			
		case TT_ADD: case TT_SUB: 
		case TT_AND: case TT_OR: case TT_XOR:
		case TT_SLT: case TT_SLTU:
			prs3RegType();
			break;
			
		case TT_ADDI:
			prsImmType();
			break;
			
		case TT_LW: case TT_SW:
			prsMemType();
			break;
			
		case TT_BEQ: case TT_BNE:
			prsBranchType();
			break;
			
		case TT_J: case TT_JAL:
			prsJumpType();
			break;
			
		case TT_JR:
			prsJumpRegType();
			break;
			
		default: throw ESyntax(E_NOTANINST);
	}
}


// <mne_2regtype> ::= "sll" | "srl" | "sra" 
// <2regtype>     ::= <mne_2regtype> <reg> "," <reg>
void Assembler::prs2RegType() {
	DEBUG("Assembler::prs2RegType()");
	oc o;
	o.opcode = 0;
	switch (curtok) {
		case TT_SLL:  o.reg.func =  0; break;
		case TT_SRL:  o.reg.func =  1; break;
		case TT_SRA:  o.reg.func =  2; break;
		default: throw ESyntax(ER_INV);
	}
	nextToken();
	o.reg.rd = prsReg();
	if (curtok!=TT_COMMA) throw ESyntax(ER_COMMAEXPECTED);
	nextToken();
	o.reg.rt = prsReg();

	curseg->put(o.opcode);
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}

// <mne_3regtype> ::= "add" | "sub" | "and" | "or" | "xor" | "slt" | "sltu"
// <3regtype>     ::= <mne_3regtype> <reg> "," <reg> "," <reg>
void Assembler::prs3RegType() {
	DEBUG("Assembler::prs3RegType()");
	oc o;
	o.opcode = 0;
	switch (curtok) {
		case TT_ADD:  o.reg.func =  8; break;
		case TT_SUB:  o.reg.func =  9; break;
		case TT_AND:  o.reg.func = 10; break;
		case TT_OR:   o.reg.func = 11; break;
		case TT_XOR:  o.reg.func = 12; break;
		case TT_SLT:  o.reg.func = 14; break;
		case TT_SLTU: o.reg.func = 15; break;
		default: throw ESyntax(ER_INV);
	}
	nextToken();
	o.reg.rd = prsReg();
	if (curtok!=TT_COMMA) throw ESyntax(ER_COMMAEXPECTED);
	nextToken();
	o.reg.rs = prsReg();
	if (curtok!=TT_COMMA) throw ESyntax(ER_COMMAEXPECTED);
	nextToken();
	o.reg.rt = prsReg();

	curseg->put(o.opcode);
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}


// <mne_immtype>     ::= "addi"
// <immtype>         ::= <mne_immtype> <reg> "," <reg> "," <E0>
void Assembler::prsImmType() {
	DEBUG("Assembler::prsImmType()");
	oc o;
	INT imm;
	
	o.opcode = 0;
	o.imm.op = 1;
	if (curtok!=TT_ADDI) throw ESyntax(EI_INV);
	nextToken();
	o.imm.rt = prsReg();
	if (curtok!=TT_COMMA) throw ESyntax(EI_COMMAEXPECTED);
	nextToken();
	o.imm.rs = prsReg();
	if (curtok!=TT_COMMA) throw ESyntax(EI_COMMAEXPECTED);
	nextToken();
	imm = prsExpr0(passCount==psFinal);
	if (imm>63 || imm<-64) throw ESyntax(EI_OUTOFRANGE);
	o.imm.imm = imm;

	curseg->put(o.opcode);
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}


// <mne_memtype>     ::= "lw" | "sw"
// <memtype>         ::= <mne_memtype> <reg> "," <ident(type==word/ascii/asciiz>"("<reg>")"
void Assembler::prsMemType() {
	DEBUG("Assembler::prsMemType()");
	oc o;
	INT offset;
	o.opcode = 0;
	switch (curtok) {
		case TT_LW:  o.mem.op = 6; break;
		case TT_SW:  o.mem.op = 7; break;
		default: throw ESyntax(EM_INV);
	}
	nextToken();
	o.mem.rt = prsReg();
	if (curtok!=TT_COMMA) throw ESyntax(EM_COMMAEXPECTED);
	nextToken();
	// ident, number
	if (curtok==TT_ID) {
		if (passCount==psFinal) {
			Ident i = lookup(sc.getLex());
			if (i.type==Ident::ID_WORD || i.type==Ident::ID_ASCII || i.type==Ident::ID_ASCIIZ) offset = i.addr;
			else throw ESyntax(EM_TYPEERROR);
		} else offset = 0;
		nextToken();
	} else offset = prsExpr2(passCount==psFinal);
	// look for base-register
	if (curtok==TT_LP) {
		nextToken();
		o.mem.rs = prsReg();
		if (curtok!=TT_RP) throw ESyntax(E_MISSINGPAR);
		nextToken();
	} else
		throw ESyntax(EM_MISSINGBASEREG);
		
 
 	if (passCount==psFinal && (offset>63 || offset<-64) ) throw ESyntax(EM_UNREACHABLE);
 	o.mem.offset = offset;
	curseg->put(o.opcode);
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}


// <mne_branchtype>  ::= "beq" | "bne"
// <branchtype>      ::= <mne_branchtype> <reg> "," <reg> "," <ident(type==label)>
void Assembler::prsBranchType() {
	DEBUG("Assembler::prsBranchType()");
	oc o;
	INT offset;
	
	switch (curtok) {
		case TT_BEQ: o.bra.op = 4; break;
		case TT_BNE: o.bra.op = 5; break;
		default: throw ESyntax(EB_INV);
	}
	nextToken();
	o.bra.rs = prsReg();
	if (curtok!=TT_COMMA) throw ESyntax(EB_COMMAEXPECTED);
	nextToken();
	o.bra.rt = prsReg();
	if (curtok!=TT_COMMA) throw ESyntax(EB_COMMAEXPECTED);
	nextToken();

	if (curtok==TT_EOL) throw ESyntax(EB_MISSINGDEST);
	if (curtok==TT_ID) {
		if (passCount==psFinal) {
			Ident i = lookup(sc.getLex());
			if (i.type!=Ident::ID_LABEL) throw ESyntax(EB_TARGETNOTLABEL);
			if (curseg!=i.segm) throw ESyntax(EB_INTERSEGMENT);
			offset = i.addr;
		} else offset=0;
		nextToken();
	} else {
		offset = prsExpr0(passCount==psFinal);
	}
	offset -= (curseg->getCount()+1);
	if (passCount==psFinal && (offset>63 || offset<-64) ) throw ESyntax(EB_UNREACHABLE);
	o.bra.offset = offset;
	curseg->put(o.opcode);
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}


// <mne_jumptype>    ::= "j" | "jal"
// <jumptype>        ::= <mne_jumptype> <ident(type==label)> 
void Assembler::prsJumpType() {
	oc o;
	WORD target;
	DEBUG("Assembler::prsJumpType()");
	o.opcode = 0;
	switch (curtok) {
		case TT_J: o.jmp.op = 2; break;
		case TT_JAL: o.jmp.op = 3; break;
		default: throw ESyntax(EJ_INV);
	}
	nextToken();
	
	if (curtok==TT_EOL) throw ESyntax(EB_MISSINGDEST);
	if (curtok==TT_ID) {
		if (passCount==psFinal) {
			Ident i = lookup(sc.getLex());
			if (i.type!=Ident::ID_LABEL) throw ESyntax(EJ_TARGETNOTLABEL);
			if (curseg!=i.segm) throw ESyntax(EJ_INTERSEGMENT);
			target = i.addr;
		} else target=0;
		nextToken();
	} else {
		target = prsExpr0(passCount==psFinal);
	}
	// check that upper 3 bits of target is equal to upper 3 bits of PC (segment counter)
	if (passCount==psFinal && (target&0xE000)!=(curseg->getCount()&0xE000)) throw ESyntax(EJ_UNREACHABLE);
	o.jmp.target = target&0x1FFF;
	curseg->put(o.opcode);
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}


// <mne_jumpregtype> ::= "jr"
// <jumpregtype>     ::= <mne_jumpregtype> <reg>
void Assembler::prsJumpRegType() {
	oc o;
	INT offset;
	DEBUG("Assembler::prsJumpRegType()");
	o.opcode = 0;
	if (curtok!=TT_JR) throw ESyntax(EJ_INV);
	nextToken();
	o.reg.func = 4;
	o.reg.rs = prsReg();
	curseg->put(o.opcode);
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}
