#ifndef ASM_H
#define ASM_H
///////////////////////////////////////////////////////////////////////////////
// asm.h
// -----
// 
// Header file for the MiniMIPS assembler. The MiniMIPS assembler is a two-pass
// assembler for the MiniMIPS CPU. The assembler outputs a simple memory-image
// for loading directly into the VHDL simulator. It could easily be modified
// for more than one segment since the mechanism is already incorporated into
// the source.
// It uses the scanner declared in "scanner.h"
// The assembler starts by parsing the file once, placing the result in the
// unrealised segment "code". The segment is then realized and the file is 
// parsed again - this time using the mapping from the first parsing. 
// In the first pass, constants MUST be resolved to a value, eg.:
// 	b = a+1
// 	a = 10
// is invalid, since b cannot be resolved in first pass. By swapping the two
// lines, it becomes valid since both a and be can be resolved by first pass.
//
// Some functions has been left over from the original R1-assembler - namely
// methods for serveral segments.
//
// History:
//   2000-10-16	1.0	Modified from R1-assembler by Rune M. Barnkob (c973431)
//   2000-10-17	1.1	Found errors in parsing of SLL, SRL, SRA
//   2000-10-17	1.2	Missed 8k boundaries in J and JAL - corrected
//   2000-10-20	1.3	Sometimes first pass would cause EB_OUTOFRANGE
//   2000-10-30	1.4	Added some facilities for "code" and "data" segment
//   2000-10-30	1.5	Support for .org directive (some work in Segment::?)
//   2000-11-01	1.6	Unary minus added (needed that for -1)
//   2000-11-03	1.7	Corrected output for V-Systems
//   2000-11-09	1.8	Should have been killed for those horrific
//             	   	bugs in JR and ADDI - fixed now...stupid me...

#include "minimips.h"
#include <string>
#include <map>
#include "scanner.h"

// This declares the segment-structures
class Segment;
typedef Segment* pSeg;
struct SegHeader {
	char name[32];
	WORD length;
};

// An identifier consists of a name and a location in memory plus a type.
class Ident;

// The assembler itself.
class Assembler {

	public:
		Assembler();
		~Assembler();


		// assemble produces a map-file and an object file from the argument
		// supplied first by doing a two-pass.
		int assemble( string, string, string );

	protected:
		// Scanner used to read file as tokens
		Scanner sc;

		// Current line-number
		int lineno;

		// Segments. Default initialised with "code" and "data"
		// curseg set to "code"
		map<string, pSeg> segments;
		pSeg curseg;

		// Identifiers
		map<string, Ident> idents;

		// Pass-count for assembler. Don't generate code in first pass.
		enum { psFirst, psFinal } passCount;

		// current token in scanner
		ScTokens curtok;

		// End-of-file
		bool eof;

		// move to the next token in the input
		void nextToken();

		// Check for identifier in both segment- and ident-lists!
		bool idExists(string id);

		// look-up identifier in database
		Ident lookup(string id);


		// *** EXPRESSIONS ***
		// If resolve=true, a value must be determined during parsing! 
		// If resolve=false, 0 is returned if unresolved
		// parses expressions of <E0> ::= <E1> { ("+" | "-") <E1> }
		INT prsExpr0(bool resolve=false);
		
		// parses expressions of <E1> ::= <E2> { "*" <E2> }
		INT prsExpr1(bool resolve=false);
		
		// parses expressions of <E2> ::= <value> | "(" <E0> ")"
		INT prsExpr2(bool resolve=false);

		// <value> ::= <const> | <ident(type==const)> | "@" <ident(type!=const)>
		INT prsValue(bool resolve=false);


		// *** LINES ***
		
		// prsLine will parse a single line in the source file.
		// parsing will start looking for a directive and hand control to 
		// prsDirective() if found; look for a label (identifier) and pass 
		// control to prsLabel() if found, and if none of those matches, let
		// prsInst() handle the line (assuming an instruction exists on the line).
		void prsLine( string );

		// "." <directive>
		void prsDirective();

		// parses a label
		void prsLabel();
		// and a possible storage specifier following the label
		void prsStorage(string);



		// *** INSTRUCTIONS ***
		
		// parses a single register and returns the number of the register
		int prsReg();

		// starts parsing an instruction - branches into prs...() according to instruction
		void prsInst();

		// parses SLL, SRL, SRA
		void prs2RegType();
		
		// parses ADD, SUB, AND, OR, XOR, SLT, SLTU
		void prs3RegType();

		// parses ADDI
		void prsImmType();

		// parses LW, SW
		void prsMemType();

		// parses BEQ, BNE
		void prsBranchType();

		// parses J, JAL
		void prsJumpType();

		// parses JR
		void prsJumpRegType();
};

#endif