///////////////////////////////////////////////////////////////////////////////
// disasm.cpp
// ----------
//
// See disasm.h
//
// History:
//   2000-10-16	1.0	Rune M. Barnkob (c973431)
//   2000-10-17	1.1	Corrected some of the output
//   2000-10-20	1.2	Somehow rs and rt got switched in lw and sw
//   2000-10-30	1.3	Just noticed addi outputs in hex - not dec
//   2000-11-03	1.4	Added constructor with selectable flags

#include "disasm.h"
#include <strstream>
#include <iomanip>


// Constructs a new DisAsm object. The objects field intopcode (the opcode used
// if <<-operator is used) is initialized either to the supplied value or to 0
DisAsm::DisAsm() : intopcode(0), intat(0), intflags(0) {}

DisAsm::DisAsm(WORD i) : intopcode(i), intat(0), intflags(0) {}

DisAsm::DisAsm(WORD i, WORD a) : intopcode(i), intat(a), intflags(opcode_bin | has_addr | print_addr) {}

DisAsm::DisAsm(WORD i, WORD a, int flags) : intopcode(i), intat(a), intflags(flags) {}




// Disassemble an instruction	
string DisAsm::disasm(WORD i, WORD a=0, int flags=0) {
	oc opcode;
	ostrstream o;
	opcode.opcode = i;
	
	// Write address (if any)
	if (flags&print_addr) {
		o << hex << setw(4) << setfill('0') << a << "h:\t";
	}
		
	// Write hex-version of opcode (if requested)
	if (flags&opcode_hex) {
		o << "0x" << hex << setw(4) << setfill('0') << i << "\t";
	}

	// Write binary version of opcode (if requested)
	if (flags&opcode_bin) {
		// decode to binary
		for (int lp1=0; lp1<3; lp1++) { o << (i&0x8000?"1":"0"); i<<=1; }
		o << ".";
		for (int lp1=0; lp1<3; lp1++) { o << (i&0x8000?"1":"0"); i<<=1; }
		o << ".";
		for (int lp1=0; lp1<3; lp1++) { o << (i&0x8000?"1":"0"); i<<=1; }
		o << ".";
		for (int lp1=0; lp1<3; lp1++) { o << (i&0x8000?"1":"0"); i<<=1; }
		o << ".";
		for (int lp1=0; lp1<4; lp1++) { o << (i&0x8000?"1":"0"); i<<=1; }
		o << "\t";
	}
	
	// decode the instruction...
	switch (opcode.reg.op) {
		// alu-operations ; reg-format
		case 0: { 
			switch (opcode.reg.func) {
				case 0: 
					o << "sll\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rt << ends;
					break;
				
				case 1:
					o << "srl\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rt << ends;
					break;
					
				case 2:
					o << "sra\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rt << ends;
					break;

				case 4:
					o << "jr\t" << "$" << opcode.reg.rs << ends;
					break;

				case 8:
					o << "add\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rs << ", $" << opcode.reg.rt << ends;
					break;
					
				case 9:
					o << "sub\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rs << ", $" << opcode.reg.rt << ends;
					break;
					
				case 10:
					o << "and\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rs << ", $" << opcode.reg.rt << ends;
					break;
					
				case 11:
					o << "or\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rs << ", $" << opcode.reg.rt << ends;
					break;
					
				case 12:
					o << "xor\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rs << ", $" << opcode.reg.rt << ends;
					break;
					
				case 14:
					o << "slt\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rs << ", $" << opcode.reg.rt << ends;
					break;
					
				case 15:
					o << "sltu\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rs << ", $" << opcode.reg.rt << ends;
					break;
					
				default:
					o << "???\t" << "$" << opcode.reg.rd << ", $" << opcode.reg.rs << ", $" << opcode.reg.rt << ends;
			}
		}
		
		// addi ; imm-format
		case 1: 
			o << "addi\t$" << opcode.imm.rt << ", $" << opcode.imm.rs << ", " << dec << opcode.imm.imm << ends;
			break;
		
		// j ; jmp-format
		case 2: 
			o << "j\t0x" << setw(4) << setfill('0') << hex << ((a+1)&0xE000|opcode.jmp.target) << ends;
			break;
		
		// jal ; jmp-format
		case 3: 
			o << "jal\t0x" << setw(4) << setfill('0') << hex << ((a+1)&0xE000|opcode.jmp.target) << ends;
			break;
		
		// beq ; bra-format
		case 4: if (flags&has_addr) { // if we have an address for the instruction, calculate destination
			WORD bt = ((a+1)+opcode.bra.offset);
			o << "beq\t$" << opcode.bra.rs << ", $" << opcode.bra.rt << ", 0x" << hex << setw(4) << setfill('0') << bt << ends;
			break;
		} else // else use dec. relative value
			o << "beq\t$" << opcode.bra.rs << ", $" << opcode.bra.rt << ", pc+" << dec << opcode.bra.offset << ends;
		
		// bne ; bra-format
		case 5: if (flags&has_addr) { // if we have an address for the instruction, calculate destination
			WORD bt = ((a+1)+opcode.bra.offset);
			o << "bne\t$" << opcode.bra.rs << ", $" << opcode.bra.rt << ", 0x" << hex << setw(4) << setfill('0') << bt << ends;
			break;
		} else // else use dec. relative value
			o << "bne\t$" << opcode.bra.rs << ", $" << opcode.bra.rt << ", pc+" << dec << opcode.bra.offset << ends;
		
		// lw ; mem-format
		case 6: 
			o << "lw\t$" << opcode.mem.rt << ", " << dec << opcode.mem.offset << "($" << opcode.mem.rs << ")" << ends;
			break;
		
		// sw ; mem-format
		case 7: 
			o << "sw\t$" << opcode.mem.rt << ", " << dec << opcode.mem.offset << "($" << opcode.mem.rs << ")" << ends;
			break;
	}
	
	return o.str();
}
	

// Output the internal opcode field disassembled.
ostream& operator<<(ostream& a, DisAsm d) {
	return (a << d.disasm(d.intopcode, d.intat, d.intflags));
}