-- -----------------------------------------------------------------------------
--
--  Title      :  Mini-MIPS cache entity
--             :
--  Developers :  Rune M. Barnkob -- c973431@student.dtu.dk
--             :
--  Purpose    :  This entity contains a cache system for the MiniMIPS 
--             :  CPU design.
--             :  The cache is a direct mapped cache with 1 word entries.
--             :  The cache guarantees either a miss-signal (equivalent to 
--             :  mem's hold signal) or data on down-edge.
--             :  Writes use a write-allocate, write-back protocol with dirty
--             :  bits, and sometimes causes a write-miss resulting in a late
--             :  write-back taking lots of time.
--             :
--  Notes      :  This contains one process with two basic branches - one on
--             :  up-edge (writes) and one on down-edge (reads).
--             :
--  Revision   :  1.0    2001-01-12    initial version
--             :                                                   
--              
-- -----------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use STD.textio.all;
use WORK.types.all;

entity cache is
  generic(
    Entries: natural   := 8;          -- number of bits to use for entries
    Treqtomiss: time   := 2 ns;       -- time from req. neg-read-edge to miss
    Tcachedata: time   := 5 ns);      -- time from req. neg-read-edge to data
  
  port(
    clk:        in  std_logic;     -- The clock.
    reset:      in  std_logic;     -- Reset signal.
    cpu_addr:   in  halfword_t;    -- The CPU address bus.
    cpu_data:   inout halfword_t;  -- The CPU data bus.
    cpu_rw:     in  std_logic;     -- Read/Write signal.
    cpu_req:    in  std_logic;     -- Request memory operation signal.
    cpu_miss:   out std_logic;     -- Miss-signal, asserted on cache-miss
    mem_addr:   out halfword_t;    -- The address passed to real memory
    mem_data:   inout halfword_t;  -- The data read/written to real memory
    mem_rw:     out std_logic;     -- The read/write signal to memory
    mem_req:    out std_logic;     -- The request operation to real memory
    mem_hold:   in  std_logic);    -- Hold-signal from memory
end cache;

architecture specification of cache is
  type cache_entry is record
    valid: std_logic;
    dirty: std_logic;
    tag:   std_logic_vector(15 downto Entries);
    data:  halfword_t;
  end record;
  type cache_mem is array(0 to (2**Entries)-1) of cache_entry;
begin
  process 
    variable cache: cache_mem;
    variable cache_ent: integer;
  begin
    wait on clk, reset;
    cpu_data <= halfword_z;
    
    if reset='1' then
      for i in 0 to (2**Entries)-1 loop
        cache(i).valid := '0';
      end loop;
      cpu_miss <= '0';
      cpu_data <= halfword_z;
      mem_req <= '0';
      mem_rw <= '1';
      mem_addr <= halfword_zero;
      mem_data <= halfword_z;
    end if;
    
    -- when down-edge, reply by either data or cpu_miss assert
    if cpu_req='1' and clk='0' and clk'event then
      cache_ent:=conv_integer(unsigned(cpu_addr(Entries-1 downto 0)));
      -- detect cache hit
      if cache(cache_ent).valid='1' and cache(cache_ent).tag=cpu_addr(15 downto Entries) then
        -- cache hit
        if cpu_rw='1' then
          cpu_miss <= '0' after Treqtomiss;
          -- read from cache (fast)
          cpu_data <= cache(cache_ent).data after Tcachedata;
        else
          -- write to cache, possible write-back
          cpu_miss <= '0' after Treqtomiss;
          wait on clk until clk='1';
          cache(cache_ent).valid := '1';
          cache(cache_ent).dirty := '1';
          cache(cache_ent).tag := cpu_addr(15 downto Entries);
          cache(cache_ent).data := cpu_data;
        end if;
      else
        -- cache miss
        if cpu_rw='1' then
          -- read miss, init mem access
          cpu_miss <= '1' after Treqtomiss;
          wait on clk until (clk='1');
          mem_addr <= cpu_addr;
          mem_data <= halfword_z;
          mem_req <= '1';
          mem_rw <= '1';
          wait on clk until (clk='1' and mem_hold='0'); -- wait for mem data ready
          -- read mem data to cache
          cache(cache_ent).valid := '1';
          cache(cache_ent).dirty := '0';
          cache(cache_ent).tag := cpu_addr(15 downto Entries);
          cache(cache_ent).data := mem_data;
          mem_req <= '0';
          -- wait for down-edge to put cpu data out
          wait on clk until clk='0';
          cpu_data <= cache(cache_ent).data after Tcachedata;
          cpu_miss <= '0' after Treqtomiss;
        else
          -- write miss
          if cache(cache_ent).valid='1' and cache(cache_ent).dirty='1' then
            -- entry is valid and dirty and must be written to memory
            cpu_miss <= '1' after Treqtomiss;
            wait on clk until (clk='1');
            mem_req <= '1';
            mem_rw <= '0';
            mem_addr <= cache(cache_ent).tag & cpu_addr(Entries-1 downto 0);
            wait on clk until clk='0';
            mem_data <= cache(cache_ent).data;
            cache(cache_ent).valid := '1';
            cache(cache_ent).dirty := '1';
            cache(cache_ent).tag := cpu_addr(15 downto Entries);
            cache(cache_ent).data := cpu_data;
            wait on clk until (clk='1' and mem_hold='0');
            mem_req <= '0';
            mem_data <= halfword_z;
            mem_addr <= halfword_z;
            wait on clk until clk='0';
            cpu_miss <= '0' after Treqtomiss;
          else
            -- entry not valid or not dirty and can be replaced
            cpu_miss <= '0' after Treqtomiss;
            wait on clk until clk='1';
            cache(cache_ent).valid := '1';
            cache(cache_ent).dirty := '1';
            cache(cache_ent).tag := cpu_addr(15 downto Entries);
            cache(cache_ent).data := cpu_data;
          end if;
        end if;
      end if;  
    end if;
  end process;
end specification;