-- -----------------------------------------------------------------------------
--
--  Title      :  Components for the MiniMIPS module
--             :
--  Developers :  Niels Haandbk -- c958307@student.dtu.dk
--             :  Rune M. Barnkob -- c973431@student.dtu.dk
--             :
--  Purpose    :  This package contains a number of components used in the
--             :  structural version of the MiniMIPS project. It is highly
--             :  based upon the file "comp.vhd" used for the MODULO project.
--             :  (most of it is rip-offs from MODULO project)
--             :
--  Note       :  All the components have a generic parameter that sets the 
--             :  bit-width of the component. This defaults to 16 bits.
--             :
--  Revision   :  1.0    2000-12-21  Initial version based on "comp.vhd" v. 1.1
--
-- -----------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;


-------------------------------------------------------------------------------- 
-- A generic tri-state buffer. Defaults to a width of 16 bits. Note the special
-- statement that assigns the input to the output. It is similar to a simple
-- IF-statement but can be used outside a process.
--------------------------------------------------------------------------------
entity tri is                                
  generic (N: natural := 16; Tpd: time := 2 ns); -- Width of inputs and delay
  port(
    data_in:  in  std_logic_vector(N downto 1);   -- Input.
    data_out: out std_logic_vector(N downto 1);   -- Output.
    en:       in  std_logic                       -- Enable signal. 
  );                     
end tri;

architecture behaviour of tri is
begin
  data_out <= data_in after Tpd when en = '1' or en = 'H' else 
              (others => 'Z') after Tpd when en = '0' or en = 'L' else
              (others => 'X') after Tpd;
end behaviour;


-------------------------------------------------------------------------------- 
-- A generic latch. Width defaults to 16 bits. 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity latch is                               
  generic (N: natural := 16; Tpd: time := 2 ns);  -- Width of inputs and delay
  port(
    en:       in  std_logic;                      -- Enable signal.
    data_in:  in  std_logic_vector(N downto 1);   -- Input data.
    data_out: out std_logic_vector(N downto 1)    -- Output data.
  );  
end latch;

architecture behaviour of latch is
begin
  process (en, data_in)
  begin
    if en='1' then
        data_out <= data_in after Tpd;
    end if;
  end process;
end behaviour;


-------------------------------------------------------------------------------- 
-- A generic flop-flip. Width defaults to 16 bits. 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity reg is                               
  generic (N: natural := 16; Tpd: time := 2 ns);  -- Width of inputs and delay
  port(
    clk:      in  std_logic;                      -- Clock signal
    en:       in  std_logic;                      -- Enable signal.
    data_in:  in  std_logic_vector(N downto 1);   -- Input data.
    data_out: out std_logic_vector(N downto 1)    -- Output data.
  );  
end reg;

architecture behaviour of reg is
begin
  process (en, clk, data_in)
  begin
    if en='1' and clk'event and clk='1' then
        data_out <= data_in after Tpd;
    end if;
  end process;
end behaviour;


-------------------------------------------------------------------------------- 
-- A very simple generic unsigned adder (defaults to 16 bit).
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;


architecture behaviour of add is
begin
  process (A, B)
  begin
    C <= unsigned(A) + unsigned(B) after Tpd;
  end process;
end;


-------------------------------------------------------------------------------- 
-- An advanced ALU that works on numbers in two's complement representation. The 
-- width defaults to 16 bits. The ALU has functions defined in "types.vhd" 
-- in the signal "func":
-- The propagation delay of add/sub/slt/sltu is a linear function of Tpd with 
-- respect to W - this to emulate a ripple-carry-adder.
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity alu is  
  generic (W: natural := 16; Tpd: time := 2 ns); -- Width of inputs and delay
  port(
    A, B: in std_logic_vector(W downto 1);   -- Input operands.
    func: FuncCode;                          -- Function
    C:    out std_logic_vector(W downto 1)   -- Result.
    );  
end alu;

architecture behaviour of alu is
  signal i_C: std_logic_vector(W downto 1); 
  constant zero: std_logic_vector(W downto 1) := (others => '0');
  constant one: std_logic_vector(W downto 1) := (others => '0') & '1';
begin
  alu_assign: C <= i_C;
  alu_function: process (A, B, func) is
  begin
    case func is
      when funct_sll =>
        i_C <= A(14 downto 0) & '0' after Tpd;
  
      when funct_srl =>
        i_C <= '0' & A(15 downto 1) after Tpd;
  
      when funct_sra =>
        i_C <= A(15) & A(15 downto 1) after Tpd;
  
      when funct_add =>
        i_C <= signed(A) + signed(B) after Tpd*W;
  
      when funct_sub =>
        i_C <= signed(A) - signed(B) after Tpd*W;
  
      when funct_and =>
        for index in A'range loop
          i_C(index) <= A(index) and B(index) after Tpd;
        end loop;
        
      when funct_or =>
        for index in A'range loop
          i_C(index) <= A(index) or B(index) after Tpd;
        end loop;
        
      when funct_xor =>
        for index in A'range loop
          i_C(index) <= A(index) xor B(index) after Tpd;
        end loop;
        
      when funct_slt =>
        if (signed(A)<signed(B)) then
          i_C <= zero after Tpd*W;
        else
          i_C <= one after Tpd*W;
        end if;
  
      when funct_sltu =>
        if (stdl_to_natural(regs(rs))<stdl_to_natural(regs(rt))) then
          i_C <= zero after Tpd*W;
        else
          i_C <= one after Tpd*W;
        end if;
        
      when others =>
        report "alu() - invalid function code received"
        severity warning;
        i_C <= (others => 'X');
        
    end case;
  end process;  
end behaviour;

-------------------------------------------------------------------------------- 
-- A simple clock generator. The period is specified in a generic and defaults
-- to 50 ns.
-------------------------------------------------------------------------------- 
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity clock is
  generic (period:    time := 50 ns);
  port(
    clk:          out std_logic := '0'
  );
end clock;

architecture behaviour of clock is
begin
  process       
  begin
    clk <= '1', '0' after period/2;
    wait for period;
  end process;
end behaviour;

-------------------------------------------------------------------------------
-- A generic multiplexor for 4 W-bit wide inputs. Defaults to 16 bit.
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity mux4 is
  generic (W: natural := 16; Tpd: time := 2 ns);
  port(
    ctrl: in std_logic_vector(1 downto 0);
    data0: in std_logic_vector(W downto 1);
    data1: in std_logic_vector(W downto 1);
    data2: in std_logic_vector(W downto 1);
    data3: in std_logic_vector(W downto 1);
    output: out std_logic_vector(W downto 1)
  );
end mux4;

architecture behaviour of mux4 is
begin
  output <= data0 after Tpd when ctrl = "00" else
            data1 after Tpd when ctrl = "01" else
            data2 after Tpd when ctrl = "10" else
            data3 after Tpd when ctrl = "11" else
            (others => 'X') after Tpd;
end behaviour;
  
-------------------------------------------------------------------------------
-- A generic multiplexor for 2 W-bit wide inputs. Defaults to 16 bit.
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity mux2 is
  generic (W: natural := 16);
  port(
    ctrl: in std_logic;
    data0: in std_logic_vector(W downto 1);
    data1: in std_logic_vector(W downto 1);
    output: out std_logic_vector(W downto 1)
  );
end mux2;

architecture behaviour of mux2 is
begin
  output <= data0 after Tpd when ctrl = '0' else
            data1 after Tpd when ctrl = '1' else
            (others => 'X') after Tpd;
end behaviour;


-------------------------------------------------------------------------------
-- A package with component declarations for all the previously
-- declared components.
-------------------------------------------------------------------------------
package components is
  component tri is
    generic (N: natural := 16; Tpd: time := 2 ns); -- Width of inputs and delay
    port(
      data_in:  in  std_logic_vector(N downto 1);   -- Input.
      data_out: out std_logic_vector(N downto 1);   -- Output.
      en:       in  std_logic                       -- Enable signal. 
    );                     
  end component;

  component latch is                               
    generic (N: natural := 16; Tpd: time := 2 ns);  -- Width of inputs and delay
    port(
      en:       in  std_logic;                      -- Enable signal.
      data_in:  in  std_logic_vector(N downto 1);   -- Input data.
      data_out: out std_logic_vector(N downto 1)    -- Output data.
    );  
  end component;

  component reg is                               
    generic (N: natural := 16; Tpd: time := 2 ns);  -- Width of inputs and delay
    port(
      clk:      in  std_logic;                      -- Clock signal
      en:       in  std_logic;                      -- Enable signal.
      data_in:  in  std_logic_vector(N downto 1);   -- Input data.
      data_out: out std_logic_vector(N downto 1)    -- Output data.
    );  
  end component;

  component add is
    generic (N: natural := 16; Tpd: time := 2 ns);  -- Width of inputs and delay
    port(
      A, B: in std_logic_vector(N downto 1);
      C:    out std_logic_vector(N downto 1)
    );
  end component;
  
  component alu is  
    generic (W: natural := 16; Tpd: time := 2 ns); -- Width of inputs and delay
    port(
      A, B: in std_logic_vector(W downto 1);   -- Input operands.
      func: FuncCode;                          -- Function
      C:    out std_logic_vector(W downto 1)   -- Result.
      );  
  end component;

  component clock is
    generic (period:    time := 50 ns);
    port(
      clk:          out std_logic := '0'
    );
  end component;
  
  component mux4 is
    generic (W: natural := 16; Tpd: time := 2 ns);
    port(
      ctrl: in std_logic_vector(1 downto 0);
      data0: in std_logic_vector(W downto 1);
      data1: in std_logic_vector(W downto 1);
      data2: in std_logic_vector(W downto 1);
      data3: in std_logic_vector(W downto 1);
      output: out std_logic_vector(W downto 1)
    );
  end component;

  component mux2 is
    generic (W: natural := 16);
    port(
      ctrl: in std_logic;
      data0: in std_logic_vector(W downto 1);
      data1: in std_logic_vector(W downto 1);
      output: out std_logic_vector(W downto 1)
    );
  end component;

end components;