-------------------------------------------------------------------------------
-- cpu2.vhd
-- --------
-- Stages and CPU for the 5-stage pipeline version of the MiniMIPS CPU.
--
-- Contains the 5 entities:
--   InstFetch - Fetches instructions from IM
--   InstDecode - Decodes the instructions
--   Execute - Executes ALU operations
--   DataMem - Accesses data memory
--   WriteBack - Writes to RF (in practice; nothing)
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973431 Rune M. Barnkob--
-- Last Update :  2000/12/11
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

-------------------------------------------------------------------------------
-- InstFetch interfaces with the IM. The entity has its own register for PC
-- and must thus be clocked same as rest.
-------------------------------------------------------------------------------
entity InstFetch is port(
  IF_CO_STALL:    out bit_t;        -- stall-signal to rest of CPU (stalls entire CPU)
  IF_CI_CLK:      in  bit_t;        -- clock signal for IF stage
  IF_CI_RESET:    in  bit_t;        -- resets PC-register
  IF_CI_ENABLE:   in  bit_t;        -- enables internal register (used when stalling)
  -- forwarded inputs
  ID_DI_IF_newPC: in  halfword_t;   -- if a new PC must be forced in from ID
  ID_CI_IF_muxPC: in  bit_t;        -- then set this bit.
  -- output to pipeline register
  IF_DO_I:        out halfword_t;   -- the instruction read from IM
  IF_DO_PC:       out halfword_t;   -- PC+1
  -- I-memory interface
  im_addr:        out halfword_t;   -- Address bus for instruction.
  im_d:           inout halfword_t; -- The instruction bus.
  im_req:         out bit_t;        -- Request signal for instruction.
  im_rw:          out bit_t;        -- Read/Write signal for instruction.
  im_hold:        in  bit_t         -- Hold signal for instruction.
  );
end entity InstFetch;

architecture behaviour of InstFetch is
  signal if_d_curpc: halfword_t := halfword_zero;
  signal if_d_nextpc : halfword_t;
  signal if_d_tmppc:  halfword_t;
begin  -- behaviour
  -- purpose: Handles the PC register. If reset then go to Z on bus and reset reg. 
  -- type   : clocked
  -- inputs : IF_CI_CLK, IF_CI_RESET, IF_CI_RESET
  -- outputs: if_d_curpc
  IF_PC: process (IF_CI_CLK, IF_CI_RESET)
  begin  -- process
    if IF_CI_CLK'event and IF_CI_CLK='1' and IF_CI_ENABLE='1' then
      if_d_curpc <= if_d_nextpc;
    end if;
    if IF_CI_RESET='1' then
      im_d <= halfword_z;
      if_d_curpc <= halfword_zero;
    end if;
  end process IF_PC;

  -- purpose: the combinational circuit preceeding the PC register (a mux).
  -- type   : combinational
  -- inputs : ID_DI_IF_newPC, ID_D_IF_muxPC, if_d_tmppc
  -- outputs: if_d_nextpc
  IF_muxPC: process (ID_DI_IF_newPC, ID_CI_IF_muxPC, if_d_tmppc)
  begin  -- process
    if ID_CI_IF_muxPC = '1' then
      if_d_nextpc <= ID_DI_IF_newPC;
   else
      if_d_nextpc <= if_d_tmppc;
   end if;
  end process IF_muxPC;

  -- purpose: +1 adder for PC
  -- type   : combinational
  -- inputs : if_d_curpc
  -- outputs: IF_DO_PC, if_d_tmppc
  IF_ADDPC: process (if_d_curpc)
  begin  -- process
    if_d_tmppc <= unsigned(if_d_curpc) + 1;
    IF_DO_PC <= unsigned(if_d_curpc) + 1;
  end process IF_ADDPC;

  -- purpose: interfaces to memory and stalls CPU if nescessary
  -- type   : combinational
  -- inputs : im_hold, if_d_curpc, im_d
  -- outputs: IF_CO_STALL
  IF_IM: process (im_hold, im_d, if_d_curpc)
  begin  -- process
    im_addr <= if_d_curpc;
    im_rw <= '1';
    im_req <= '1';
    IF_CO_STALL <= im_hold;
    IF_DO_I <= im_d;
  end process IF_IM;

end architecture behaviour;

-------------------------------------------------------------------------------
-- InstDecode decodes the instruction and prepares for the next stages in the
-- pipeline. It handles forwarding and branches.
-- The meaning of the signals can be read in the documentation.
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity InstDecode is port(
  ID_CO_STALL:    out bit_t;      -- this stage needs to stall CPU (never)
  ID_CI_CLK:      in  bit_t;      -- clock for registers in RegFile
  ID_CI_RESET:    in  bit_t;      -- resets RegFile
  ID_CI_ENABLE:   in  bit_t;      -- enables register file clock
  -- input from pipeline register
  ID_DI_PC:       in  halfword_t; -- PC+1 stored
  ID_DI_I:        in  halfword_t; -- Instruction read by IF
  -- forwarding from WriteBack stage
  WB_CI_ID_WB:    in  bit_t;
  WB_CI_ID_WBR:   in  std_logic_vector(2 downto 0);
  WB_DI_ID_RES:   in  halfword_t;
  -- forwarding from DataMem stage
  DM_CI_ID_WB:    in  bit_t;
  DM_CI_ID_WBR:   in  std_logic_vector(2 downto 0);
  DM_DI_ID_RES:   in  halfword_t;
  -- forwarding from Execute stage
  EX_CI_ID_WB:    in  bit_t;
  EX_CI_ID_WBR:   in  std_logic_vector(2 downto 0);
  EX_DI_ID_RES:   in  halfword_t;
  -- new PC's if branches are performed
  ID_DO_IF_newPC: out halfword_t;
  ID_CO_IF_muxPC: out bit_t;
  -- outputs to pipeline register
  ID_DO_PC:       out halfword_t;  -- PC+1 (needed for JAL
  ID_DO_RS:       out halfword_t;  -- contents of register RS (after forwarding)
  ID_DO_RT:       out halfword_t;  -- contents of register RT (after forwarding)
  ID_DO_IMM:      out std_logic_vector(6 downto 0);  -- immediate read from I[6:0]
  ID_CO_WB:       out bit_t;       -- this instruction will write-back to a register
  ID_CO_WBR:      out std_logic_vector(2 downto 0);  -- the register it will write to
  ID_CO_DM:       out bit_t;       -- this instruction accesses data memory
  ID_CO_AC:       out std_logic_vector(3 downto 0);  -- control for the ALU
  ID_CO_EX_muxA:  out bit_t;       -- control for muxA in EX
  ID_CO_EX_muxB:  out std_logic_vector(1 downto 0)  -- control for muxB in EX
  );
end entity InstDecode;

architecture behaviour of InstDecode is
  signal id_d_op : OpCode;
  signal id_d_func : FuncCode;
  signal id_a_rs, id_a_rt, id_a_rd : std_logic_vector(2 downto 0);
  signal id_d_im : std_logic_vector(6 downto 0);
  signal id_d_regA, id_d_regB : halfword_t; -- Registers read from register file
  signal id_d_fwA, id_d_fwB : halfword_t;  -- Data after forwarding
  signal id_c_muxA, id_c_muxB : std_logic_vector(1 downto 0); -- Control for forwarding muxes
  signal id_c_breq : bit_t;             -- two values equal in ID_COMPARE
  signal id_c_brctr : std_logic_vector(1 downto 0);  -- control for ID_MUXPC
  signal id_d_bradd:  halfword_t; -- result of branch target calculation
  signal id_d_immse : halfword_t; -- sign-extended immediat for branch adder
  signal id_c_muxwbr : std_logic_vector(1 downto 0); -- control for write-back-reg mux
  constant forward_rf : std_logic_vector(1 downto 0) := "00"; -- forwards register file
  constant forward_ex : std_logic_vector(1 downto 0) := "01"; -- forwards execute stage
  constant forward_dm : std_logic_vector(1 downto 0) := "10"; -- forwards data-mem
  constant forward_wb : std_logic_vector(1 downto 0) := "11"; -- forwards write-back
  constant mw_0 : std_logic_vector(1 downto 0) := "00";
  constant mw_7 : std_logic_vector(1 downto 0) := "01";
  constant mw_rt : std_logic_vector(1 downto 0) := "10";
  constant mw_rd : std_logic_vector(1 downto 0) := "11";
  constant bsel_zero: std_logic_vector(1 downto 0) := "00";
  constant bsel_ra : std_logic_vector(1 downto 0) := "01";
  constant bsel_bradd : std_logic_vector(1 downto 0) := "10";
  constant bsel_jmp : std_logic_vector(1 downto 0) := "11";
begin  -- behaviour
  ID_CO_STALL <= '0';
  ID_DO_PC <= ID_DI_PC;
  ID_DO_RS <= id_d_fwA;
  ID_DO_RT <= id_d_fwB;

  -- purpose: This process splits the instruction into single lines of opcode, 
  --          funccode, rd, rs, rt and imm plus passes imm on
  -- type   : combinational
  -- inputs : ID_DI_I
  -- outputs: id_d_op, id_d_func, id_a_rs, id_a_rt, id_a_rd, id_d_im
  ID_SPLIT: process (ID_DI_I)
  begin  -- process
    id_d_op <= ID_DI_I(15 downto 13);
    id_d_func <= ID_DI_I(3 downto 0);
    id_a_rs <= ID_DI_I(12 downto 10);
    id_a_rt <= ID_DI_I(9 downto 7);
    id_a_rd <= ID_DI_I(6 downto 4);
    id_d_im <= ID_DI_I(6 downto 0);
    ID_DO_IMM <= ID_DI_I(6 downto 0);
  end process;

  -- purpose: Register file
  -- type   : clocked and reset-able
  -- inputs : ID_DO_RS, ID_DO_RT, ID_CI_CLK, WB_CI_WB, WB_CI_WBR, WB_DI_RES, ID_CI_RESET
  -- outputs: id_d_regA, id_d_regB
  ID_RF: process (id_a_rs, id_a_rt, ID_CI_CLK, ID_CI_RESET)
    -- The register file itself
    type regfile_t is array(0 to 7) of halfword_t;
    variable regs : regfile_t := (others => halfword_zero);
  begin  -- process ID_RF
    id_d_regA <= regs( conv_integer(unsigned(id_a_rs)) );
    id_d_regB <= regs( conv_integer(unsigned(id_a_rt)) );
    if ID_CI_CLK'event and ID_CI_CLK='1' and ID_CI_ENABLE='1' and WB_CI_ID_WB='1' and WB_CI_ID_WBR/="000" then
      regs( conv_integer(unsigned(WB_CI_ID_WBR)) ) := WB_DI_ID_RES;
    end if;
    if ID_CI_RESET='1' then
      regs := (others => halfword_zero);
    end if;
  end process ID_RF;

  -- purpose: Forwarding unit
  -- type   : combinational
  -- inputs : *_C_ID_WB, *_C_ID_WBR
  -- outputs: id_c_muxA, id_c_muxB
  ID_FORWARDUNIT: process (WB_CI_ID_WB, WB_CI_ID_WBR, DM_CI_ID_WB, DM_CI_ID_WBR, EX_CI_ID_WBR, EX_CI_ID_WBR, id_a_rs, id_a_rt)
  begin
    if EX_CI_ID_WB='1' and EX_CI_ID_WBR=id_a_rs and id_a_rs/="000" then
      id_c_muxA <= forward_ex;
    elsif DM_CI_ID_WB='1' and DM_CI_ID_WBR=id_a_rs and id_a_rs/="000" then
      id_c_muxA <= forward_dm;
    elsif WB_CI_ID_WB='1' and WB_CI_ID_WBR=id_a_rs and id_a_rs/="000" then
      id_c_muxA <= forward_wb;
    else
      id_c_muxA <= forward_rf;
    end if;

    if EX_CI_ID_WB='1' and EX_CI_ID_WBR=id_a_rt and id_a_rt/="000" then
      id_c_muxB <= forward_ex;
    elsif DM_CI_ID_WB='1' and DM_CI_ID_WBR=id_a_rt and id_a_rt/="000" then
      id_c_muxB <= forward_dm;
    elsif WB_CI_ID_WB='1' and WB_CI_ID_WBR=id_a_rt and id_a_rt/="000" then
      id_c_muxB <= forward_wb;
    else
      id_c_muxB <= forward_rf;
    end if;
  end process ID_FORWARDUNIT;

  -- purpose: Does forwarding from the different stages'
  -- type   : combinational
  -- inputs : WB_DI_ID_RES, DM_DI_ID_RES, EX_DI_ID_RES, id_d_regA, id_c_muxA
  -- outputs: id_d_fwA
  ID_MUXA: process (WB_DI_ID_RES, DM_DI_ID_RES, EX_DI_ID_RES, id_d_regA, id_c_muxA)
  begin  -- process ID_MUXA
    case id_c_muxA is
        when forward_rf => id_d_fwA <= id_d_regA;
        when forward_ex => id_d_fwA <= EX_DI_ID_RES;
        when forward_dm => id_d_fwA <= DM_DI_ID_RES;
        when forward_wb => id_d_fwA <= WB_DI_ID_RES;
        when others => id_d_fwA <= id_d_regA;
    end case;

  end process ID_MUXA;

  -- purpose: Does forwarding of B-register
  -- type   : combinational
  -- inputs : id_d_regB, EX_DI_ID_RES, DM_DI_ID_RES, WB_DI_ID_RES, id_c_muxB
  -- outputs: id_d_fwB
  ID_MUXB: process (id_d_regB, EX_DI_ID_RES, DM_DI_ID_RES, WB_DI_ID_RES, id_c_muxB)
  begin  -- process ID_MUXB
    case id_c_muxB is
      when forward_rf => id_d_fwB <= id_d_regB;
      when forward_ex => id_d_fwB <= EX_DI_ID_RES;
      when forward_dm => id_d_fwB <= DM_DI_ID_RES;
      when forward_wb => id_d_fwB <= WB_DI_ID_RES;
      when others => id_d_fwB <= id_d_regA;
    end case;
  end process ID_MUXB;

  -- purpose: Compares the two values for branch-decision
  -- type   : combinational
  -- inputs : id_d_fwA, id_d_fwB
  -- outputs: id_c_breq
  ID_COMPARE: process (id_d_fwA, id_d_fwB)
  begin  -- process ID_COMPARE
    if id_d_fwA = id_d_fwB then
      id_c_breq <= '1';
    else
      id_c_breq <= '0';
    end if;
  end process ID_COMPARE;

  -- purpose: Looks for and performs branches/jumps by adjusting ID_CI_IF_muxPC
  -- type   : combinational
  -- inputs : id_c_breq, ID_DI_I
  -- outputs: ID_CI_IF_muxPC
  ID_BRANCHDETECT: process (id_c_breq, id_d_op, id_d_func)
  begin  -- process ID_BRANCHDETECT
    case id_d_op is
      when op_j | op_jal =>
        ID_CO_IF_muxPC <= '1';
        id_c_brctr <= bsel_jmp;

      when op_beq =>
        ID_CO_IF_muxPC <= id_c_breq;
        id_c_brctr <= bsel_bradd;

      when op_bne =>
        ID_CO_IF_muxPC <= not id_c_breq;
        id_c_brctr <= bsel_bradd;

      when op_Rformat =>
        if id_d_func=funct_jr then
          ID_CO_IF_muxPC <= '1';
          id_c_brctr <= bsel_ra;
        else 
          ID_CO_IF_muxpc <= '0';
        end if;

      when others =>
        ID_CO_IF_muxPC <= '0';
    end case;
  end process ID_BRANCHDETECT;

  -- purpose: Chooses the right address to pass back to the IF stage in case of a jump/branch
  -- type   : combinational
  -- inputs : id_c_brctr, id_d_bradd, id_d_jadd, id_d_fwA, ID_DI_I
  -- outputs: ID_DI_IF_newPC
  ID_MUXPC: process (id_c_brctr, id_d_bradd, id_d_fwA, ID_DI_I, ID_DI_PC)
  begin  -- process ID_MUXPC
    case id_c_brctr is
      when bsel_bradd => ID_DO_IF_newPC <= id_d_bradd;
      when bsel_jmp => ID_DO_IF_newPC <= ID_DI_PC(15 downto 13) & ID_DI_I(12 downto 0);
      when bsel_ra => ID_DO_IF_newPC <= id_d_fwA;
      when others => ID_DO_IF_newPC <= halfword_zero;
    end case;
  end process ID_MUXPC;

  -- purpose: Sign-extends immediates from the instruction
  -- type   : combinational
  -- inputs : id_d_im
  -- outputs: id_d_immse
  ID_SE: process (id_d_im)
  begin  -- process ID_SE
    id_d_immse(15 downto 7) <= (others => id_d_im(6));
    id_d_immse(6 downto 0) <= id_d_im;
  end process ID_SE;

  -- purpose: Does branch-destination address calculation
  -- type   : combinational
  -- inputs : ID_DI_PC, id_d_immse
  -- outputs: id_d_bradd
  ID_BRANCHADD: process (ID_DI_PC, id_d_immse)
  begin  -- process ID_BRANCHADD
    id_d_bradd <= unsigned(ID_DI_PC) + unsigned(id_d_immse);
  end process ID_BRANCHADD;

  -- purpose: multiplexor for destination register
  -- type   : combinational
  -- inputs : id_c_muxwbr, ID_DO_RT, id_d_rd
  -- outputs: ID_CO_WBR
  ID_MUXWBR: process (id_c_muxwbr, id_a_rt, id_a_rd)
  begin  -- process ID_MUXWBR
    case id_c_muxwbr is
      when mw_7 => ID_CO_WBR <= "111";
      when mw_rt => ID_CO_WBR <= id_a_rt;
      when mw_rd => ID_CO_WBR <= id_a_rd;
      when others => ID_CO_WBR <= "000";
    end case;
  end process ID_MUXWBR;

  -- purpose: Calculate control signals for the rest of the pipeline
  -- type   : combinational
  -- inputs : id_d_op, id_d_func
  -- outputs: ID_C_EXmuxA, ID_C_EXmuxB, ID_CO_DM, ID_CO_WB, id_c_muxwbr
  ID_CONTROL: process (id_d_op, id_d_func)
  begin  -- process ID_CONTROL
    if id_d_op=op_sw or id_d_op=op_lw then
      ID_CO_DM <= '1';
    else
      ID_CO_DM <= '0';
    end if;
    case id_d_op is
      when op_Rformat =>
        ID_CO_WB <= '1';
        id_c_muxwbr <= mw_rd;
        if id_d_func= funct_jr then
          ID_CO_WB <= '0';
        end if;
        ID_CO_AC <= id_d_func;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "00";

      when op_addi =>
        ID_CO_WB <= '1';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_rt;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "01";

      when op_j =>
        ID_CO_WB <= '0';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_0;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "00";

      when op_jal =>
        ID_CO_WB <= '1';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_7;
        ID_CO_EX_muxA <= '1';
        ID_CO_EX_muxB <= "11";

      when op_beq | op_bne =>
        ID_CO_WB <= '0';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_0;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "00";

      when op_lw =>
        ID_CO_WB <= '1';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_rt;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "01";

      when op_sw =>
        ID_CO_WB <= '0';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_rt;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "01";

      when others =>
        report "EINVOP - invalid/unstable value for instruction" severity error;
    end case;
  end process ID_CONTROL;
end architecture behaviour;


-------------------------------------------------------------------------------
-- Executes the instruction on the ALU (if nescesary)
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity Execute is port(
  EX_CO_STALL: out bit_t;
  EX_CI_DM:    in  bit_t;
  EX_CI_AC:    in  std_logic_vector(3 downto 0);  -- alu control
  EX_DI_RS:    in  halfword_t;    -- first operand for the ALU
  EX_DI_RT:    in  halfword_t;    -- second operand for the ALU
  EX_DI_PC:    in  halfword_t;    -- PC+1 for this instruction
  EX_DI_IMM:   in  std_logic_vector(6 downto 0); -- immediate from I[6:0]
  EX_CI_WB:    in  bit_t;         -- this instruction does write-back to a register
  EX_CI_WBR:   in  std_logic_vector(2 downto 0);  -- write-back register
  EX_CI_muxA:  in  bit_t;         -- (pc) or (rs) on muxA
  EX_CI_muxB:  in  std_logic_vector(1 downto 0);  -- (imm_se), (rt) or (1) on muxB
  EX_DO_RES:   out halfword_t;    -- result of ALU operation
  EX_CO_DM:    out bit_t;         -- straight-through version of DM
  EX_CO_WB:    out bit_t;         -- --"-- WB
  EX_CO_WBR:   out std_logic_vector(2 downto 0);  -- --"-- WBR
  EX_DO_RT:    out halfword_t     -- --"-- RT
  );
end entity Execute;

architecture behavior of Execute is
  constant halfword_one: halfword_t := "0000000000000001";
  signal ex_d_aluA : halfword_t;        -- A-operand
  signal ex_d_aluB : halfword_t;        -- B-operand
  signal ex_d_imse : halfword_t;        -- sign-extended version of immediate
begin  -- behavioral
  EX_CO_STALL <= '0';

  -- purpose: sign-extender for immediate value from instruction
  -- type   : combinational
  -- inputs : EX_DI_IMM
  -- outputs: ex_d_imse
  EX_SE: process (EX_DI_IMM)
  begin  -- process EX_SE
    ex_d_imse(15 downto 7) <= (others => EX_DI_IMM(6));
    ex_d_imse(6 downto 0) <= EX_DI_IMM;
  end process EX_SE;

  -- purpose: handles muxA
  -- type   : combinational
  -- inputs : EX_DI_PC, EX_DI_RS, EX_CI_muxA
  -- outputs: ex_d_aluA
  EX_muxA: process (EX_DI_PC, EX_DI_RS, EX_CI_muxA)
  begin  -- process EX_muxA
    if EX_CI_muxA = '1' then
      ex_d_aluA <= EX_DI_PC;
    else
      ex_d_aluA <= EX_DI_RS;
    end if;
  end process EX_muxA;

  -- purpose: handles EX_muxB
  -- type   : combinational
  -- inputs : ex_d_imse, EX_DI_RT, EX_CI_muxB
  -- outputs: ex_d_aluB
  EX_muxB: process (ex_d_imse, EX_DI_RT, EX_CI_muxB)
  begin  -- process EX_muxB
    case EX_CI_muxB is
      when "00" => ex_d_aluB <= EX_DI_RT;
      when "01" => ex_d_aluB <= ex_d_imse;
      when "11" => ex_d_aluB <= halfword_one;
      when others => ex_d_aluB <= halfword_zero;
    end case;
  end process EX_muxB;

  -- purpose: alu functional unit
  -- type   : combinational
  -- inputs : ex_d_aluA, ex_d_aluB, ex_c_alu
  -- outputs: ex_do_res
  EX_ALU: process (ex_d_aluA, ex_d_aluB, EX_CI_AC)
  begin  -- process EX_ALU
    case EX_CI_AC is
      when funct_sll =>
        EX_DO_RES <= ex_d_aluB(14 downto 0) & '0';
      when funct_srl =>
        EX_DO_RES <= '0' & ex_d_aluB(15 downto 1);
      when funct_sra =>
        EX_DO_RES <= ex_d_aluB(15) & ex_d_aluB(15 downto 1);
      when funct_jr =>
        EX_DO_RES <= ex_d_aluA;
      when funct_add =>
        EX_DO_RES <= signed(ex_d_aluA) + signed(ex_d_aluB);
      when funct_sub =>
        EX_DO_RES <= signed(ex_d_aluA) - signed(ex_d_aluB);
      when funct_and =>
        for index in EX_DO_RES'range loop
          EX_DO_RES(index) <= ex_d_aluA(index) and ex_d_aluB(index);
        end loop;
      when funct_or =>
        for index in EX_DO_RES'range loop
          EX_DO_RES(index) <= ex_d_aluA(index) or ex_d_aluB(index);
        end loop;
      when funct_xor =>
        for index in EX_DO_RES'range loop
          EX_DO_RES(index) <= ex_d_aluA(index) xor ex_d_aluB(index);
        end loop;
      when funct_slt =>
        if signed(ex_d_aluA)<signed(ex_d_aluB) then
          EX_DO_RES <= halfword_one;
        else
          EX_DO_RES <= halfword_zero;
        end if;
      when funct_sltu =>
        if unsigned(ex_d_aluA)<unsigned(ex_d_aluB) then
          EX_DO_RES <= halfword_one;
        else
          EX_DO_RES <= halfword_zero;
        end if;
      when others =>
        report "EINVFUNC - invalid/unstable function code in ALU" severity error;
    end case;
  end process EX_ALU;

  -- purpose: pass-through process for wires
  -- type   : combinational
  -- inputs : EX_CI_DM, EX_CI_WB, EX_CI_WBR, EX_DI_RT
  -- outputs: EX_CO_DM, EX_CO_WB, EX_CI_WBR, EX_DO_RT
  process (EX_CI_DM, EX_CI_WB, EX_CI_WBR, EX_DI_RT)
  begin  -- process
    EX_CO_DM <= EX_CI_DM;
    EX_CO_WB <= EX_CI_WB;
    EX_CO_WBR <= EX_CI_WBR;
    EX_DO_RT <= EX_DI_RT;
  end process;
end architecture behavior;


-------------------------------------------------------------------------------
-- Accesses DM (if DM_CI_DM is set) or by-passes the stage (if DM_CI_DM is cleared)
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity DataMem is port(
  DM_CO_STALL: out bit_t;
  DM_CI_CLK:   in bit_t;              -- the clock!
  DM_CI_DM:    in bit_t;              -- this instruction accesses DM
  DM_CI_WB:    in bit_t;              -- write-back in register (not write-back in DM)
  DM_CI_WBR:   in std_logic_vector(2 downto 0);  -- write-back register
  DM_DI_RES:   in halfword_t;         -- address of memory cell
  DM_DI_RT:    in halfword_t;         -- contents of RT for this instruction (if write to DM)
  DM_CO_WB:    out bit_t;             -- write-back to register
  DM_CO_WBR:   out std_logic_vector(2 downto 0);  -- write-back register number
  DM_DO_RES:   out halfword_t;        -- either a by-pass of DM_DI_RES (if DM is cleared) or
                                      -- whatever was read from DM
  dm_hold:     in bit_t;              -- wait for memory
  dm_d:        inout halfword_t;      -- data read/written to memory
  dm_req:      out bit_t;             -- req. signal for this stage
  dm_rw:       out bit_t;             -- read (1) or write (0)
  dm_addr:     out halfword_t         -- address to access
  );
end entity DataMem;


architecture behaviour of DataMem is
  signal dm_d_data : halfword_t;        -- data read from memory

begin  -- behavioural

  -- purpose: DM_muxDM selects memory or by-pass
  -- type   : combinational
  -- inputs : DM_CI_DM, dm_d_data DM_DI_RES
  -- outputs: dm_do_res
  DM_muxDM: process (DM_CI_DM, dm_d_data, DM_DI_RES)
  begin  -- process
    if DM_CI_DM='1' then
      DM_DO_RES <= dm_d_data;
    else
      DM_DO_RES <= DM_DI_RES;
    end if;
  end process DM_muxDM;

  -- purpose: handles memory interfacing - the waiting for memory to complete
  -- type   : combinational
  -- inputs : DM_DI_RES, DM_CI_WB, DM_CI_DM, DM_DI_RT, DM_CI_CLK
  -- outputs: dm_req, dm_addr, dm_rw, dm_req, DM_CO_STALL
  DM_DM: process (DM_DI_RES, DM_CI_WB, DM_CI_DM, DM_DI_RT, DM_CI_CLK)
  begin  -- process
    if DM_CI_DM='1' then
      dm_req <= '1';
      dm_addr <= DM_DI_RES;
      if DM_CI_WB='1' then
        dm_rw <= '1';                     -- read from memory if write-back
        dm_d <= halfword_z;
      else
        dm_rw <= '0';
        if DM_CI_CLK'event and DM_CI_CLK='0' then
          dm_d <= DM_DI_RT;
        end if;
      end if;
      dm_addr <= DM_DI_RES;
    else
      dm_req <= '0';
      dm_rw <= '0';
      dm_d <= halfword_z;
      dm_addr <= halfword_z;
    end if;
  end process DM_DM;

  -- purpose: Reads data off memory bus
  -- type   : combinational
  -- inputs : dm_d
  -- outputs: dm_d_data
  DM_DMread: process (dm_d, dm_hold)
  begin  -- process DM_DMread
    dm_d_data <= dm_d;
    DM_CO_STALL <= dm_hold;
  end process DM_DMread;

  -- purpose: pass-through process
  -- type   : combinational
  -- inputs : DM_CI_WB, DM_CI_WBR
  -- outputs: DM_CO_WB, DM_CO_WBR
  process (DM_CI_WB, DM_CI_WBR)
  begin  -- process
    DM_CO_WB <= DM_CI_WB;
    DM_CO_WBR <= DM_CI_WBR;
  end process;
end architecture behaviour;



-------------------------------------------------------------------------------
-- Write-Back does no real work
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity WriteBack is
  port(
    WB_CI_WB:  in bit_t;
    WB_CI_WBR: in std_logic_vector(2 downto 0);
    WB_DI_RES: in halfword_t;
    WB_CO_WB:  out bit_t;
    WB_CO_WBR: out std_logic_vector(2 downto 0);
    WB_DO_RES: out halfword_t
  );
end entity WriteBack;

architecture behaviour of WriteBack is
begin
  WB_CO_WB <= WB_CI_WB;
  WB_CO_WBR <= WB_CI_WBR;
  WB_DO_RES <= WB_DI_RES;
end behaviour;


--------------------------------------------------------------------------------
-- A generic flop-flip. Width defaults to 16 bits.
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity reg is
  generic (N: natural := 16; Tpd: time := 2 ns);  -- Width of inputs and delay
  port(
    clk:      in  std_logic;                      -- Clock signal
    en:       in  std_logic;                      -- Enable signal.
    reset:    in  std_logic;                      -- Reset Register to 0
    data_in:  in  std_logic_vector(N downto 1);   -- Input data.
    data_out: out std_logic_vector(N downto 1)    -- Output data.
  );
end reg;

architecture behaviour of reg is
begin
  reg_int: process (en, clk, reset)
  begin
    if en='1' and clk'event and clk='1' then
      data_out <= data_in after Tpd;
    end if;
    if reset='1' then
      data_out <= (others => '0') after Tpd;
    end if;
  end process;
end behaviour;

--------------------------------------------------------------------------------
-- A single bit D-flop-flip.
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity dff is
  generic (Tpd: time := 2 ns);  -- delay...
  port(
    clk:      in  std_logic;           -- Clock signal
    en:       in  std_logic;           -- Enable signal.
    reset:    in  std_logic;           -- Reset Register to 0
    data_in:  in  std_logic;           -- Input data.
    data_out: out std_logic            -- Output data.
  );
end dff;

architecture behaviour of dff is
begin
  dff_int: process (en, clk, reset)
  begin
    if en='1' and clk'event and clk='1' then
      data_out <= data_in after Tpd;
    end if;
    if reset='1' then
      data_out <= '0' after Tpd;
    end if;
  end process;
end behaviour;


-------------------------------------------------------------------------------
-- The CPU entity itself
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity cpu2 is
    port (
      clk :       IN    bit_t;            -- The clock.
      reset:      IN    bit_t;            -- The reset signal.
      instaddr:   OUT   halfword_t;       -- Address bus for instruction.
      inst:       INOUT halfword_t;       -- The instruction bus.
      instreq:    OUT   bit_t;            -- Request signal for instruction.
      instrw:     OUT   bit_t;            -- Read/Write signal for instruction.
      insthold:   IN    bit_t;            -- Hold signal for instruction.
      dataaddr:   OUT   halfword_t;       -- Address bus for data.
      data:       INOUT halfword_t;       -- The data bus.
      datareq:    OUT   bit_t;            -- Request signal for data.
      datarw:     OUT   bit_t;            -- Read/Write signal for data.
      datahold:   IN    bit_t             -- Hold signal for data.
      );
end entity cpu2;

--------------------------------------------------------------------------------
-- The structural architecture of the CPU2-design. Direct Entity Instantiation 
-- is used (avaiable only in VHDL-93, should not be used when trying to 
-- implement...)
--------------------------------------------------------------------------------
architecture structure of cpu2 is
  subtype wbr_t is std_logic_vector(2 downto 0);
  subtype imm_t is std_logic_vector(6 downto 0);
-- Signal declarations:
  signal stall:         bit_t;
  signal if_en, ifid_en, idex_en, exdm_en, dmwb_en:   bit_t; -- enable for pipeline registers
  signal IF_CO_STALL:    bit_t; 
  signal IF_DO_I:        halfword_t; --> ID_DI_I
  signal IF_DO_PC:       halfword_t; --> ID_DI_PC
  signal ID_D_IF_newPC:  halfword_t; -- connection between ID_DI_IF_newPC and ID_DO_ID_pc
  signal ID_C_IF_muxPC:  bit_t;      
  signal ID_CO_STALL:    bit_t;
  signal ID_DI_PC:       halfword_t;
  signal ID_DI_I:        halfword_t;
  signal ID_DO_PC:      halfword_t; --> EX_DI_PC
  signal ID_DO_RS:       halfword_t; --> EX_DI_RS
  signal ID_DO_RT:       halfword_t; --> EX_DI_RT
  signal ID_DO_IMM:      imm_t;  --> EX_DI_IMM
  signal ID_CO_WB:       bit_t;  --> EX_CI_WB
  signal ID_CO_WBR:      std_logic_vector(2 downto 0);  --> EX_CI_WBR
  signal ID_CO_DM:       bit_t;  --> EX_CI_DM
  signal ID_CO_AC:       std_logic_vector(3 downto 0);  --> EX_CI_AC
  signal ID_CO_EX_muxA:  bit_t;  --> EX_CI_muxA
  signal ID_CO_EX_muxB:  std_logic_vector(1 downto 0);  --> EX_CI_muxB
  signal EX_CO_STALL:    bit_t;
  signal EX_CI_DM:       bit_t;
  signal EX_CI_AC:       std_logic_vector(3 downto 0);  
  signal EX_DI_RS:       halfword_t;  
  signal EX_DI_RT:       halfword_t;  
  signal EX_DI_PC:       halfword_t;  
  signal EX_DI_IMM:      imm_t;
  signal EX_CI_WB:       bit_t; 
  signal EX_CI_WBR:      std_logic_vector(2 downto 0);  
  signal EX_CI_muxA:     bit_t;  
  signal EX_CI_muxB:     std_logic_vector(1 downto 0);  
  signal EX_DO_RES:      halfword_t;  --> DM_DI_RES
  signal EX_CO_DM:      bit_t; --> DM_CI_DM, EX_C_ID_DM
  signal EX_CO_WB:      bit_t; --> DM_CI_WB, EX_CI_ID_WBR
  signal EX_CO_WBR:     std_logic_vector(2 downto 0);  --> DM_CI_WBR, EX_DI_ID_RES
  signal EX_DO_RT:      halfword_t; --> DM_DI_RT
  signal DM_CO_STALL:    bit_t;
  signal DM_CI_DM:       bit_t; 
  signal DM_CI_WB:       bit_t; 
  signal DM_CI_WBR:      std_logic_vector(2 downto 0);  
  signal DM_DI_RES:      halfword_t;  
  signal DM_DI_RT:       halfword_t;
  signal DM_CO_WB:      bit_t; --> WB_CI_WB, DM_CI_ID_WB
  signal DM_CO_WBR:     std_logic_vector(2 downto 0);  --> WB_CI_WBR, DM_CI_ID_WBR
  signal DM_DO_RES:     halfword_t;  --> WB_DO_RES, DM_DI_ID_RES
  signal WB_CI_WB:       bit_t;
  signal WB_CI_WBR:      std_logic_vector(2 downto 0);
  signal WB_DI_RES:      halfword_t;
  signal WB_CO_WB:      bit_t; --> WB_CI_ID_WB
  signal WB_CO_WBR:     std_logic_vector(2 downto 0); --> WB_CI_ID_WBR
  signal WB_DO_RES:     halfword_t; --> WB_DI_ID_RES
begin  -- structure
  -- pipeline stage IF
  stageIF: entity work.InstFetch port map (
    IF_CO_STALL => IF_CO_STALL,
    IF_CI_CLK => clk,
    IF_CI_RESET => reset,
    IF_CI_ENABLE => if_en,
    IF_DO_I => IF_DO_I,
    IF_DO_PC => IF_DO_PC,
    ID_DI_IF_newPC => ID_D_IF_newPC,
    ID_CI_IF_muxPC => ID_C_IF_muxPC,
    im_addr => instaddr,
    im_d => inst,
    im_req => instreq,
    im_rw => instrw,
    im_hold => insthold
  );
  -- pipeline registers IFID
  IFID_I: entity work.reg port map ( clk => clk, en => ifid_en, reset => reset, data_in => IF_DO_I, data_out => ID_DI_I );
  IFID_PC: entity work.reg port map ( clk => clk, en => ifid_en, reset => reset, data_in => IF_DO_PC, data_out => ID_DI_PC );
  -- pipeline stage ID
  stageID: entity work.InstDecode port map (
    ID_CO_STALL => ID_CO_STALL,
    ID_CI_CLK => clk,
    ID_CI_RESET => reset,
    ID_CI_ENABLE => ifid_en,
    ID_DI_PC => ID_DI_PC,
    ID_DI_I => ID_DI_I,
    -- forwarding from write-back stage
    WB_CI_ID_WB => WB_CO_WB,
    WB_CI_ID_WBR => WB_CO_WBR,
    WB_DI_ID_RES => WB_DO_RES,
    -- forwarding from DataMem stage
    DM_CI_ID_WB => DM_CO_WB,
    DM_CI_ID_WBR => DM_CO_WBR,
    DM_DI_ID_RES => DM_DO_RES,
    -- forwarding from Execute stage
    EX_CI_ID_WB => EX_CO_WB,
    EX_CI_ID_WBR => EX_CO_WBR,
    EX_DI_ID_RES => EX_DO_RES,
    -- new PC's if branches are performed
    ID_DO_IF_newPC => ID_D_IF_newPC,
    ID_CO_IF_muxPC => ID_C_IF_muxPC,
    -- outputs to pipeline register
    ID_DO_PC => ID_DO_PC,
    ID_DO_RS => ID_DO_RS,
    ID_DO_RT => ID_DO_RT,
    ID_DO_IMM => ID_DO_IMM,
    ID_CO_WB => ID_CO_WB,
    ID_CO_WBR => ID_CO_WBR,
    ID_CO_DM => ID_CO_DM,
    ID_CO_AC => ID_CO_AC,
    ID_CO_EX_muxA => ID_CO_EX_muxA,
    ID_CO_EX_muxB => ID_CO_EX_muxB
  );
  -- pipeline register IDEX
  IDEX_PC: entity work.reg port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_DO_PC, data_out => EX_DI_PC );
  IDEX_RS: entity work.reg port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_DO_RS,  data_out => EX_DI_RS );
  IDEX_RT: entity work.reg port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_DO_RT,  data_out => EX_DI_RT );
  IDEX_IMM: entity work.reg generic map (N=>7) port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_DO_IMM, data_out => EX_DI_IMM );
  IDEX_WB: entity work.dff port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_CO_WB, data_out => EX_CI_WB );
  IDEX_WBR: entity work.reg generic map (N=>3) port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_CO_WBR, data_out => EX_CI_WBR );
  IDEX_DM: entity work.dff port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_CO_DM, data_out => EX_CI_DM );
  IDEX_AC: entity work.reg generic map (N=>4) port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_CO_AC, data_out => EX_CI_AC );
  IDEX_muxA: entity work.dff port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_CO_EX_muxA, data_out => EX_CI_muxA );
  IDEX_muxB: entity work.reg generic map (N=>2) port map ( clk => clk, en => idex_en, reset => reset, data_in => ID_CO_EX_muxB, data_out => EX_CI_muxB );
  -- pipeline stage EX
  stageEX: entity work.Execute port map (
    EX_CO_STALL => EX_CO_STALL,
    EX_CI_DM => EX_CI_DM,
    EX_CI_AC => EX_CI_AC,
    EX_DI_RS => EX_DI_RS,
    EX_DI_RT => EX_DI_RT,
    EX_DI_PC => EX_DI_PC,
    EX_DI_IMM => EX_DI_IMM,
    EX_CI_WB => EX_CI_WB,
    EX_CI_WBR => EX_CI_WBR,
    EX_CI_muxA => EX_CI_muxA,
    EX_CI_muxB => EX_CI_muxB,
    EX_DO_RES => EX_DO_RES,
    EX_CO_DM => EX_CO_DM,
    EX_CO_WB => EX_CO_WB,
    EX_CO_WBR => EX_CO_WBR,
    EX_DO_RT => EX_DO_RT
  );  
  -- pipeline register EXDM
  EXDM_RES: entity work.reg port map ( clk => clk, en => exdm_en, reset => reset, data_in => EX_DO_RES, data_out => DM_DI_RES );
  EXDM_RT: entity work.reg port map ( clk => clk, en => exdm_en, reset => reset, data_in => EX_DO_RT, data_out => DM_DI_RT );
  EXDM_WBR: entity work.reg generic map (N=>3) port map ( clk => clk, en => exdm_en, reset => reset, data_in => EX_CO_WBR, data_out => DM_CI_WBR );
  EXDM_DM: entity work.dff port map ( clk => clk, en => exdm_en, reset => reset, data_in => EX_CO_DM, data_out => DM_CI_DM );
  EXDM_WB: entity work.dff port map ( clk => clk, en => exdm_en, reset => reset, data_in => EX_CO_WB, data_out => DM_CI_WB );
  -- pipeline stage DM
  stageDM: entity work.DataMem port map (
    DM_CO_STALL => DM_CO_STALL,
    DM_CI_CLK => clk,
    DM_CI_DM => DM_CI_DM,
    DM_CI_WB => DM_CI_WB,
    DM_CI_WBR => DM_CI_WBR,
    DM_DI_RES => DM_DI_RES,
    DM_DI_RT => DM_DI_RT,
    DM_CO_WB => DM_CO_WB,
    DM_CO_WBR => DM_CO_WBR,
    DM_DO_RES => DM_DO_RES,
    dm_hold => datahold,
    dm_d => data,
    dm_req => datareq,
    dm_rw => datarw,
    dm_addr => dataaddr
  );    
  -- pipeline register DMWB
  DMWB_RES: entity work.reg port map ( clk => clk, en => dmwb_en, reset => reset, data_in => DM_DO_RES, data_out => WB_DI_RES );
  DMWB_WBR: entity work.reg generic map (N=>3) port map ( clk => clk, en => dmwb_en, reset => reset, data_in => DM_CO_WBR, data_out => WB_CI_WBR );
  DMWB_WB: entity work.dff port map ( clk => clk, en => dmwb_en, reset => reset, data_in => DM_CO_WB, data_out => WB_CI_WB );
  -- pipeline stage WB
  stageWB: entity work.WriteBack port map (
    WB_CI_WB => WB_CI_WB,
    WB_CI_WBR => WB_CI_WBR,
    WB_DI_RES => WB_DI_RES,
    WB_CO_WB => WB_CO_WB,
    WB_CO_WBR => WB_CO_WBR,
    WB_DO_RES => WB_DO_RES
  );
  
  -- purpose: watches stall-signals from each stage and (if nescessary - later vers)
  --        : stalls the entire CPU if one of the signals are asserted.
  --        : NOTE: reset stalls CPU.
  -- type   : combinatorial
  -- input  : IF_CO_STALL, ID_CO_STALL, EX_CO_STALL, DM_CO_STALL, reset
  -- output : stall 
  STALL_ARBITRATOR: process (IF_CO_STALL, ID_CO_STALL, EX_CO_STALL, DM_CO_STALL, reset)
  begin
    if reset='1' then
      stall <= '1';
    else
      stall <= IF_CO_STALL or ID_CO_STALL or EX_CO_STALL or DM_CO_STALL;
    end if;
  end process;
  -- concurrent signal assignments to all enable signals.
  if_en <= not stall;
  ifid_en <= not stall;
  idex_en <= not stall;
  exdm_en <= not stall;
  dmwb_en <= not stall;
  
end architecture structure;


--------------------------------------------------------------------------------
-- The specification of the minimips CPU. This version uses at least two cycles
-- since it interfaces to memory correctly and needs one cycle to read 
-- instruction and another cycle to read/write memory (if nescessary).
-- WARNING: THIS VERSION IS MULTI-CYCLE AND HAS NO
--          DELAYED BRANCHES OR LOADS!!!
--------------------------------------------------------------------------------
architecture specification of cpu2 is 
  -- Signals used to output internal data during simulation. Intended for debugging only!
  signal view_reg: regType;    -- view of register content.
begin
  ----------------------------------------------------------------------------
  -- The main process. You should insert your code here. Make sure that it is	
  -- as short as possible.                            
  ----------------------------------------------------------------------------
  CPU_SPEC: process
    variable oc: OpCode;
    variable fc: FuncCode;
    variable rs: RegAddr;
    variable rt: RegAddr;
    variable rd: RegAddr;
    variable imm: halfword_t;  -- sign-extended immediate
    variable target: std_logic_vector(12 downto 0);
    variable PC: halfword_t;        -- Program Counter (PC)
    variable regs: RegType;      -- array with all registers

  begin 
    if reset='1' then
      -- reset CPU
      PC := halfword_zero;
      regs := (others => halfword_zero);
      -- reset bus interface
      inst <= halfword_z;
      instaddr <= halfword_zero;
      instreq <= '0';
      instrw <= '1';
      data <= halfword_z;
      dataaddr <= halfword_zero;
      datareq <= '0';
      datarw <= '1';
      wait until reset /= '1';
    end if;
    

    if clk'event and clk = '1' then
      -- read instruction and move program counter
      instrw <= '1';
      instreq <= '1';
      instaddr <= PC;
      wait on clk, reset until (clk='1' and insthold='0');
      PC := unsigned(PC) + 1;
      instreq <= '0';
      -- split instruction into components
      oc := inst(15 downto 13);
      rs := RegAddr( conv_integer(unsigned(inst(12 downto 10))) );
      rt := RegAddr( conv_integer(unsigned(inst(9 downto 7))) );
      rd := RegAddr( conv_integer(unsigned(inst(6 downto 4))) );
      fc := inst(3 downto 0);
      imm(6 downto 0) := inst(6 downto 0);
      imm(15 downto 7) := (others=>inst(6));
      target := inst(12 downto 0);
  
      -- branch on opcode
      case oc is
        when op_Rformat =>
          -- sub-split on function during ALU-type instructions
          case fc is
            when funct_sll =>
              regs(rd) := regs(rt)(14 downto 0) & '0';
              
            when funct_srl =>
              regs(rd) := '0' & regs(rt)(15 downto 1);
              
            when funct_sra =>
              regs(rd) := regs(rt)(15) & regs(rt)(15 downto 1);
              
            when funct_jr =>
              PC := regs(rs);
              
            when funct_add =>
              regs(rd) := signed(regs(rs)) + signed(regs(rt));
              
            when funct_sub =>
              regs(rd) := signed(regs(rs)) - signed(regs(rt));

            when funct_and =>
              for index in regs(rd)'range loop
                regs(rd)(index) := regs(rs)(index) and regs(rt)(index);
              end loop;
            
            when funct_or =>
              for index in regs(rd)'range loop
                regs(rd)(index) := regs(rs)(index) or regs(rt)(index);
              end loop;
            
            when funct_xor =>
              for index in regs(rd)'range loop
                regs(rd)(index) := regs(rs)(index) xor regs(rt)(index);
              end loop;
            
            when funct_slt =>
              if signed(regs(rs))<signed(regs(rt)) then
                regs(rd) := "0000000000000001";
              else
                regs(rd) := "0000000000000000";
              end if;
              
            when funct_sltu =>
              if unsigned(regs(rs))<unsigned(regs(rt)) then
                regs(rd) := "0000000000000001";
              else
                regs(rd) := "0000000000000000";
              end if;
            
            when others =>
              report "EINVFUNC - invalid function received";
              null;
          end case;
        
        when op_addi =>
          regs(rt) := signed(regs(rs)) + signed(imm);
          
        when op_j =>
          -- PC && 0xE000 + target
          PC := PC(15 downto 13) & target;
          
        when op_jal =>
          regs(7) := PC;
          PC := PC(15 downto 13) & target;
          
        when op_beq =>
          if (regs(rs)=regs(rt)) then 
            PC := unsigned(PC) + unsigned(imm);
          end if;
        
        when op_bne =>
          if (regs(rs)/=regs(rt)) then 
            PC := unsigned(PC) + unsigned(imm);
          end if;
        
        when op_lw =>
          wait on clk, reset until clk='1';
          dataaddr <= unsigned(regs(rs)) + unsigned(imm);
          datareq <= '1';
          datarw <= '1';
          wait on clk, reset until (clk='1' and datahold='0');
          regs(rt) := data;
          datareq <= '0';
        
        when op_sw =>
          wait on clk, reset until clk='1';
          dataaddr <= unsigned(regs(rs)) + unsigned(imm);
          datareq <= '1';
          datarw <= '0';
          data <= regs(rt);
          wait on clk, reset until (clk='1' and datahold='0');
          datareq <= '0';
          datarw <= '1';
          data <= halfword_z;
          
        -- If opcode hasn't matched any other case, it is invalid.
        when others =>
          report "EINVOPCODE - invalid opcode received";
          null;
          
      end case;
    end if;
    -- Model a hardwired $0.
    regs(0) := "0000000000000000";
    
    -- output registercontent on signal view_reg.
    view_reg <= regs;
    
    wait on clk, reset;
  end process;
end specification;

------------------------<< EOF >>----------------------------------------------
