-- -----------------------------------------------------------------------------
--
--  Title      :  Mini-MIPS memory entity.
--             :
--  Developers :  Niels Haandbk -- c958307@student.dtu.dk
--             :  Rune M. Barnkob -- c973431@student.dtu.dk
--             :
--  Purpose    :  This design contains a small memory for the Mini-MIPS CPU.
--             :  The memory loads data from a file specified as a generic.
--             :  This version is a bug-fix on Niels original code, which could
--             :  not do writes with 0 wait states (his design used down-edge
--             :  for writing, which is incorrect according to specs - up-edge
--             :  must be used for this).
--             :
--  Notes      :  This contains one process with two basic branches - one on
--             :  up-edge (writes) and one on down-edge (reads).
--             :
--  Revision   :  1.0    1999-09-12    initial version
--             :  2.0    2000-12-21    rewritten to fix bug
--             :  2.1    2001-01-12    data held to long on bus during reads
--              
-- -----------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- This is the memory model that should be used in task two. Write a small test 
-- bench for the memory and examine the behaviour if needed.
--------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use STD.textio.all;
use WORK.types.all;

entity mem3 is
    generic (filename:    string  := "";      -- Name of memory load file
             n:           natural := 0;       -- Number of hold cycles(wait states)
             Tclock:      time := 50 ns);     -- Length of clock-period.
    port (clk:        in  std_logic;              -- The clock.
          addr:       in  std_logic_vector(16 downto 1);         -- The address bus.
          data:       inout std_logic_vector(16 downto 1);       -- The data bus.
          rw:         in  std_logic;              -- Read/Write signal.
          req:        in  std_logic;              -- Request memory operation signal.
          hold:       out std_logic);             -- Hold signal.
end mem3;   

--------------------------------------------------------------------------------
-- This architecture for the memory is sensitive to both clock edges.
-- If a memory operation is requested then 'holdcnt' is set to the number of
-- hold-cycles. 'holdcnt' is decremented on every appropriate clock edge when it 
-- is greater than zero. When 'holdcnt' is equal to or less than one then the 
-- operation is done and data is either written to or read from the data bus. 
-- Notice how signal attributes are used to ensure that all signals have been 
-- held stable for the appropriate amount of time.
--------------------------------------------------------------------------------

architecture behaviour OF mem3 is
  procedure LoadMemoryFromFile(variable memory: out MemType;
                               constant filename: in string) is
    file memfile:     text open READ_MODE is filename;
    variable memline: line;
    variable tmp:     bit_vector(15 downto 0);    
  begin
    for i in memory'RANGE loop
      exit when ENDFILE(memfile);
      READLINE(memfile, memline);
      READ(memline, tmp);
      memory(i) := TO_STDLOGICVECTOR(tmp);
    end loop;
  end LoadMemoryFromFile;

  signal holdcnt:   NATURAL;
begin    
  -- process : Memory
  -- purpose : Does memory interface simulation
  -- type    : sequential, clocked on both edges
  -- input   : 
  memory: process
    variable memory:    MemType;  -- memory array  
  begin
    ------------------------------------------------------------------------
    -- Start by resetting all signals to acceptable values:
    ------------------------------------------------------------------------
    data <= (others => 'Z');
    holdcnt <= 0;
    
    if (filename /= "") then      
      LoadMemoryFromFile(memory, filename);   
    end if;

    ------------------------------------------------------------------------
    -- Endless loop for memory access.
    ------------------------------------------------------------------------
    hold <= '0';
    loop
      -- free bus
--      if req='0' then 
        data <= (others => 'Z');
--     	end if;
      
      -- something happened on the clock and memory accessed
      if clk'event and req='1' then
        -- down-edge, used if reading - bus only driven on clk='0'
        if rw='1' and clk='0' then
          -- if first time then set hold-count
          if holdcnt = 0 then
            hold <= '1';
            holdcnt <= n;
          end if;
          -- if a hold-count exists, decrease it
          if holdcnt > 0 then
            holdcnt <= holdcnt - 1;
          end if;
          -- if this is final hold and signals where stable through period, put
          -- data on bus
          if holdcnt <= 1 then
            if req'stable(n * Tclock) and rw'stable(n * Tclock) and addr'stable(n * Tclock) then
              hold <= '0'; -- release hold and put data if addr known
              if not is_X(addr) then
                data <= memory(CONV_INTEGER(UNSIGNED(addr)));
              else
                data <= (others => 'X');
              end if; -- is_X(addr)
            end if; -- stable
          end if; -- holdcnt
          -- end if read_write='1'
        -- set hold on down-edge if n>0
        elsif rw='0' and clk='0' then
          -- if first time then set hold-count
          if holdcnt = 0 then
            hold <= '1';
            holdcnt <= n;
          end if;
          -- if a hold-count exists, decrease it
          if holdcnt > 0 then
            holdcnt <= holdcnt - 1;
          end if;
          -- if this is final wait cycle, take away hold signal
          if holdcnt <= 1 then
            if req'stable(n * Tclock) and rw'stable(n * Tclock) and addr'stable(n * Tclock) then
               hold <= '0';
             end if;
          end if;          
        -- up-edge, used if writing, data caught on up-edge when holdcnt=1
        elsif rw='0' and clk='1' and holdcnt<=1 then 
          -- if this is final hold and signals where stable through period, put
          -- data on bus
          if req'stable(n * Tclock) and rw'stable(n * Tclock) and addr'stable(n * Tclock) then
            if not is_X(addr) then
              if data'stable(n * Tclock) then
                memory(CONV_INTEGER(UNSIGNED(addr))) := data;
              else
                memory(CONV_INTEGER(UNSIGNED(addr))) := (others => 'X');
              end if; -- data'stable
            else
              memory(0) := (others => 'X');
            end if; -- is_X(addr)
          end if;  -- req'stable
        end if;  -- rw='0' and clk='1' and holdcnt=1
        
      end if; -- clk'event and req='1'
      wait on clk; -- wait for clock event
    end loop; -- process wrapper loop
  end process;
end behaviour;
