--	 -----------------------------------------------------------------------------
--
--  Title      :  Testbench for task two of the Mini-MIPS project.
--             :
--  Developers :  Niels Haandbk -- c958307@student.dtu.dk
--             :  Mikael Lund Jepsen -- c973811@
--             :  Rune M. Barnkob -- c973431@
--             :
--  Purpose    :  This design contains a testbench for the CPU build in task 
--             :  two. It uses two external memory banks, one for instructions
--             :  and one for data.                           .                
--             :
--  Revision   :  1.0    99-09-09    Initial version
--             :  1.1    00-12-12    Added multiple configurations
--             :  1.2    00-12-29    Switched to mem3 architecture
--             :  1.3    01-01-12    Added cache structure and one config
-- -----------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use WORK.types.all;

entity mips is
end mips;

architecture structure of mips is
    component clock 
        generic (period:        time := 50 ns);
        port (clk:              out std_logic := '0');
    end component;

    component mem3 
        generic (filename:      string  := "";      
                 n:             natural := 0;
                 Tclock:        time := 50 ns);     
        port (clk:        in    bit_t;              
              addr:       in    halfword_t;         
              data:       inout halfword_t;       
              rw:         in    bit_t;              
              req:        in    bit_t;              
              hold:       out   bit_t);             
    end component;   

    component cpu2 
        port (clk :       in    bit_t;           
              reset:      in    bit_t;           
              instaddr:   out   halfword_t;      
              inst:       inout halfword_t;      
              instreq:    out   bit_t;           
              instrw:     out   bit_t;           
              insthold:   in    bit_t;           
              dataaddr:   out   halfword_t;      
              data:       inout halfword_t;      
              datareq:    out   bit_t;           
              datarw:     out   bit_t;           
              datahold:   in    bit_t);          
    end component;

    signal clk:      bit_t;
    signal reset:    bit_t;

    signal im_addr: halfword_t;
    signal im_bus: halfword_t;
    signal im_req:  bit_t;
    signal im_rw:   bit_t;
    signal im_hold: bit_t;

    signal dm_addr: halfword_t;
    signal dm_bus: halfword_t;
    signal dm_req:  bit_t;
    signal dm_rw:   bit_t;
    signal dm_hold: bit_t;

begin
    reset <= '1', 'L' AFTER 57 ns;

    SysClk: clock
        port map (clk      => clk);

    Processor: cpu2
        port map (clk      => clk,
                  reset    => reset,
                  instaddr => im_addr,
                  inst     => im_bus,
                  instreq  => im_req,
                  instrw   => im_rw,
                  insthold => im_hold,
                  dataaddr => dm_addr,
                  data     => dm_bus,
                  datareq  => dm_req,
                  datarw   => dm_rw,
                  datahold => dm_hold);

    InstMem: mem3
        port map (addr     => im_addr,
                  data     => im_bus,
                  req      => im_req,
                  rw       => im_rw,
                  hold     => im_hold,
                  clk      => clk);

    DataMem: mem3
        port map (addr     => dm_addr,
                  data     => dm_bus,
                  req      => dm_req,
                  rw       => dm_rw,
                  hold     => dm_hold,
                  clk      => clk);

end structure;

architecture cache_struc of mips is
    component clock 
        generic (period:        time := 50 ns);
        port (clk:              out std_logic := '0');
    end component;

    component mem3 
        generic (filename:      string  := "";      
                 n:             natural := 0;
                 Tclock:        time := 50 ns);     
        port (clk:        in    bit_t;              
              addr:       in    halfword_t;         
              data:       inout halfword_t;       
              rw:         in    bit_t;              
              req:        in    bit_t;              
              hold:       out   bit_t);             
    end component;   

    component cpu2 
        port (clk :       in    bit_t;           
              reset:      in    bit_t;           
              instaddr:   out   halfword_t;      
              inst:       inout halfword_t;      
              instreq:    out   bit_t;           
              instrw:     out   bit_t;           
              insthold:   in    bit_t;           
              dataaddr:   out   halfword_t;      
              data:       inout halfword_t;      
              datareq:    out   bit_t;           
              datarw:     out   bit_t;           
              datahold:   in    bit_t);          
    end component;

component cache is
  generic(
    Entries: natural   := 8;          -- number of bits to use for entries
    Treqtomiss: time   := 5 ns);      -- time from request to miss
  
  port(
    clk:        in  std_logic;     -- The clock.
    reset:      in  std_logic;     -- Reset signal.
    cpu_addr:   in  halfword_t;    -- The CPU address bus.
    cpu_data:   inout halfword_t;  -- The CPU data bus.
    cpu_rw:     in  std_logic;     -- Read/Write signal.
    cpu_req:    in  std_logic;     -- Request memory operation signal.
    cpu_miss:   out std_logic;    -- Miss-signal, asserted on cache-miss
    mem_addr:   out halfword_t;    -- The address passed to real memory
    mem_data:   inout halfword_t;  -- The data read/written to real memory
    mem_rw:     out std_logic;     -- The read/write signal to memory
    mem_req:    out std_logic;     -- The request operation to real memory
    mem_hold:   in  std_logic);    -- Hold-signal from memory
end component;

    signal clk:      bit_t;
    signal reset:    bit_t;

    signal ic_addr: halfword_t;
    signal ic_bus: halfword_t;
    signal ic_req:  bit_t;
    signal ic_rw:   bit_t;
    signal ic_miss: bit_t;
    signal im_addr: halfword_t;
    signal im_bus: halfword_t;
    signal im_req:  bit_t;
    signal im_rw:   bit_t;
    signal im_hold: bit_t;

    signal dc_addr: halfword_t;
    signal dc_bus: halfword_t;
    signal dc_req:  bit_t;
    signal dc_rw:   bit_t;
    signal dc_miss: bit_t;
    signal dm_addr: halfword_t;
    signal dm_bus: halfword_t;
    signal dm_req:  bit_t;
    signal dm_rw:   bit_t;
    signal dm_hold: bit_t;

begin
    reset <= '1', 'L' AFTER 57 ns;

    SysClk: clock
        port map (clk      => clk);

    Processor: cpu2
        port map (clk      => clk,
                  reset    => reset,
                  instaddr => ic_addr,
                  inst     => ic_bus,
                  instreq  => ic_req,
                  instrw   => ic_rw,
                  insthold => ic_miss,
                  dataaddr => dc_addr,
                  data     => dc_bus,
                  datareq  => dc_req,
                  datarw   => dc_rw,
                  datahold => dc_miss);
    
    InstCache: cache
        port map (clk => clk,
                  reset => reset,
                  cpu_addr => ic_addr,
                  cpu_data => ic_bus,
                  cpu_rw => ic_rw,
                  cpu_req => ic_req,
                  cpu_miss => ic_miss,
                  mem_addr => im_addr,
                  mem_data => im_bus,
                  mem_rw => im_rw,
                  mem_req => im_req,
                  mem_hold => im_hold);

    InstMem: mem3
        port map (addr     => im_addr,
                  data     => im_bus,
                  req      => im_req,
                  rw       => im_rw,
                  hold     => im_hold,
                  clk      => clk);

    DataCache: cache
        port map (clk => clk,
                  reset => reset,
                  cpu_addr => dc_addr,
                  cpu_data => dc_bus,
                  cpu_rw => dc_rw,
                  cpu_req => dc_req,
                  cpu_miss => dc_miss,
                  mem_addr => dm_addr,
                  mem_data => dm_bus,
                  mem_rw => dm_rw,
                  mem_req => dm_req,
                  mem_hold => dm_hold);

    DataMem: mem3
        port map (addr     => dm_addr,
                  data     => dm_bus,
                  req      => dm_req,
                  rw       => dm_rw,
                  hold     => dm_hold,
                  clk      => clk);

end cache_struc;

-------------------------------------------------------------------------------- 
-- A configuration for the testbench. This configuration will make the cpu load
-- the file "fib2.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration test_fib of mips is
    for structure
        for SysClk: clock
            use entity WORK.clock(behaviour)
                generic map(period => 50 ns);
        end for;

        for Processor: cpu2
                --use entity WORK.cpu2(specification);
                use entity WORK.cpu2(structure);
        end for;
    
        for InstMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\fib2.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    
        for DataMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\fib2.src.bin",
                            n        => 1,
                            Tclock   => 50 ns);
        end for;
    end for;
    
end test_fib;

configuration tc_fib of mips is
    for cache_struc
        for SysClk: clock
            use entity WORK.clock(behaviour)
                generic map(period => 50 ns);
        end for;

        for Processor: cpu2
                --use entity WORK.cpu2(specification);
                use entity WORK.cpu2(structure);
        end for;
    
        for InstCache: cache
            use entity WORK.cache(specification)
                generic map(Entries  => 4);
        end for;
                
        for InstMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\fib2.src.bin",
                            n        => 4,
                            Tclock   => 50 ns);
        end for;

        for DataCache: cache
            use entity WORK.cache(specification)
                generic map(Entries  => 4);
        end for;
    
        for DataMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\fib2.src.bin",
                            n        => 8,
                            Tclock   => 50 ns);
        end for;
    end for;
end tc_fib;

-------------------------------------------------------------------------------- 
-- A configuration for the testbench. This configuration will make the cpu load
-- the file "sqrt2.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration test_sqrt of mips is
    for structure
        for SysClk: clock
            use entity WORK.clock(behaviour)
                generic map(period => 50 ns);
        end for;

        for Processor: cpu2
                use entity WORK.cpu2(structure);
        end for;
    
        for InstMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\sqrt2.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    
        for DataMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\sqrt2.src.bin",
                            n        => 1,
                            Tclock   => 50 ns);
        end for;
    end for;
end test_sqrt;

-------------------------------------------------------------------------------- 
-- A configuration for the testbench. This configuration will make the cpu load
-- the file "pipetest.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration test_pipe of mips is
    for structure
        for SysClk: clock
            use entity WORK.clock(behaviour)
                generic map(period => 50 ns);
        end for;

        for Processor: cpu2
                use entity WORK.cpu2(structure);
        end for;
    
        for InstMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\pipetest.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    
        for DataMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\pipetest.src.bin",
                            n        => 1,
                            Tclock   => 50 ns);
        end for;
    end for;
end test_pipe;

configuration tc_pipe of mips is
    for cache_struc
        for SysClk: clock
            use entity WORK.clock(behaviour)
                generic map(period => 50 ns);
        end for;

        for Processor: cpu2
                use entity WORK.cpu2(structure);
        end for;
    
        for InstMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\pipetest.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    
        for DataMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\pipetest.src.bin",
                            n        => 1,
                            Tclock   => 50 ns);
        end for;
    end for;
end tc_pipe;

-------------------------------------------------------------------------------- 
-- A configuration for the testbench. This configuration will make the cpu load
-- the file "fwdtest.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration test_forward of mips is
    for structure
        for SysClk: clock
            use entity WORK.clock(behaviour)
                generic map(period => 50 ns);
        end for;

        for Processor: cpu2
                use entity WORK.cpu2(structure);
        end for;
    
        for InstMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\fwdtest.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    
        for DataMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\fwdtest.src.bin",
                            n        => 1,
                            Tclock   => 50 ns);
        end for;
    end for;
end test_forward;

-------------------------------------------------------------------------------- 
-- A configuration for the testbench. This configuration will make the cpu load
-- the file "br_jump.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration test_brjump of mips is
    for structure
        for SysClk: clock
            use entity WORK.clock(behaviour)
                generic map(period => 50 ns);
        end for;

        for Processor: cpu2
                use entity WORK.cpu2(structure);
        end for;
    
        for InstMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\br_jump.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    
        for DataMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\br_jump.src.bin",
                            n        => 1,
                            Tclock   => 50 ns);
        end for;
    end for;
end test_brjump;

-------------------------------------------------------------------------------- 
-- A configuration for the testbench. This configuration will make the cpu load
-- the file "sortheap.src.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration test_sort of mips is
    for structure
        for SysClk: clock
            use entity WORK.clock(behaviour)
                generic map(period => 50 ns);
        end for;

        for Processor: cpu2
                use entity WORK.cpu2(structure);
        end for;
    
        for InstMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\sortheap.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    
        for DataMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\sortheap.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    end for;
end test_sort;

-------------------------------------------------------------------------------- 
-- A configuration for the testbench. This configuration will make the cpu load
-- the file "erast.src.bin" when a simulation is started.
-------------------------------------------------------------------------------- 

configuration test_eras of mips is
    for structure
        for SysClk: clock
            use entity WORK.clock(behaviour)
                generic map(period => 50 ns);
        end for;

        for Processor: cpu2
                use entity WORK.cpu2(structure);
        end for;
    
        for InstMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\erast.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    
        for DataMem: mem3
            use entity WORK.mem3(behaviour)
                generic map(filename => "..\minimips_prog\erast.src.bin",
                            n        => 0,
                            Tclock   => 50 ns);
        end for;
    end for;
end test_eras;
