-------------------------------------------------------------------------------
-- ALU.vhd
-- -------
-- This is the ALU entity for the MiniMIPS pipeline Execute stage. 
-- The ALU is composed of an array of 16 single-bit ALU's, each supporting
-- only a very limited set of operations, but it allows for a very small
-- ALU. The ALU has been engineered to reduce the LCC (logic cell count). Best 
-- size so far is a LCC of 111.
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973816 Emil Heinze
--             :  c973431 Rune M. Barnkob
--
-- Last Update :  2001/01/28
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- Executes the instruction on the ALU (if nescesary)
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity ALU is port (
  A: in halfword_t;
  B: in halfword_t;
  C: out halfword_t;
  CTR: in std_logic_vector(3 downto 0)
  );
end entity ALU;

architecture behaviour of ALU is
  component singlebitALU is port (
    Binv: in  bit_t;
    Cin:  in  bit_t;
    l_op:   in  std_logic_vector(1 downto 0);
    a_op:   in  std_logic_vector(1 downto 0);
    A:    in  bit_t;
    B:    in  bit_t;
    R:    out bit_t;
    prevbit: in bit_t;
    nextbit: in bit_t;
    slt:  in  bit_t;
    Radd: out bit_t; -- bypass for the adder result
    Cout: out bit_t
  );
  end component singlebitALU;
  
  
  signal carrychain: halfword_t;
  signal sub: bit_t;
  signal msb_add: bit_t;
  signal zero: bit_t;
  signal sar: bit_t;
  signal slt: bit_t;
  signal l_op: std_logic_vector(1 downto 0);
  signal a_op: std_logic_vector(1 downto 0);
begin
  -- subtract-detect
  SUB_DETECT: sub <= '1' when CTR=funct_sub or CTR=funct_slt or CTR=funct_sltu else '0';
  zero <= '0';
  process(CTR)
  begin
    case CTR is
      when funct_sll =>
        l_op <= "00";
        a_op <= "10";
      when funct_srl | funct_sra =>
        l_op <= "00";
        a_op <= "01";
      when funct_and =>
        l_op <= "00";
        a_op <= "00";
      when funct_or =>
        l_op <= "01";
        a_op <= "00";
      when funct_xor =>
        l_op <= "10";
        a_op <= "00";
      when funct_slt | funct_sltu =>
        l_op <= "11";
        a_op <= "00";
      when funct_add | funct_sub =>
        l_op <= "00";
        a_op <= "11";
      when others =>
        l_op <= "00";
        a_op <= "11";
    end case;
  end process;
  
  -- it is nescessary to correct the MSB of the addition in order to correctly
  -- do SLT/SLTU
  slt <= not carrychain(15) when CTR=funct_sltu else (carrychain(15) xor A(15) xor (not B(15)));
  
  alu_0: singlebitALU port map (
    Binv => sub,
    Cin => sub,
    l_op => l_op, a_op => a_op,
    A => A(0),
    B => B(0),
    R => C(0),
    prevbit => zero,
    nextbit => B(1),
    slt => slt,
    Cout => carrychain(0)
  );

  alu_1: singlebitALU port map (
    Binv => sub, Cin => carrychain(0), l_op => l_op, a_op => a_op, A => A(1), B => B(1), 
    R => C(1), prevbit => B(0), nextbit => B(2), slt => zero, 
    Cout => carrychain(1) );

  alu_2: singlebitALU port map (
    Binv => sub, Cin => carrychain(1), l_op => l_op, a_op => a_op, A => A(2), B => B(2), 
    R => C(2), prevbit => B(1), nextbit => B(3), slt => zero, 
    Cout => carrychain(2) );

  alu_3: singlebitALU port map (
    Binv => sub, Cin => carrychain(2), l_op => l_op, a_op => a_op, A => A(3), B => B(3), 
    R => C(3), prevbit => B(2), nextbit => B(4), slt => zero, 
    Cout => carrychain(3) );

  alu_4: singlebitALU port map (
    Binv => sub, Cin => carrychain(3), l_op => l_op, a_op => a_op, A => A(4), B => B(4), 
    R => C(4), prevbit => B(3), nextbit => B(5), slt => zero, 
    Cout => carrychain(4) );

  alu_5: singlebitALU port map (
    Binv => sub, Cin => carrychain(4), l_op => l_op, a_op => a_op, A => A(5), B => B(5), 
    R => C(5), prevbit => B(4), nextbit => B(6), slt => zero, 
    Cout => carrychain(5) );

  alu_6: singlebitALU port map (
    Binv => sub, Cin => carrychain(5), l_op => l_op, a_op => a_op, A => A(6), B => B(6), 
    R => C(6), prevbit => B(5), nextbit => B(7), slt => zero, 
    Cout => carrychain(6) );

  alu_7: singlebitALU port map (
    Binv => sub, Cin => carrychain(6), l_op => l_op, a_op => a_op, A => A(7), B => B(7), 
    R => C(7), prevbit => B(6), nextbit => B(8), slt => zero, 
    Cout => carrychain(7) );
    
  alu_8: singlebitALU port map (
    Binv => sub, Cin => carrychain(7), l_op => l_op, a_op => a_op, A => A(8), B => B(8), 
    R => C(8), prevbit => B(7), nextbit => B(9), slt => zero, 
    Cout => carrychain(8) );
    
  alu_9: singlebitALU port map (
    Binv => sub, Cin => carrychain(8), l_op => l_op, a_op => a_op, A => A(9), B => B(9), 
    R => C(9), prevbit => B(8), nextbit => B(10), slt => zero, 
    Cout => carrychain(9) );
    
  alu_a: singlebitALU port map (
    Binv => sub, Cin => carrychain(9), l_op => l_op, a_op => a_op, A => A(10), B => B(10), 
    R => C(10), prevbit => B(9), nextbit => B(11), slt => zero, 
    Cout => carrychain(10) );
    
  alu_b: singlebitALU port map (
    Binv => sub, Cin => carrychain(10), l_op => l_op, a_op => a_op, A => A(11), B => B(11), 
    R => C(11), prevbit => B(10), nextbit => B(12), slt => zero, 
    Cout => carrychain(11) );
    
  alu_c: singlebitALU port map (
    Binv => sub, Cin => carrychain(11), l_op => l_op, a_op => a_op, A => A(12), B => B(12), 
    R => C(12), prevbit => B(11), nextbit => B(13), slt => zero, 
    Cout => carrychain(12) );
    
  alu_d: singlebitALU port map (
    Binv => sub, Cin => carrychain(12), l_op => l_op, a_op => a_op, A => A(13), B => B(13), 
    R => C(13), prevbit => B(12), nextbit => B(14), slt => zero, 
    Cout => carrychain(13) );

  alu_e: singlebitALU port map (
    Binv => sub, Cin => carrychain(13), l_op => l_op, a_op => a_op, A => A(14), B => B(14), 
    R => C(14), prevbit => B(13), nextbit => B(15), slt => zero, 
    Cout => carrychain(14) );

  sar <= B(15) when CTR=funct_sra else zero;
  alu_f: singlebitALU port map (
    Binv => sub, Cin => carrychain(14), l_op => l_op, a_op => a_op, A => A(15), B => B(15), 
    R => C(15), prevbit => B(14), nextbit => sar, slt => zero, 
    Cout => carrychain(15), Radd => msb_add );
  
end;
