-------------------------------------------------------------------------------
-- BusInt.vhd
-- ------------
-- Bus interface for the processor. Handles sequencing IM and DM request to 
-- only one memory instance. IM has higher priority than DM and is served 
-- first. The serving is performed through a Mealy-type state machine (to allow
-- single cycle access which saves a wait state for IM/DM).
-- This design also supports multiplier hold during memory serving.
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973431 Rune M. Barnkob
--             :  c973816 Emil Heinze
--
-- Last Update :  2001/01/28 - made dual-port interface and moved tri-state to
--                             external module
--                2001/01/30 - new next-state logic supporting mu_hold and
--                             start DM access in first cycle if I-cache hits.
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity BusInt is
  port(
    -- global signals
    clk       : in  bit_t;
    reset     : in  bit_t;
    
    -- interface towards the cpu core
    im_addr   : in  halfword_t;
    im_datain : in  halfword_t;
    im_dataout: out halfword_t;
    im_rw     : in  bit_t;
    im_req    : in  bit_t;
    im_hold   : out bit_t;
    dm_addr   : in  halfword_t;
    dm_datain : in  halfword_t;
    dm_dataout: out halfword_t;
    dm_rw     : in  bit_t;
    dm_req    : in  bit_t;
    dm_hold   : out bit_t;
    mu_hold   : in  bit_t;  -- The CPU signals that the EX stage holds due to multiplier latency
   
    -- interface towards memory
    mem_addr    : out halfword_t;
    mem_datain  : in  halfword_t;
    mem_dataout : out halfword_t;
    mem_rw      : out bit_t;
    mem_req     : out bit_t;
    mem_hold    : in  bit_t
  );
end entity BusInt;

architecture behaviour of BusInt is
  -- The four states for the bus interface:
  --   ready: ready to receive a request - IM highest pri., granted in 1st cycle
  --   serve_im: currently serving IM
  --   serve_dm: currently serving DM
  --   wait_mult: waiting (holding) due to multiplier
  type states is (ready, serve_im, serve_dm, wait_mult);
  signal curr_st:   states;     -- the current state of the interface
  signal nextstate: states;     -- the next state (after the clock)
  signal source:    bit_t;      -- signal that represents which source is using the memory now ('0': IM, '1': DM)
  signal im_stored: halfword_t; -- signal for IM register
  signal im_pass:   bit_t;      -- pass memory data-in and store when '1' else pass stored
  signal dm_stored: halfword_t; -- signal for DM register
  signal dm_pass:   bit_t;      -- pass memory data-in and store when '1' else pass stored
  signal gl_hold:   bit_t;      -- global hold signal (same on IM and DM)

begin
  -- purpose: set all signals to internal and external systems
  -- type   : combinatorial
  -- input  : curr_st, im_req, dm_req, mem_hold, mu_hold
  -- output : nextstate, source, im_pass, dm_pass, gl_hold, mem_req
  NEXTSTATE_LOGIC: process(curr_st, im_req, dm_req, mem_hold, mu_hold)
  begin
    case curr_st is
      when ready =>
        gl_hold <= mem_hold or mu_hold;
        mem_req <= im_req or dm_req;
        if im_req='1' then
          nextstate <= serve_im;
          source <= '0';
          dm_pass <= '0';
          im_pass <= '1';
        elsif dm_req='1' then
          nextstate <= serve_dm;
          source <= '1';
          dm_pass <= '1';
          im_pass <= '0';
        elsif mu_hold='1' then
          nextstate <= wait_mult;
          source <= '0';
          dm_pass <= '0';
          im_pass <= '0';
        else
          nextstate <= ready;
          source <= '0';
          dm_pass <= '0';
          im_pass <= '0';
        end if;
        
      when serve_im =>
        source <= '0';
        mem_req <= '1';
        dm_pass <= '0';
        im_pass <= '1';
        gl_hold <= mem_hold or dm_req or mu_hold;
        if mem_hold='1' then
          nextstate <= serve_im;
        elsif dm_req='1' then
          nextstate <= serve_dm;
        elsif mu_hold='1' then
          nextstate <= wait_mult;
        else
          nextstate <= ready;
        end if;
        
      when serve_dm => 
        source <= '1';
        mem_req <= '1';
        dm_pass <= '1';
        im_pass <= '0';
        gl_hold <= mem_hold or mu_hold;
        if mem_hold='1' then
          nextstate <= serve_dm;
        elsif mu_hold='1' then
          nextstate <= wait_mult;
        else
          nextstate <= ready;
        end if;
        
      when wait_mult =>
        source <= '0';
        mem_req <= '0';
        dm_pass <= '0';
        im_pass <= '0';
        gl_hold <= mu_hold;
        if mu_hold='1' then
          nextstate <= wait_mult;
        else
          nextstate <= ready;
        end if;
      
      when others =>
        source <= '0';
        im_pass <= '0';
        dm_pass <= '0';
        gl_hold <='0';
        nextstate <= ready;
      
    end case;
  end process NEXTSTATE_LOGIC;
  
  -- purpose: state register for the Mealy machine
  -- type   : sequential
  -- input  : reset, clk
  -- output : curr_st
  STATE_REG: process(clk, reset)
  begin
    if reset/='1' then
      curr_st <= ready;
    elsif clk'event and clk='1' then
      curr_st <= nextstate;
    end if;
  end process STATE_REG;

  -- purpose: register for catching data read from memory to IM (to make sure data ready when gl_hold='0')
  -- type   : sequential
  -- input  : clk, reset
  -- output : im_intdo
  CATCH_IM: process(clk, reset)
  begin
    if reset/='1' then
      im_stored <= halfword_zero;
    elsif clk'event and clk='1' then
      if im_pass='1' then
        im_stored <= mem_datain;
      end if;
    end if;
  end process CATCH_IM;

  -- bypass the register during the IM access cycle
  IM_MUX: im_dataout <= mem_datain when im_pass='1' else im_stored;  
  
  -- purpose: register for catching data read from memory to DM (to make sure data ready when gl_hold='0')
  -- type   : sequential
  -- input  : clk, reset
  -- output : im_intdo
  CATCH_DM: process(clk, reset)
  begin
    if reset/='1' then
      dm_stored <= halfword_zero;
    elsif clk'event and clk='1' then
      if dm_pass='1' then
        dm_stored <= mem_datain;
      end if;
    end if;
  end process CATCH_DM;

  -- bypass the register during the DM access cycle
  DM_MUX: dm_dataout <= mem_datain when dm_pass='1' else dm_stored;  
  
  -- pass DM's dataout directly to memory (IM not capable of writing)
  DATA_TO_MEM: mem_dataout <= dm_datain;
  
  -- same global hold signal on both IM and DM
  SET_IMHOLD: im_hold <= gl_hold;
  SET_DMHOLD: dm_hold <= gl_hold;
  
  -- mux'es to memory
  MUX_ADDR: mem_addr <= dm_addr when source='1' else im_addr;
  MUX_RW: mem_rw <= dm_rw when source='1' else im_rw;

end architecture behaviour;
