-------------------------------------------------------------------------------
-- DataMem.vhd
-- -----------
-- Data Memory Access stage of the pipeline
-- Accesses DM (if DM_CI_DM is set) or by-passes the stage (if DM_CI_DM is cleared)
-- The stage assumes data on dm_din to be valid when the CPU is unstalled.
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973431 Rune M. Barnkob
--             :  c973816 Emil Heinze
--
-- Last Update :  2001/01/25
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity DataMem is port(
  DM_CO_STALL: out bit_t;             -- this stage needs to stall entire CPU
  DM_CI_DM:    in  bit_t;             -- this instruction accesses DM
  DM_CI_WB:    in  bit_t;             -- write-back in register (not write-back in DM)
  DM_CI_WBR:   in  std_logic_vector(2 downto 0);  -- write-back register number
  DM_DI_RES:   in  halfword_t;        -- address of memory cell
  DM_DI_RT:    in  halfword_t;        -- contents of RT for this instruction (if write to DM)
  DM_CO_WB:    out bit_t;             -- write-back to register
  DM_CO_WBR:   out std_logic_vector(2 downto 0);  -- write-back register number
  DM_DO_RES:   out halfword_t;        -- either a by-pass of DM_DI_RES (if DM is cleared) or
                                      -- whatever was read from DM
  dm_hold:     in  bit_t;             -- wait for memory
  dm_din:      in  halfword_t;        -- data read from memory
  dm_dout:     out halfword_t;        -- data written to memory
  dm_req:      out bit_t;             -- req. signal for this stage
  dm_rw:       out bit_t;             -- read (1) or write (0)
  dm_addr:     out halfword_t         -- address to access
  );
end entity DataMem;


architecture behaviour of DataMem is
begin  -- behavioural
  -- purpose: DM_muxDM selects memory or by-pass
  -- type   : combinational
  -- inputs : DM_CI_DM, dm_din, DM_DI_RES
  -- outputs: DM_DO_RES
  DM_muxDM: DM_DO_RES <= dm_din when DM_CI_DM='1' else DM_DI_RES;

  -- purpose: handles memory interfacing - the waiting for memory to complete
  -- type   : combinational
  -- inputs : DM_DI_RES, DM_CI_WB, DM_CI_DM, DM_DI_RT, DM_CI_CLK, dm_hold
  -- outputs: dm_req, dm_addr, dm_rw, dm_dout, DM_CO_STALL
  DM_SETADDR  : dm_addr <= DM_DI_RES;  -- we force memory address continously - sometimes not to read
  DM_SETDATA  : dm_dout <= DM_DI_RT;  -- output data
  DM_SETSTALL : DM_CO_STALL <= dm_hold;  -- when memory holds, we need to stall pipeline
  DM_SETREQ   : dm_req <= DM_CI_DM;  -- only request if dm access
  DM_SETRW    : dm_rw <= DM_CI_WB and DM_CI_DM;  -- set rw as write-back

  -- purpose: pass-through process 
  WB_PASS  : DM_CO_WB <= DM_CI_WB;
  WBR_PASS : DM_CO_WBR <= DM_CI_WBR;
end architecture behaviour;

