-------------------------------------------------------------------------------
-- Execute.vhd
-- -----------
-- Execute stage of the pipeline
-- Executes the instruction on the ALU (if nescesary)
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973431 Rune M. Barnkob
--             :  c973816 Emil Heinze
--
-- Last Update :  2001/01/20
--             :  2001/01/25 - concurrent statements used
--             :  2001/01/30 - added multiplier
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity Execute is port(
  EX_CO_STALL:  out bit_t;
  EX_CI_CLK:    in  bit_t;
  EX_CI_RESET:  in  bit_t;
  EX_CI_DM:     in  bit_t;
  EX_CI_AC:     in  std_logic_vector(3 downto 0);  -- alu control
  EX_DI_RS:     in  halfword_t;    -- first operand for the ALU
  EX_DI_RT:     in  halfword_t;    -- second operand for the ALU
  EX_DI_PC:     in  halfword_t;    -- PC+1 for this instruction
  EX_DI_IMM:    in  std_logic_vector(6 downto 0); -- immediate from I[6:0]
  EX_CI_WB:     in  bit_t;         -- this instruction does write-back to a register
  EX_CI_WBR:    in  std_logic_vector(2 downto 0);  -- write-back register
  EX_CI_muxA:   in  bit_t;         -- (pc) or (rs) on muxA
  EX_CI_muxB:   in  std_logic_vector(1 downto 0);  -- (imm_se), (rt) or (1) on muxB
  EX_DO_RES:    out halfword_t;    -- result of ALU operation
  EX_CO_DM:     out bit_t;         -- straight-through version of DM
  EX_CO_WB:     out bit_t;         -- --"-- WB
  EX_CO_WBR:    out std_logic_vector(2 downto 0);  -- --"-- WBR
  EX_DO_RT:     out halfword_t     -- --"-- RT
  );
end entity Execute;

architecture behaviour of Execute is
  component ALU is port (
    A: in halfword_t;
    B: in halfword_t;
    C: out halfword_t;
    CTR: in std_logic_vector(3 downto 0)
  );
  end component ALU;
  component Mult is port (
    -- global
    clk: in bit_t;
    reset: in bit_t;
    -- input: when start='1', the multiplication starts
    A: in halfword_t;
    B: in halfword_t;
    Start: in bit_t;
    -- output: when Ready='1' result is ready
    Hi: out halfword_t;
    Lo: out halfword_t;
    Ready: out bit_t
    );
  end component Mult;

  signal ex_d_aluA : halfword_t;        -- A-operand
  signal ex_d_aluB : halfword_t;        -- B-operand
  signal ex_d_aluR : halfword_t;        -- Result from ALU
  signal ex_d_mulRLo : halfword_t;      -- Lo result from mult
  signal ex_d_mulRHi : halfword_t;      -- Hi result from mult
  signal ex_c_mulStart : bit_t;         -- Multiplier should start calc.
  signal ex_c_mulReady : bit_t;         -- Result is ready from mult.
  signal ex_c_muxres : std_logic_vector(1 downto 0); -- controls the output mux
  signal ex_d_imse : halfword_t;        -- sign-extended version of immediate
  constant halfword_one: halfword_t := "0000000000000001";
  
begin  -- behavioral
  -- This stage stalls if we try to fetch a mult result that isnt ready yet
  EX_CO_STALL <= '1' when (EX_CI_AC=funct_mflo or EX_CI_AC=funct_mfhi) and ex_c_mulReady='0' else '0';

  -- purpose: sign-extender for immediate value from instruction
  -- type   : combinational
  -- inputs : EX_DI_IMM
  -- outputs: ex_d_imse
  EX_SE: ex_d_imse <= EX_DI_IMM(6)&EX_DI_IMM(6)&EX_DI_IMM(6)&EX_DI_IMM(6)&EX_DI_IMM(6)&EX_DI_IMM(6)&EX_DI_IMM(6)&EX_DI_IMM(6)&EX_DI_IMM(6) & EX_DI_IMM;

  -- purpose: handles muxA
  -- type   : combinational
  -- inputs : EX_DI_PC, EX_DI_RS, EX_CI_muxA
  -- outputs: ex_d_aluA
  EX_muxA: ex_d_aluA <= EX_DI_PC when EX_CI_muxA='1' else EX_DI_RS;

  -- purpose: handles EX_muxB
  -- type   : combinational
  -- inputs : ex_d_imse, EX_DI_RT, EX_CI_muxB
  -- outputs: ex_d_aluB
  EX_muxB: ex_d_aluB <= ex_d_imse when EX_CI_muxB="01" else
                        EX_DI_RT when EX_CI_muxB="00" else
                        halfword_one;-- when EX_CI_muxB="11" else
                        --halfword_zero;
  
  -- purpose: alu functional unit
  -- inputs : ex_d_aluA, ex_d_aluB, ex_c_alu
  -- outputs: ex_do_res
  EX_ALU: ALU port map (
    A => ex_d_aluA,
    B => ex_d_aluB,
    C => ex_d_aluR,
    CTR => EX_CI_AC
  );
  
  -- purpose: Multiplier
  -- type   : component
  EX_MULT: mult port map (
    clk => EX_CI_CLK,
    reset => EX_CI_RESET,
    A => ex_d_aluA,
    B => ex_d_aluB,
    Start => ex_c_mulStart,
    Hi => ex_d_mulRHi,
    Lo => ex_d_mulRLo,
    Ready => ex_c_mulReady
  );

  -- 
  EX_MULSTART: ex_c_mulStart <= '1' when EX_CI_AC=funct_umul else '0';
  
  -- sets control signals for output mux
  EX_OUTMUXCTR: ex_c_muxres <= "10" when EX_CI_AC=funct_mfhi else
                               "11" when EX_CI_AC=funct_mflo else
                               "00";
  -- adjusts the output according to ex_c_muxres
  EX_OUTMUX: EX_DO_RES <= ex_d_mulRHi when ex_c_muxres="10" else
                          ex_d_mulRLo when ex_c_muxres="11" else
                          ex_d_aluR;

  -- purpose: pass-through process for wires
  -- type   : combinational
  -- inputs : EX_CI_DM, EX_CI_WB, EX_CI_WBR, EX_DI_RT
  -- outputs: EX_CO_DM, EX_CO_WB, EX_CI_WBR, EX_DO_RT
  EX_CO_DM <= EX_CI_DM;
  EX_CO_WB <= EX_CI_WB;
  EX_CO_WBR <= EX_CI_WBR;
  EX_DO_RT <= EX_DI_RT;
end architecture behaviour;
