-------------------------------------------------------------------------------
-- InstDecode.vhd
-- --------------
-- InstDecode decodes the instruction and prepares for the next stages in the
-- pipeline. It handles forwarding and branches.
-- The meaning of the signals can be read in the documentation.
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973431 Rune M. Barnkob
--             :  c973816 Emil Heinze
--
-- Last Update :  2001/01/17 - WB always cleared if write-back to register 0.
--                2001/01/20 - minor cosmetics
--                2001/01/25 - major cosmetics
--                2001/01/26 - minor cosmetics
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity InstDecode is port(
  ID_CO_STALL:    out bit_t;      -- this stage needs to stall CPU (never)
  ID_CI_CLK:      in  bit_t;      -- clock for registers in RegFile
  ID_CI_RESET:    in  bit_t;      -- resets RegFile
  ID_CI_ENABLE:   in  bit_t;      -- enables register file clock
  -- input from pipeline register
  ID_DI_PC:       in  halfword_t; -- PC+1 stored
  ID_DI_I:        in  halfword_t; -- Instruction read by IF
  -- forwarding from WriteBack stage
  WB_CI_ID_WB:    in  bit_t;
  WB_CI_ID_WBR:   in  std_logic_vector(2 downto 0);
  WB_DI_ID_RES:   in  halfword_t;
  -- forwarding from DataMem stage
  DM_CI_ID_WB:    in  bit_t;
  DM_CI_ID_WBR:   in  std_logic_vector(2 downto 0);
  DM_DI_ID_RES:   in  halfword_t;
  -- forwarding from Execute stage
  EX_CI_ID_WB:    in  bit_t;
  EX_CI_ID_WBR:   in  std_logic_vector(2 downto 0);
  EX_DI_ID_RES:   in  halfword_t;
  -- new PC's if branches are performed
  ID_DO_IF_newPC: out halfword_t;
  ID_CO_IF_muxPC: out bit_t;
  -- outputs to pipeline register
  ID_DO_PC:       out halfword_t;  -- PC+1 (needed for JAL
  ID_DO_RS:       out halfword_t;  -- contents of register RS (after forwarding)
  ID_DO_RT:       out halfword_t;  -- contents of register RT (after forwarding)
  ID_DO_IMM:      out std_logic_vector(6 downto 0);  -- immediate read from I[6:0]
  ID_CO_WB:       out bit_t;       -- this instruction will write-back to a register
  ID_CO_WBR:      out std_logic_vector(2 downto 0);  -- the register it will write to
  ID_CO_DM:       out bit_t;       -- this instruction accesses data memory
  ID_CO_AC:       out std_logic_vector(3 downto 0);  -- control for the ALU
  ID_CO_EX_muxA:  out bit_t;       -- control for muxA in EX
  ID_CO_EX_muxB:  out std_logic_vector(1 downto 0)  -- control for muxB in EX
  );
end entity InstDecode;


architecture behaviour of InstDecode is
  -- The register file itself
  type regfile_t is array(0 to 7) of halfword_t;
  signal regs : regfile_t;
  signal id_d_op : OpCode;
  signal id_d_func : FuncCode;
  signal id_a_rs, id_a_rt, id_a_rd : std_logic_vector(2 downto 0);
  signal id_d_im : std_logic_vector(6 downto 0);
  signal id_d_regA, id_d_regB : halfword_t; -- Registers read from register file
  signal id_d_fwA, id_d_fwB : halfword_t;  -- Data after forwarding
  signal id_c_muxA, id_c_muxB : std_logic_vector(1 downto 0); -- Control for forwarding muxes
  signal id_c_breq : bit_t;             -- two values equal in ID_COMPARE
  signal id_c_brctr : std_logic_vector(1 downto 0);  -- control for ID_MUXPC
  signal id_d_bradd:  halfword_t; -- result of branch target calculation
  signal id_d_immse : halfword_t; -- sign-extended immediat for branch adder
  signal id_c_muxwbr : std_logic_vector(1 downto 0); -- control for write-back-reg mux
  signal id_c_wb : std_logic; -- temporary write-back signal used to detect writes to reg. 0
  signal id_c_wbr : std_logic_vector(2 downto 0); -- temporary internal write back reg. no
  constant forward_rf : std_logic_vector(1 downto 0) := "00"; -- forwards register file
  constant forward_ex : std_logic_vector(1 downto 0) := "01"; -- forwards execute stage
  constant forward_dm : std_logic_vector(1 downto 0) := "10"; -- forwards data-mem
  constant forward_wb : std_logic_vector(1 downto 0) := "11"; -- forwards write-back
  constant mw_0 : std_logic_vector(1 downto 0) := "00";
  constant mw_7 : std_logic_vector(1 downto 0) := "01";
  constant mw_rt : std_logic_vector(1 downto 0) := "10";
  constant mw_rd : std_logic_vector(1 downto 0) := "11";
  constant bsel_zero: std_logic_vector(1 downto 0) := "00";
  constant bsel_ra : std_logic_vector(1 downto 0) := "01";
  constant bsel_bradd : std_logic_vector(1 downto 0) := "10";
  constant bsel_jmp : std_logic_vector(1 downto 0) := "11";
begin  -- behaviour
  ID_CO_STALL <= '0';
  ID_DO_PC <= ID_DI_PC;
  ID_DO_RS <= id_d_fwA;
  ID_DO_RT <= id_d_fwB;

  -- purpose: This process splits the instruction into single lines of opcode, 
  --          funccode, rd, rs, rt and imm plus passes imm on
  -- type   : combinational
  -- inputs : ID_DI_I
  -- outputs: id_d_op, id_d_func, id_a_rs, id_a_rt, id_a_rd, id_d_im
  ID_EXTRACTOP  : id_d_op <= ID_DI_I(15 downto 13);
  ID_EXTRACTFUNC: id_d_func <= ID_DI_I(3 downto 0);
  ID_EXTRACTRS  : id_a_rs <= ID_DI_I(12 downto 10);
  ID_EXTRACTRT  : id_a_rt <= ID_DI_I(9 downto 7);
  ID_EXTRACTRD  : id_a_rd <= ID_DI_I(6 downto 4);
  ID_EXTRACTIMM : id_d_im <= ID_DI_I(6 downto 0);
  ID_PASSIMM    : ID_DO_IMM <= ID_DI_I(6 downto 0);

  -- purpose: Register file
  -- type   : clocked and reset-able
  -- inputs : WB_CI_WB, WB_CI_WBR, WB_DI_RES, ID_CI_RESET, id_a_rs, id_a_rt, ID_CI_CLK
  -- outputs: id_d_regA, id_d_regB
  ID_FETCHA: id_d_regA <= regs( conv_integer(unsigned(id_a_rs)) );
  ID_FETCHB: id_d_regB <= regs( conv_integer(unsigned(id_a_rt)) );
  ID_RF: process (ID_CI_CLK, ID_CI_RESET)
  begin  -- process ID_RF
    if ID_CI_RESET/='1' then
      regs(0) <= halfword_zero;
      regs(1) <= halfword_zero;
      regs(2) <= halfword_zero;
      regs(3) <= halfword_zero;
      regs(4) <= halfword_zero;
      regs(5) <= halfword_zero;
      regs(6) <= halfword_zero;
      regs(7) <= halfword_zero;
    elsif ID_CI_CLK'event and ID_CI_CLK='1' then
      if ID_CI_ENABLE='1' and WB_CI_ID_WB='1' then -- and WB_CI_ID_WBR/="000" then
        regs( conv_integer(unsigned(WB_CI_ID_WBR)) ) <= WB_DI_ID_RES;
      end if;
    end if;
  end process ID_RF;

  -- purpose: Forwarding unit
  -- type   : combinational
  -- inputs : *_C_ID_WB, *_C_ID_WBR
  -- outputs: id_c_muxA, id_c_muxB
  ID_FWDA: id_c_muxA <= forward_ex when EX_CI_ID_WB='1' and EX_CI_ID_WBR=id_a_rs else
                        forward_dm when DM_CI_ID_WB='1' and DM_CI_ID_WBR=id_a_rs else
                        forward_wb when WB_CI_ID_WB='1' and WB_CI_ID_WBR=id_a_rs else
                        forward_rf;
  ID_FWDB: id_c_muxB <= forward_ex when EX_CI_ID_WB='1' and EX_CI_ID_WBR=id_a_rt else
                        forward_dm when DM_CI_ID_WB='1' and DM_CI_ID_WBR=id_a_rt else
                        forward_wb when WB_CI_ID_WB='1' and WB_CI_ID_WBR=id_a_rt else
                        forward_rf;

  -- purpose: Does forwarding from the different stages'
  -- type   : combinational
  -- inputs : ...
  -- outputs: id_d_fwA
  ID_MUXA: id_d_fwA <= EX_DI_ID_RES when id_c_muxA=forward_ex else
                       DM_DI_ID_RES when id_c_muxA=forward_dm else
                       WB_DI_ID_RES when id_c_muxA=forward_wb else
                       id_d_regA;                       

  -- purpose: Does forwarding of B-register
  -- type   : combinational
  -- inputs : ...
  -- outputs: id_d_fwB
  ID_MUXB: id_d_fwB <= EX_DI_ID_RES when id_c_muxB=forward_ex else
                       DM_DI_ID_RES when id_c_muxB=forward_dm else
                       WB_DI_ID_RES when id_c_muxB=forward_wb else
                       id_d_regB;

  -- purpose: Compares the two values for branch-decision
  -- type   : combinational
  -- inputs : ...
  -- outputs: id_c_breq
  ID_COMPARE: id_c_breq <= '1' when id_d_fwA=id_d_fwB else '0';
  
  -- purpose: Looks for and performs branches/jumps by adjusting ID_CI_IF_muxPC
  -- type   : combinational
  -- inputs : id_c_breq, id_d_op, id_d_func
  -- outputs: ID_CI_IF_muxPC
  ID_BRANCHDETECT: process (id_c_breq, id_d_op, id_d_func)
  begin  -- process ID_BRANCHDETECT
    case id_d_op is
      when op_j | op_jal =>
        id_c_brctr <= bsel_jmp;
        ID_CO_IF_muxPC <= '1';

      when op_beq =>
        id_c_brctr <= bsel_bradd;
        ID_CO_IF_muxPC <= id_c_breq;

      when op_bne =>
        id_c_brctr <= bsel_bradd;
        ID_CO_IF_muxPC <= not id_c_breq;

      when op_Rformat =>
        id_c_brctr <= bsel_ra; -- causes slightly unstable values at outputs of ID_DO_IF_newPC
        if id_d_func=funct_jr then
          ID_CO_IF_muxPC <= '1';
        else 
          ID_CO_IF_muxPC <= '0';
        end if;

      when others =>
        id_c_brctr <= bsel_zero;
        ID_CO_IF_muxPC <= '0';
        
    end case;
  end process ID_BRANCHDETECT;

  -- purpose: Chooses the right address to pass back to the IF stage in case of a jump/branch
  -- type   : combinational
  -- inputs : id_c_brctr, id_d_bradd, id_d_jadd, id_d_fwA, ID_DI_I
  -- outputs: ID_DI_IF_newPC
  ID_MUXPC: ID_DO_IF_newPC <= id_d_bradd when id_c_brctr=bsel_bradd else
                              id_d_fwA when id_c_brctr=bsel_ra else
                              ID_DI_PC(15 downto 13) & ID_DI_I(12 downto 0);-- when id_c_brctr=bsel_jmp else
                              --halfword_zero;

  -- purpose: Sign-extends immediates from the instruction
  -- type   : combinational
  -- inputs : id_d_im
  -- outputs: id_d_immse
  ID_SE: id_d_immse <= id_d_im(6)&id_d_im(6)&id_d_im(6)&id_d_im(6)&id_d_im(6)&id_d_im(6)&id_d_im(6)&id_d_im(6)&id_d_im(6)&id_d_im;

  -- purpose: Does branch-destination address calculation
  -- type   : combinational
  -- inputs : ID_DI_PC, id_d_immse
  -- outputs: id_d_bradd
  ID_BRANCHADD: id_d_bradd <= unsigned(ID_DI_PC) + unsigned(id_d_immse);

  -- purpose: multiplexor for destination register
  -- type   : combinational
  -- inputs : id_c_muxwbr, id_a_rt, id_a_rd
  -- outputs: id_c_wbr
  ID_MUXWBR: id_c_wbr <= "111"   when id_c_muxwbr=mw_7 else
                         id_a_rt when id_c_muxwbr=mw_rt else
                         id_a_rd; -- when id_c_muxwbr=mw_rd else
                         --"000";

  -- passes internal write-back register signal to outputs
  ID_PASSWBR: ID_CO_WBR <= id_c_wbr; 

  -- purpose: Calculate control signals for the rest of the pipeline
  -- type   : combinational
  -- inputs : id_d_op, id_d_func
  -- outputs: ID_C_EXmuxA, ID_C_EXmuxB, ID_CO_DM, ID_CO_WB, id_c_muxwbr
  ID_CONTROL: process (id_d_op, id_d_func)
  begin  -- process ID_CONTROL
    case id_d_op is
      when op_Rformat =>
        id_c_muxwbr <= mw_rd;
        if id_d_func=funct_jr or id_d_func=funct_umul then
          id_c_wb <= '0';
        else
          id_c_wb <= '1';
        end if;
        ID_CO_AC <= id_d_func;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "00";
        ID_CO_DM <= '0';

      when op_addi =>
        id_c_wb <= '1';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_rt;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "01";
        ID_CO_DM <= '0';

      when op_j =>
        id_c_wb <= '0';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_0;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "00";
        ID_CO_DM <= '0';

      when op_jal =>
        id_c_wb <= '1';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_7;
        ID_CO_EX_muxA <= '1';
        ID_CO_EX_muxB <= "11";
        ID_CO_DM <= '0';

      when op_beq | op_bne =>
        id_c_wb <= '0';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_0;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "00";
        ID_CO_DM <= '0';

      when op_lw =>
        id_c_wb <= '1';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_rt;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "01";
        ID_CO_DM <= '1';

      when op_sw =>
        id_c_wb <= '0';
        ID_CO_AC <= funct_add;
        id_c_muxwbr <= mw_rt;
        ID_CO_EX_muxA <= '0';
        ID_CO_EX_muxB <= "01";
        ID_CO_DM <= '1';

      when others =>
        report "EINVOP - invalid/unstable value for instruction" severity error;
    end case;
  end process ID_CONTROL;
  
  -- Detects a write-back to register zero and disables that write-back
  ID_DETECTWBRZERO: ID_CO_WB <= '0' when id_c_wbr="000" else id_c_wb;

end architecture behaviour;
