-------------------------------------------------------------------------------
-- InstFetch.vhd
-- -------------
-- InstFetch interfaces with the IM. The entity has its own register for PC
-- and must thus be clocked same as rest.
-- InstFetch assumes data on im_d to be valid once the CPU is unstalled.
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973431 Rune M. Barnkob
--             :  c973816 Emil Heinze
--
-- Last Update :  2001/01/15
--             :  2001/01/20
--             :  2001/01/25
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity InstFetch is port(
  IF_CO_STALL:    out bit_t;        -- stall-signal to rest of CPU (stalls entire CPU)
  IF_CI_CLK:      in  bit_t;        -- clock signal for IF stage
  IF_CI_RESET:    in  bit_t;        -- resets PC-register
  IF_CI_ENABLE:   in  bit_t;        -- enables internal register (used when stalling)
  -- forwarded inputs
  ID_DI_IF_newPC: in  halfword_t;   -- if a new PC must be forced in from ID
  ID_CI_IF_muxPC: in  bit_t;        -- then set this bit.
  -- output to pipeline register
  IF_DO_I:        out halfword_t;   -- the instruction read from IM
  IF_DO_PC:       out halfword_t;   -- PC+1
  -- I-memory interface
  im_addr:        out halfword_t;   -- Address bus for instruction.
  im_d:           in  halfword_t;   -- The instruction bus. one-way from v.2
  im_req:         out bit_t;        -- Request signal for instruction.
  im_rw:          out bit_t;        -- Read/Write signal for instruction.
  im_hold:        in  bit_t         -- Hold signal for instruction.
  );
end entity InstFetch;

architecture behaviour of InstFetch is
  signal if_d_curpc: halfword_t;
  signal if_d_nextpc : halfword_t;
  signal if_d_tmppc:  halfword_t;
begin  -- behaviour
  -- purpose: Handles the PC register. If reset then go to Z on bus and reset reg. 
  -- type   : sequential
  -- inputs : IF_CI_CLK, IF_CI_RESET
  -- outputs: if_d_curpc
  IF_PC: process (IF_CI_CLK, IF_CI_RESET)
  begin  -- process
    if IF_CI_RESET/='1' then
      if_d_curpc <= halfword_zero;
    elsif (IF_CI_CLK'event and IF_CI_CLK='1') then
      if IF_CI_ENABLE='1' then
        if_d_curpc <= if_d_nextpc;
      end if;
    end if;
  end process IF_PC;

  -- purpose: the combinational circuit preceeding the PC register (a mux).
  -- type   : combinational
  -- inputs : ID_DI_IF_newPC, ID_D_IF_muxPC, if_d_tmppc
  -- outputs: if_d_nextpc
  IF_muxPC: if_d_nextpc <= ID_DI_IF_newPC when ID_CI_IF_muxPC='1' else if_d_tmppc;

  -- purpose: +1 adder for PC
  -- type   : combinational
  -- inputs : if_d_curpc
  -- outputs: if_d_tmppc
  IF_ADDPC: if_d_tmppc <= unsigned(if_d_curpc) + 1;
  
  -- purpose: pass adder result to outputs
  -- type   : combinational
  -- inputs : if_d_tmppc
  -- outputs: IF_DO_PC
  IF_PASSPC: IF_DO_PC <= if_d_tmppc;

  -- purpose: interfaces to memory and stalls CPU if nescessary
  -- type   : combinational
  -- inputs : im_hold, if_d_curpc, im_d
  -- outputs: IF_CO_STALL
  IF_ADDR: im_addr <= if_d_curpc;
  IF_RNOTW: im_rw <= '1';
  IF_REQUEST: im_req <= '0' when IF_CI_RESET/='1' else '1';
  IF_STALLER: IF_CO_STALL <= im_hold;
  IF_DATAREAD: IF_DO_I <= im_d;
end architecture behaviour;

