-------------------------------------------------------------------------------
-- MemDualPort.vhd
-- --------------
-- This is a memory entity featuring a dual-port read/write interface for use
-- with testbenches interfacing directly to the CPU core.
-- 
-- 
--
-- Authors : ?
--
-- Version : ? - Initial version

-------------------------------------------------------------------------------
LIBRARY IEEE; USE IEEE.std_logic_1164.ALL; 
USE IEEE.std_logic_arith.ALL; USE STD.TEXTIO.ALL; 
USE WORK.types.ALL;   



ENTITY memdualport IS   
  GENERIC (filename : IN STRING  := "";  -- name of memory load file
           n        : IN INTEGER := 0);  -- number of hold cycles(wait states)    
  PORT    (Addr : IN    halfword_t := halfword_z;
	   DataWriteBus : IN halfword_t;
	   DataReadBus  : OUT halfword_t;
           RW   : IN    bit_t := 'Z';
           Req  : IN    bit_t := 'Z';
           Hold : OUT   bit_t := '0';
           Clk  : IN    bit_t);
END memdualport;   

ARCHITECTURE behaviour OF memdualport IS
type memtype is array(0 to 2**16-1) of halfword_t;
BEGIN    

  RAM : PROCESS     
    VARIABLE memory  : MemType;  -- memory array  
    VARIABLE index   : NATURAL;  -- index to memory array     
    VARIABLE Int_Hold : bit_t := '0'; 
    VARIABLE HoldCount : INTEGER := 0;
   
    -- This procedure loads memory from a file.
    PROCEDURE LoadMemoryFromFile(VARIABLE memory   : OUT MemType;    
                                 CONSTANT filename : IN  STRING) IS     
                                  FILE     memfile : TEXT IS filename;    
                                  VARIABLE memline : LINE; 
                                  VARIABLE tmp     : BIT_VECTOR(15 downto 0);  
    BEGIN
      FOR i IN memory'RANGE LOOP      
        EXIT WHEN ENDFILE(memfile);        
        READLINE(memfile, memline);       
        READ(memline, tmp);     
        memory(i) := TO_STDLOGICVECTOR(tmp);   
      END LOOP;
    END LoadMemoryFromFile;  

  BEGIN 
    IF (filename /= "") THEN      
      LoadMemoryFromFile(memory, filename);   
    END IF;
    LOOP  -- forever for memory access   
   
      -- Read address on negative clockedge.
      -- If address is undefined (tristated or just X's) then address is set to 
      -- a pseudo-random value. This also happens if the address is changed during
      -- a memory transaction.  
      IF Clk'EVENT AND Clk = '0' AND Int_Hold = '0' AND NOT Is_X(Addr) THEN
        index := CONV_INTEGER(UNSIGNED(Addr));   
      ELSIF (Addr'EVENT AND Int_Hold = '1') OR IS_X(Addr) THEN 
        index := (index * 27677) MOD 65535; 
      END IF;
      
      -- On positive clock-edge we write data to the memory. 
      IF Clk'EVENT AND Clk = '1' THEN
        -- Make sure the memory isn't driving the bus.
        DataReadBus <= halfword_z AFTER Tpd;   
        IF (Req = '1') AND (RW = '0') AND (Int_Hold = '0') THEN  
          --memory(index) := Data;
		memory(index) := DataWriteBus;  
        END IF; 
        
      ELSIF Clk'EVENT AND Clk = '0' THEN
        -- This generates the hold signal. 
        IF (Req = '1') AND (Int_Hold = '0') AND (n /= 0) THEN  
          Int_Hold :='1';
          HoldCount := n;
          Hold <= Int_Hold AFTER Tpd;
        END IF;  
        -- This upholds the hold signal for n clockperiods.
        IF HoldCount > 0 THEN 
          HoldCount := HoldCount - 1;
        ELSE 
          Int_Hold := '0'; 
          Hold <= Int_Hold AFTER Tpd;
        END IF;
        -- For read transactions.
        IF (Req = '1') AND (RW = '1') AND (Int_Hold = '0') THEN       
          DataReadBus <= memory(index) AFTER Tpd;     
        ELSE         
          DataReadBus <= halfword_z AFTER Tpd;   
        END IF;     
      END IF;
      WAIT ON Addr,Req, RW, Clk; 
    END LOOP; 
  END PROCESS RAM;
END behaviour; 
