-------------------------------------------------------------------------------
-- cpu.vhd
-- -------
-- The structure and interface required for the CPU for the 5-stage pipeline 
-- version of the MiniMIPS CPU.
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973431 Rune M. Barnkob
--             :  c973618 Emil Heinze
--
-- History     :  2001/01/18 - made dual-port memory interface without tri-st.
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity cpu is port (
    clk :    in  bit_t;            -- The clock.
    reset:   in  bit_t;            -- The reset signal.
    im_addr: out halfword_t;       -- Address bus for instruction.
    im_din:  in  halfword_t;       -- The instruction bus.
    im_dout: out halfword_t;       -- (dummy)
    im_req:  out bit_t;            -- Request signal for instruction.
    im_rw:   out bit_t;            -- Read/Write signal for instruction.
    im_hold: in  bit_t;            -- Hold signal for instruction.
    dm_addr: out halfword_t;       -- Address bus for data.
    dm_din:  in  halfword_t;       -- The data in bus
    dm_dout: out halfword_t;       -- Data bus for output
    dm_req:  out bit_t;            -- Request signal for data.
    dm_rw:   out bit_t;            -- Read/Write signal for data.
    dm_hold: in  bit_t;            -- Hold signal for data.
    mu_hold: out bit_t             -- Multiplier needs to hold CPU waiting for data
  );
end entity cpu;

--------------------------------------------------------------------------------
-- The structural architecture of the CPU2-design. Direct Entity Instantiation 
-- is used (avaiable only in VHDL-93, should not be used when trying to 
-- implement...)
--------------------------------------------------------------------------------
architecture structure of cpu is
  component InstFetch is port(
    IF_CO_STALL:    out bit_t;        -- stall-signal to rest of CPU (stalls entire CPU)
    IF_CI_CLK:      in  bit_t;        -- clock signal for IF stage
    IF_CI_RESET:    in  bit_t;        -- resets PC-register
    IF_CI_ENABLE:   in  bit_t;        -- enables internal register (used when stalling)
    -- forwarded inputs
    ID_DI_IF_newPC: in  halfword_t;   -- if a new PC must be forced in from ID
    ID_CI_IF_muxPC: in  bit_t;        -- then set this bit.
    -- output to pipeline register
    IF_DO_I:        out halfword_t;   -- the instruction read from IM
    IF_DO_PC:       out halfword_t;   -- PC+1
    -- I-memory interface
    im_addr:        out halfword_t;   -- Address bus for instruction.
    im_d:           in  halfword_t;   -- The instruction bus. one-way from v.2
    im_req:         out bit_t;        -- Request signal for instruction.
    im_rw:          out bit_t;        -- Read/Write signal for instruction.
    im_hold:        in  bit_t         -- Hold signal for instruction.
    );
  end component InstFetch;
  
  component InstDecode is port(
    ID_CO_STALL:    out bit_t;      -- this stage needs to stall CPU (never)
    ID_CI_CLK:      in  bit_t;      -- clock for registers in RegFile
    ID_CI_RESET:    in  bit_t;      -- resets RegFile
    ID_CI_ENABLE:   in  bit_t;      -- enables register file clock
    -- input from pipeline register
    ID_DI_PC:       in  halfword_t; -- PC+1 stored
    ID_DI_I:        in  halfword_t; -- Instruction read by IF
    -- forwarding from WriteBack stage
    WB_CI_ID_WB:    in  bit_t;
    WB_CI_ID_WBR:   in  std_logic_vector(2 downto 0);
    WB_DI_ID_RES:   in  halfword_t;
    -- forwarding from DataMem stage
    DM_CI_ID_WB:    in  bit_t;
    DM_CI_ID_WBR:   in  std_logic_vector(2 downto 0);
    DM_DI_ID_RES:   in  halfword_t;
    -- forwarding from Execute stage
    EX_CI_ID_WB:    in  bit_t;
    EX_CI_ID_WBR:   in  std_logic_vector(2 downto 0);
    EX_DI_ID_RES:   in  halfword_t;
    -- new PC's if branches are performed
    ID_DO_IF_newPC: out halfword_t;
    ID_CO_IF_muxPC: out bit_t;
    -- outputs to pipeline register
    ID_DO_PC:       out halfword_t;  -- PC+1 (needed for JAL
    ID_DO_RS:       out halfword_t;  -- contents of register RS (after forwarding)
    ID_DO_RT:       out halfword_t;  -- contents of register RT (after forwarding)
    ID_DO_IMM:      out std_logic_vector(6 downto 0);  -- immediate read from I[6:0]
    ID_CO_WB:       out bit_t;       -- this instruction will write-back to a register
    ID_CO_WBR:      out std_logic_vector(2 downto 0);  -- the register it will write to
    ID_CO_DM:       out bit_t;       -- this instruction accesses data memory
    ID_CO_AC:       out std_logic_vector(3 downto 0);  -- control for the ALU
    ID_CO_EX_muxA:  out bit_t;       -- control for muxA in EX
    ID_CO_EX_muxB:  out std_logic_vector(1 downto 0)  -- control for muxB in EX
    );
  end component InstDecode;
  
  component Execute is port(
    EX_CO_STALL: out bit_t;
    EX_CI_CLK:   in  bit_t;
    EX_CI_RESET: in  bit_t;
    EX_CI_DM:    in  bit_t;
    EX_CI_AC:    in  std_logic_vector(3 downto 0);  -- alu control
    EX_DI_RS:    in  halfword_t;    -- first operand for the ALU
    EX_DI_RT:    in  halfword_t;    -- second operand for the ALU
    EX_DI_PC:    in  halfword_t;    -- PC+1 for this instruction
    EX_DI_IMM:   in  std_logic_vector(6 downto 0); -- immediate from I[6:0]
    EX_CI_WB:    in  bit_t;         -- this instruction does write-back to a register
    EX_CI_WBR:   in  std_logic_vector(2 downto 0);  -- write-back register
    EX_CI_muxA:  in  bit_t;         -- (pc) or (rs) on muxA
    EX_CI_muxB:  in  std_logic_vector(1 downto 0);  -- (imm_se), (rt) or (1) on muxB
    EX_DO_RES:   out halfword_t;    -- result of ALU operation
    EX_CO_DM:    out bit_t;         -- straight-through version of DM
    EX_CO_WB:    out bit_t;         -- --"-- WB
    EX_CO_WBR:   out std_logic_vector(2 downto 0);  -- --"-- WBR
    EX_DO_RT:    out halfword_t     -- --"-- RT
    );
  end component Execute;
  
  component DataMem is port(
    DM_CO_STALL: out bit_t;             -- this stage needs to stall entire CPU
    DM_CI_DM:    in  bit_t;             -- this instruction accesses DM
    DM_CI_WB:    in  bit_t;             -- write-back in register (not write-back in DM)
    DM_CI_WBR:   in  std_logic_vector(2 downto 0);  -- write-back register number
    DM_DI_RES:   in  halfword_t;        -- address of memory cell
    DM_DI_RT:    in  halfword_t;        -- contents of RT for this instruction (if write to DM)
    DM_CO_WB:    out bit_t;             -- write-back to register
    DM_CO_WBR:   out std_logic_vector(2 downto 0);  -- write-back register number
    DM_DO_RES:   out halfword_t;        -- either a by-pass of DM_DI_RES (if DM is cleared) or
                                        -- whatever was read from DM
    dm_hold:     in  bit_t;             -- wait for memory
    dm_din:      in  halfword_t;        -- data read from memory
    dm_dout:     out halfword_t;        -- data written to memory
    dm_req:      out bit_t;             -- req. signal for this stage
    dm_rw:       out bit_t;             -- read (1) or write (0)
    dm_addr:     out halfword_t         -- address to access
    );
  end component DataMem;
  
  component WriteBack is port(
      WB_CI_WB:  in bit_t;
      WB_CI_WBR: in std_logic_vector(2 downto 0);
      WB_DI_RES: in halfword_t;
      WB_CO_WB:  out bit_t;
      WB_CO_WBR: out std_logic_vector(2 downto 0);
      WB_DO_RES: out halfword_t
    );
  end component WriteBack;

  subtype wbr_t is std_logic_vector(2 downto 0);
  subtype imm_t is std_logic_vector(6 downto 0);
-- Signal declarations:
  signal stall:         bit_t;
  signal if_en, ifid_en, idex_en, exdm_en, dmwb_en:   bit_t; -- enable for pipeline registers
  signal IF_CO_STALL:   bit_t; 
  signal IF_DO_I:       halfword_t; --> ID_DI_I
  signal IF_DO_PC:      halfword_t; --> ID_DI_PC
  signal ID_D_IF_newPC: halfword_t; -- connection between ID_DI_IF_newPC and ID_DO_ID_pc
  signal ID_C_IF_muxPC: bit_t;      
  signal ID_CO_STALL:   bit_t;
  signal ID_DI_PC:      halfword_t;
  signal ID_DI_I:       halfword_t;
  signal ID_DO_PC:      halfword_t; --> EX_DI_PC
  signal ID_DO_RS:      halfword_t; --> EX_DI_RS
  signal ID_DO_RT:      halfword_t; --> EX_DI_RT
  signal ID_DO_IMM:     imm_t;  --> EX_DI_IMM
  signal ID_CO_WB:      bit_t;  --> EX_CI_WB
  signal ID_CO_WBR:     wbr_t;  --> EX_CI_WBR
  signal ID_CO_DM:      bit_t;  --> EX_CI_DM
  signal ID_CO_AC:      std_logic_vector(3 downto 0);  --> EX_CI_AC
  signal ID_CO_EX_muxA: bit_t;  --> EX_CI_muxA
  signal ID_CO_EX_muxB: std_logic_vector(1 downto 0);  --> EX_CI_muxB
  signal EX_CO_STALL:   bit_t;
  signal EX_CI_DM:      bit_t;
  signal EX_CI_AC:      std_logic_vector(3 downto 0);  
  signal EX_DI_RS:      halfword_t;  
  signal EX_DI_RT:      halfword_t;  
  signal EX_DI_PC:      halfword_t;  
  signal EX_DI_IMM:     imm_t;
  signal EX_CI_WB:      bit_t; 
  signal EX_CI_WBR:     wbr_t;
  signal EX_CI_muxA:    bit_t;  
  signal EX_CI_muxB:    std_logic_vector(1 downto 0);  
  signal EX_DO_RES:     halfword_t;  --> DM_DI_RES
  signal EX_CO_DM:      bit_t; --> DM_CI_DM, EX_C_ID_DM
  signal EX_CO_WB:      bit_t; --> DM_CI_WB, EX_CI_ID_WBR
  signal EX_CO_WBR:     wbr_t;  --> DM_CI_WBR, EX_DI_ID_RES
  signal EX_DO_RT:      halfword_t; --> DM_DI_RT
  signal DM_CO_STALL:   bit_t;
  signal DM_CI_DM:      bit_t; 
  signal DM_CI_WB:      bit_t; 
  signal DM_CI_WBR:     wbr_t;
  signal DM_DI_RES:     halfword_t;  
  signal DM_DI_RT:      halfword_t;
  signal DM_CO_WB:      bit_t; --> WB_CI_WB, DM_CI_ID_WB
  signal DM_CO_WBR:     wbr_t; --> WB_CI_WBR, DM_CI_ID_WBR
  signal DM_DO_RES:     halfword_t;  --> WB_DO_RES, DM_DI_ID_RES
  signal WB_CI_WB:      bit_t;
  signal WB_CI_WBR:     wbr_t;
  signal WB_DI_RES:     halfword_t;
  signal WB_CO_WB:      bit_t; --> WB_CI_ID_WB
  signal WB_CO_WBR:     wbr_t; --> WB_CI_ID_WBR
  signal WB_DO_RES:     halfword_t; --> WB_DI_ID_RES
  
begin  -- structure
  im_dout <= halfword_zero;
  
  -- pipeline stage IF
  stageIF: InstFetch port map (
    IF_CO_STALL => IF_CO_STALL,
    IF_CI_CLK => clk,
    IF_CI_RESET => reset,
    IF_CI_ENABLE => if_en,
    IF_DO_I => IF_DO_I,
    IF_DO_PC => IF_DO_PC,
    ID_DI_IF_newPC => ID_D_IF_newPC,
    ID_CI_IF_muxPC => ID_C_IF_muxPC,
    im_addr => im_addr,
    im_d => im_din,
    im_req => im_req,
    im_rw => im_rw,
    im_hold => im_hold
  );

  -- pipeline registers IFID
  IFID: process (clk, reset)
  begin
    if reset/='1' then
      ID_DI_I <= halfword_zero;
      ID_DI_PC <= halfword_zero;
    elsif clk'event and clk='1' then
      if ifid_en='1' then
        ID_DI_I <= IF_DO_I;
        ID_DI_PC <= IF_DO_PC;
      end if;
    end if;
  end process IFID;

  -- pipeline stage ID
  stageID: InstDecode port map (
    ID_CO_STALL => ID_CO_STALL,
    ID_CI_CLK => clk,
    ID_CI_RESET => reset,
    ID_CI_ENABLE => ifid_en,
    ID_DI_PC => ID_DI_PC,
    ID_DI_I => ID_DI_I,
    -- forwarding from write-back stage
    WB_CI_ID_WB => WB_CO_WB,
    WB_CI_ID_WBR => WB_CO_WBR,
    WB_DI_ID_RES => WB_DO_RES,
    -- forwarding from DataMem stage
    DM_CI_ID_WB => DM_CO_WB,
    DM_CI_ID_WBR => DM_CO_WBR,
    DM_DI_ID_RES => DM_DO_RES,
    -- forwarding from Execute stage
    EX_CI_ID_WB => EX_CO_WB,
    EX_CI_ID_WBR => EX_CO_WBR,
    EX_DI_ID_RES => EX_DO_RES,
    -- new PC's if branches are performed
    ID_DO_IF_newPC => ID_D_IF_newPC,
    ID_CO_IF_muxPC => ID_C_IF_muxPC,
    -- outputs to pipeline register
    ID_DO_PC => ID_DO_PC,
    ID_DO_RS => ID_DO_RS,
    ID_DO_RT => ID_DO_RT,
    ID_DO_IMM => ID_DO_IMM,
    ID_CO_WB => ID_CO_WB,
    ID_CO_WBR => ID_CO_WBR,
    ID_CO_DM => ID_CO_DM,
    ID_CO_AC => ID_CO_AC,
    ID_CO_EX_muxA => ID_CO_EX_muxA,
    ID_CO_EX_muxB => ID_CO_EX_muxB
  );

  -- pipeline register IDEX
  IDEX: process (clk, reset)
  begin
    if reset/='1' then
      EX_DI_PC <= halfword_zero;
      EX_DI_RS <= halfword_zero;
      EX_DI_RT <= halfword_zero;
      EX_DI_IMM <= "0000000";
      EX_CI_DM <= '0';
      EX_CI_AC <= "0000";
      EX_CI_WB <= '0';
      EX_CI_WBR <= "000";
      EX_CI_muxA <= '0';
      EX_CI_muxB <= "00";
    elsif clk'event and clk='1' then
      if ifid_en='1' then
        EX_DI_PC <= ID_DO_PC;
        EX_DI_RS <= ID_DO_RS;
        EX_DI_RT <= ID_DO_RT;
        EX_DI_IMM <= ID_DO_IMM;
        EX_CI_DM <= ID_CO_DM;
        EX_CI_AC <= ID_CO_AC;
        EX_CI_WB <= ID_CO_WB;
        EX_CI_WBR <= ID_CO_WBR;
        EX_CI_muxA <= ID_CO_EX_muxA;
        EX_CI_muxB <= ID_CO_EX_muxB;
      end if;
    end if;
  end process IDEX;

  -- pipeline stage EX
  stageEX: Execute port map (
    EX_CO_STALL => EX_CO_STALL,
    EX_CI_CLK => clk,
    EX_CI_RESET => reset,
    EX_CI_DM => EX_CI_DM,
    EX_CI_AC => EX_CI_AC,
    EX_DI_RS => EX_DI_RS,
    EX_DI_RT => EX_DI_RT,
    EX_DI_PC => EX_DI_PC,
    EX_DI_IMM => EX_DI_IMM,
    EX_CI_WB => EX_CI_WB,
    EX_CI_WBR => EX_CI_WBR,
    EX_CI_muxA => EX_CI_muxA,
    EX_CI_muxB => EX_CI_muxB,
    EX_DO_RES => EX_DO_RES,
    EX_CO_DM => EX_CO_DM,
    EX_CO_WB => EX_CO_WB,
    EX_CO_WBR => EX_CO_WBR,
    EX_DO_RT => EX_DO_RT
  );  

  -- pipeline register EXDM
  EXDM: process (clk, reset)
  begin
    if reset/='1' then
      DM_CI_DM <= '0';
      DM_CI_WB <= '0';
      DM_CI_WBR <= "000";
      DM_DI_RES <= halfword_zero;
      DM_DI_RT <= halfword_zero;
    elsif clk'event and clk='1' then
      if ifid_en='1' then
        DM_CI_DM <= EX_CO_DM;
        DM_CI_WB <= EX_CO_WB;
        DM_CI_WBR <= EX_CO_WBR;
        DM_DI_RES <= EX_DO_RES;
        DM_DI_RT <= EX_DO_RT;
      end if;
    end if;
  end process EXDM;

  -- pipeline stage DM
  stageDM: DataMem port map (
    DM_CO_STALL => DM_CO_STALL,
    DM_CI_DM => DM_CI_DM,
    DM_CI_WB => DM_CI_WB,
    DM_CI_WBR => DM_CI_WBR,
    DM_DI_RES => DM_DI_RES,
    DM_DI_RT => DM_DI_RT,
    DM_CO_WB => DM_CO_WB,
    DM_CO_WBR => DM_CO_WBR,
    DM_DO_RES => DM_DO_RES,
    dm_hold => dm_hold,
    dm_din => dm_din,
    dm_dout => dm_dout,
    dm_req => dm_req,
    dm_rw => dm_rw,
    dm_addr => dm_addr
  );
               
  -- pipeline register DMWB
  DMWB: process (clk, reset)
  begin
    if reset/='1' then
      WB_CI_WB <= '0';
      WB_CI_WBR <= "000";
      WB_DI_RES <= halfword_zero;
    elsif clk'event and clk='1' then
      if ifid_en='1' then
        WB_CI_WB <= DM_CO_WB;
        WB_CI_WBR <= DM_CO_WBR;
        WB_DI_RES <= DM_DO_RES;
      end if;
    end if;
  end process DMWB;
  
  -- pipeline stage WB
  stageWB: WriteBack port map (
    WB_CI_WB => WB_CI_WB,
    WB_CI_WBR => WB_CI_WBR,
    WB_DI_RES => WB_DI_RES,
    WB_CO_WB => WB_CO_WB,
    WB_CO_WBR => WB_CO_WBR,
    WB_DO_RES => WB_DO_RES
  );
  
  -- purpose: watches stall-signals from each stage and (if nescessary - later vers)
  --        : stalls the entire CPU if one of the signals are asserted.
  --        : NOTE: reset stalls CPU.
  -- type   : combinatorial
  -- input  : IF_CO_STALL, ID_CO_STALL, EX_CO_STALL, DM_CO_STALL, reset
  -- output : stall 
  STALL_ARBITRATOR: stall <= '1' when reset/='1' else
                             IF_CO_STALL or DM_CO_STALL; -- only IF or DM supports stalls!
                             --IF_CO_STALL or ID_CO_STALL or EX_CO_STALL or DM_CO_STALL;
  MU_HOLD <= EX_CO_STALL;

  -- concurrent signal assignments to all enable signals.
  if_en <= not stall;
  ifid_en <= not stall;
  idex_en <= not stall;
  exdm_en <= not stall;
  dmwb_en <= not stall;
  
end architecture structure;

------------------------<< EOF >>----------------------------------------------
