-------------------------------------------------------------------------------
-- Mult.vhd
-- --------
-- Simple shift-and-add multiplier
--
-- Author      :  c973811 Mikael Lund Jepsen
--             :  c973816 Emil Heinze
--             :  c973431 Rune M. Barnkob
--
-- Last Update :  2001/01/29
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity Mult is port (
  -- global
  clk: in bit_t;
  reset: in bit_t;
  -- input: when start='1', the multiplication starts
  A: in halfword_t;
  B: in halfword_t;
  Start: in bit_t;
  -- output: when Ready='1' result is ready
  Hi: out halfword_t;
  Lo: out halfword_t;
  Ready: out bit_t
  );
end entity Mult;

architecture behaviour of Mult is
  signal sum: std_logic_vector(16 downto 0);
  signal loreg: halfword_t;
  signal hireg: halfword_t;
  signal multor: halfword_t;
  signal count: std_logic_vector(4 downto 0);
  signal i_ready: bit_t;
  signal andmultor: halfword_t;
begin
  -- purpose: ready signal
  -- type   : sequential
  -- input  : clk, reset, start
  -- output : ready
  READY_REG: process(clk, reset)
  begin
    if reset/='1' then
      count <= "10000";
    elsif clk'event and clk='1' then
      if start='1' then
        count <= "00000";
      elsif count/="10000" then
        count <= unsigned(count) + 1;
      end if;
    end if;
  end process READY_REG;
  READY_SIG: i_ready <= '1' when count="10000" else '0';
  OUT_READY: Ready <= i_ready;
  
  -- purpose: The shift-register, that shifts the operand to the right
  --          and enters the sum as hireg
  -- type   : sequential
  -- input  : clk, reset
  -- output : loreg, hireg
  PROD_REG: process(clk, reset)
  begin
    if reset/='1' then
      loreg <= halfword_zero;
      hireg <= halfword_zero;
    elsif clk'event and clk='1' then
      if Start='1' then
        loreg <= B;
        hireg <= halfword_zero;
      elsif i_ready/='1' then
        loreg(14 downto 0) <= loreg(15 downto 1);
        loreg(15) <= sum(0);
        hireg <= sum(16 downto 1);
      end if;
    end if;
  end process PROD_REG;
  
  -- purpose: Register for the multiplier
  -- type   : sequential
  -- input  : clk, reset
  -- output : mult
  MULT_REG: process(clk, reset)
  begin
    if reset/='1' then
      multor <= halfword_zero;
    elsif clk'event and clk='1' then
      if start='1' then
        multor <= A;
      end if;
    end if;
  end process MULT_REG;
  
  AND_GATES: andmultor<= multor and loreg(0)&loreg(0)&loreg(0)&loreg(0)&
                                    loreg(0)&loreg(0)&loreg(0)&loreg(0)&
                                    loreg(0)&loreg(0)&loreg(0)&loreg(0)&
                                    loreg(0)&loreg(0)&loreg(0)&loreg(0);
  
  ADDER: sum <= unsigned('0'&andmultor) + unsigned('0'&hireg);
  
  OUT_HI: Hi <= hireg;
  
  OUT_LO: Lo <= loreg;

end;
