-------------------------------------------------------------------------------
-- im_cache.vhd
-- ---------
-- This is the instruction memory cache interface for the MiniMIPS cpu.
--
-- Author : Emil Heinze (c973816)
--        : Mikael Lund Jepsen (c973811)
--        : Rune M. Barnkob (c973431)
--
-- Version: 2001/01/26 - initial version
--
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use WORK.types.all;

entity im_cache is port(
  reset:    in  bit_t;
  clk:      in  bit_t;
  -- Interface towards CPU core
  imcpu_req:   in  bit_t; -- Request signal from CPU
  imcpu_hold:  out bit_t; -- Hold signal to CPU while accessing memory
  imcpu_rw:    in  bit_t; -- rw (expected constant =1)
  imcpu_addr:  in  std_logic_vector(15 downto 0); -- Address from CPU
  imcpu_dout:  out halfword_t; -- The data read either from cache or memory
  imcpu_din:   in  halfword_t; -- Dummy port

  -- The interface against the bus interface
  imbus_req:   out bit_t; -- request to send to bus interface
  imbus_hold:  in  bit_t; -- hold from bus interface
  imbus_rw:    out bit_t; -- rw signal to bus interface
  imbus_addr:  out halfword_t; -- address to bus interface
  imbus_dout:  out halfword_t; -- data send to businterface (dummy)
  imbus_din:   in  halfword_t  -- data received from memory
);
end entity im_cache;

architecture behaviour of im_cache is
  -- Each cache entry's contents is defined in the following record
  type cache_entry is record
    notvalid: bit_t;                         -- not valid bit (0 if entry is valid)
    tag:      std_logic_vector(15 downto 3); -- tag (upper address bits from addr)
    data:     halfword_t;                    -- The data itself
  end record;
  -- Defines the entire cache as a single type
  type cache_t is array(0 to 7) of cache_entry; 
  
  signal caddr: std_logic_vector(2 downto 0); -- index of entry to read
  signal cache: cache_t; -- the cache itself
  signal new_ent: cache_entry; -- the new entry composed of data from memory and address
  signal index_ent: cache_entry; -- the entry indexed by the current CPU core address
  signal miss: bit_t; -- signals that a miss in the cache occured
  signal c_enable: bit_t;

begin
  -- purpose: emulates the cache memory
  -- type   : sequential
  -- input  : clk, reset
  -- output : cache()
  THE_CACHE: process(clk, reset)
  begin
    if reset /= '1' then
      for lp in cache'range loop
        cache(lp).notvalid <= '1';
        cache(lp).tag <= (others=>'0');
        cache(lp).data <= halfword_zero;
      end loop;
    elsif clk'event and clk='1' then
      if c_enable='1' then
        cache(CONV_INTEGER(unsigned(caddr))) <= new_ent;
      end if;
    end if;
  end process;
  
  -- extract the index from the CPU core address
  caddr <= imcpu_addr(2 downto 0);
  -- index the cache with the address extracted
  index_ent <= cache(CONV_INTEGER(unsigned(caddr)));
  -- detect a miss in the cache
  miss <= '1' when (index_ent.tag/=imcpu_addr(15 downto 3)) or index_ent.notvalid='1' else '0';
  -- the mux selecting either data read from memory or cache data
  imcpu_dout <= imbus_din when miss='1' else index_ent.data;
  -- prepare an updated entry
  new_ent.notvalid <= '0';
  new_ent.tag <= imcpu_addr(15 downto 3);
  new_ent.data <= imbus_din;
  -- 
  c_enable <= (miss and (not imbus_hold));
  --
  imbus_req <= miss and imcpu_req;
  -- push address to bus interface
  imbus_addr <= imcpu_addr;
  -- pass output from cpu core to bus
  imbus_dout <= imcpu_din;
  -- 
  imbus_rw <= imcpu_rw;
  --
  imcpu_hold <= imbus_hold;

end architecture behaviour;
