-- -----------------------------------------------------------------------------
--
--  Title      :  Mini-MIPS memory entity.
--             :
--  Developers :  Niels Haandbk -- c958307@student.dtu.dk
--             :
--  Purpose    :  This design contains a small memory for the Mini-MIPS CPU.     
--             :  The memory loads data from a file specified as a generic.     
--             :
--  Notes      :  This architecture uses signal attributes.                     
--             :
--  Revision   :  1.0    12-9-99     Initial version
--             :                                                   
--              
-- -----------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- This is the memory model that should be used in task two. Write a small test 
-- bench for the memory and examine the behaviour if needed.
--------------------------------------------------------------------------------

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.std_logic_arith.ALL;
USE STD.textio.ALL;
USE WORK.types.ALL;

ENTITY mem2 IS   
    GENERIC (filename:    STRING  := "";      -- Name of memory load file
             n:           NATURAL := 0;       -- Number of hold cycles(wait states)
             Tclock:      TIME := 50 ns);     -- Length of clock-period.
    PORT (clk:        IN  std_logic;              -- The clock.
          addr:       IN  std_logic_vector(16 DOWNTO 1);         -- The address bus.
          data:       INOUT std_logic_vector(16 DOWNTO 1);       -- The data bus.
          rw:         IN  std_logic;              -- Read/Write signal.
          req:        IN  std_logic;              -- Request memory operation signal.
          hold:       OUT std_logic);             -- Hold signal.
END mem2;   

--------------------------------------------------------------------------------
-- This architecture for the memory is sensitive to the negative clock edge.
-- If a memory operation is requested then 'holdcnt' is set to the number of
-- hold-cycles. 'holdcnt' is decremented on every negative clock edge when it is
-- greater than zero. When 'holdcnt' is equal to or less than one then the 
-- operation is done and data is either written to or read from the data bus. 
-- Notice how signal attributes are used to ensure that all signals have been 
-- held stable for the appropriate amount of time.
--------------------------------------------------------------------------------

ARCHITECTURE behaviour OF mem2 IS
    PROCEDURE LoadMemoryFromFile(VARIABLE memory: OUT MemType;
                                 CONSTANT filename: IN STRING) IS
        FILE memfile:     TEXT OPEN READ_MODE IS filename;
        VARIABLE memline: LINE;
        VARIABLE tmp:     BIT_VECTOR(15 downto 0);    
    BEGIN
        FOR i IN memory'RANGE LOOP
            EXIT WHEN ENDFILE(memfile);
            READLINE(memfile, memline);
            READ(memline, tmp);
            memory(i) := TO_STDLOGICVECTOR(tmp);
        END LOOP;
    END LoadMemoryFromFile;

    
    SIGNAL holdcnt:   NATURAL;
BEGIN    
    --hold <= '1' WHEN holdcnt > 0 ELSE '0';
    
    PROCESS
        VARIABLE memory:    MemType;  -- memory array  
    BEGIN
        ------------------------------------------------------------------------
        -- Reset all signals to acceptable values:
        ------------------------------------------------------------------------
        data <= (OTHERS => 'Z');
        holdcnt <= 0;
        
        IF (filename /= "") THEN      
            LoadMemoryFromFile(memory, filename);   
        END IF;
    
        ------------------------------------------------------------------------
        -- Endless loop for memory access.
        ------------------------------------------------------------------------
        hold <= '0';
        LOOP
            data <= (OTHERS => 'Z');
            
            IF clk'EVENT AND clk = '0' THEN
                IF holdcnt = 0 AND req = '1' THEN
                    hold <= '1';
                    holdcnt <= n;
                END IF;
                
                IF holdcnt > 0 THEN
                    holdcnt <= holdcnt - 1;
                END IF;
                
                IF holdcnt <= 1 THEN
                    IF req = '1' AND req'STABLE(n * Tclock) AND 
                       rw'STABLE(n * Tclock) AND addr'STABLE(n * Tclock) THEN
                        IF rw = '1' AND NOT Is_X(addr) THEN
                            hold <= '0';
                            data <= memory(CONV_INTEGER(UNSIGNED(addr)));
                        ELSIF rw = '1' THEN
                            hold <= '0';
                            data <= (OTHERS => 'X');
                        ELSIF rw = '0' AND NOT Is_X(addr) THEN                                                        
                            IF data'STABLE(n * Tclock) THEN
                                hold <= '0';
                                memory(CONV_INTEGER(UNSIGNED(addr))) := data;
                            ELSE
                                hold <= '0';
                                memory(CONV_INTEGER(UNSIGNED(addr))) := (OTHERS => 'X');
                            END IF;
                        ELSIF rw = '0' THEN
                            hold <= '0';
                            memory(0) := (OTHERS => 'X');
                        END IF;
                    END IF;
                END IF;
            END IF;

            WAIT ON clk; 
        END LOOP; 
    END PROCESS;
    
END behaviour;
