-- -----------------------------------------------------------------------------
--
--  Title:		 Testbench for task one of the Mini-MIPS design project.
--
--  Developers: Niels Haandbk -- c958307@student.dtu.dk
--					 Rune M. Barnkob -- c973431@student.dtu.dk
--					 Mikael Lund Jepsen -- c973811@student.dtu.dk
--	
--  Purpose:	This design contains an entity for the CPU that must be build  
--					in task one of the Mini-MIPS design project. It contains an	 
--					architecture skeleton for the entity as well.				
--
--  Notes:		The architecture contains two memory banks (in the form of two
--					arrays) for the simplest possible memory interface.
--
--  Revision:	1.0	25-08-98	Initial version
--					1.1	27-08-99	Added more comments and a nice header.
--										All types were changed to use the new MIPS 
--										types from the file types.vhd.
--					2.0	08-11-00	Main process + additional processes for 
--										data-conversion added.
--			  
-- -----------------------------------------------------------------------------

-------------------------------------------------------------------------------- 
-- The entity for task one. It has two input signals (clk and reset) and is 
-- therefore rather useless. 
-------------------------------------------------------------------------------- 

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.std_logic_arith.ALL;
USE STD.TEXTIO.ALL;
USE WORK.types.ALL;

ENTITY cpu1 IS  
	  GENERIC (programname: STRING := "");	-- name of memory load file.
	  PORT (clk:		IN  bit_t;				-- The clock.
			reset:	  IN  bit_t);				-- The reset signal.
			
END cpu1;

-------------------------------------------------------------------------------- 
-- The specification of the Mini-MIPS processor. You should modify this so that
-- it executes the Mini-MIPS instruction set. Each instruction must execute in
-- one clock-cycle. 
-------------------------------------------------------------------------------- 

ARCHITECTURE specification OF cpu1 IS 

-- Signals used to output internal data during simulation. Intended for debugging only!
SIGNAL view_reg: regType;		-- view of register content.
SIGNAL view_pc: natural;		-- view of program counter
SIGNAL view_inst: halfword_t;	-- view of current instruction

---------------------------------------------------------
-- The following processes handles dataconversion between
-- std_logic-types and (un)signed numbers.
-- Original source of algoritm as stated below.
---------------------------------------------------------

-- Converts a std_logic_vector of any size to a natural (unsigned) number.
-- Copied and modified from "Designers Guide To VHDL", pg. 256
function stdl_to_natural( stdv: in std_logic_vector) return natural is
	variable res: natural := 0;
begin
	for index in stdv'range loop
		res := res * 2;
		case stdv(index) is
			when '0' => res := res;
			when '1' => res := res + 1;
			when 'L' => report "stdl_to_natural: weak low in convert" severity note;
			when 'H' => res := res + 1; report "stdl_to_natural: weak high in convert" severity note;
			when others =>
				report "stdl_to_natural: Undefined in convert!" severity note;
		end case;
	end loop;
	return res;
end function stdl_to_natural;

-- Converts a std_logic_vector of any size to a signed number.
-- Copied and modified from "Designers Guide To VHDL", pg. 256
function stdl_to_signed( stdv: in std_logic_vector) return integer is
	variable res: integer := 0;
begin
	for index in stdv'length-2 downto 0 loop
		res := res * 2;
		case stdv(index) is
			when '0' => res := res;
			when '1' => res := res + 1;
			when 'L' => report "stdl_to_signed: weak low in convert" severity note;
			when 'H' => res := res + 1; report "stdl_to_signed: weak high in convert" severity note;
			when others =>
				report "stdl_to_signed: Undefined in convert!" severity note;
		end case;
	end loop;
	case stdv(stdv'length-1) is
		when '0' => res := res;
		when '1' => res := res - (2**(stdv'length-1));
		when 'L' => report "stdl_to_signed: weak low in convert" severity note;
		when 'H' => res := res - (2**(stdv'length-1)); report "stdl_to_signed: weak high in convert" severity note;
		when others =>
			report "stdl_to_signed: Undefined in convert!" severity note;
	end case;
	return res;
end function stdl_to_signed;

-- Converts a natural (unsigned) number to a std_logic_vector of size "len".
-- Arguments passed: the natural number, the width of the vector to be returned.
-- Copied and (almost not) modified from "Designers Guide To VHDL", pg. 257
function natural_to_stdl( v: in natural; len: in natural ) return std_logic_vector is
	variable temp: natural := v;
	variable res: std_logic_vector(len-1 downto 0);
begin
	if (temp < 0) then 
		temp := 2**len - temp;
	end if;
	for index in res'reverse_range loop
		if (temp rem 2)=1 then res(index) := '1';
		else res(index) := '0';
		end if;
		temp := temp / 2;
		--exit when temp = 0;
	end loop;
	return res;
end function natural_to_stdl;
			
			
-- A small function for handling integer ranges. 
-- Pass the value as first argument and the lower limit as second 
-- the upper limit (plus 1) is passed as third (eg. 65536 if limit is 65535).
function wraparound( v: in integer; botlimit: in integer; toplimit: integer) return integer is
variable res: integer;
begin
	res := ((v-botlimit) mod (toplimit - botlimit)) + botlimit;
	return res;
end;


BEGIN
	  
	----------------------------------------------------------------------------
	-- The main process. You should insert your code here. Make sure that it is	
	-- as short as possible.													  
	----------------------------------------------------------------------------
	PROCESS 
		VARIABLE IMemory:   MemType;	-- Instruction-cache.
		VARIABLE DMemory:   MemType;	-- Data-cache.
		
		VARIABLE	oc: OpCode;
		VARIABLE	fc: FuncCode;
		VARIABLE	rs: RegAddr;
		VARIABLE	rt: RegAddr;
		VARIABLE	rd: RegAddr;
		VARIABLE	imm: std_logic_vector(6 downto 0);
		VARIABLE	target: std_logic_vector(12 downto 0);
		VARIABLE	inst: halfword_t;
		VARIABLE	PC: MemAddr;		-- Program Counter (PC)
		VARIABLE	regs: RegType;			-- array with all registers

	BEGIN 
		-- This bit loads a program or reports an error if programname = "".
		IF (programname /= "") THEN
			LoadMemoryFromFile(IMemory, programname);
			LoadMemoryFromFile(DMemory, programname);
		ELSE
			REPORT "No program specified!";
		END IF;
		
		-- reset CPU
		PC := 0;
		regs(0) := halfword_zero;
		regs(1) := halfword_zero;
		regs(2) := halfword_zero;
		regs(3) := halfword_zero;
		regs(4) := halfword_zero;
		regs(5) := halfword_zero;
		regs(6) := halfword_zero;
		regs(7) := halfword_zero;

		-- Endless loop.
		LOOP 
			IF clk'EVENT AND clk = '1' THEN
				-- read instruction and move program counter
				inst := IMemory(PC);
				view_pc <= PC;
				view_inst <= inst;
				PC := PC + 1;
				-- split instruction into components
				oc := inst(15 downto 13);
				rs := RegAddr( stdl_to_natural(inst(12 downto 10)) );
				rt := RegAddr( stdl_to_natural(inst(9 downto 7)) );
				rd := RegAddr( stdl_to_natural(inst(6 downto 4)) );
				fc := inst(3 downto 0);
				imm := inst(6 downto 0);
				target := inst(12 downto 0);
		
				-- branch on opcode
				case oc is
					when op_Rformat =>
						-- sub-split on function during ALU-type instructions
						case fc is
							when funct_sll =>
								regs(rd) := regs(rt)(14 downto 0) & '0';
								
							when funct_srl =>
								regs(rd) := '0' & regs(rt)(15 downto 1);
								
							when funct_sra =>
								regs(rd) := regs(rt)(15) & regs(rt)(15 downto 1);
								
							when funct_jr =>
								PC := stdl_to_natural(regs(rs));
								
							when funct_add =>
								regs(rd) := natural_to_stdl( (stdl_to_natural(regs(rs)) + stdl_to_natural(regs(rt))) mod 2**16, 16 );
								
							when funct_sub =>
								regs(rd) := natural_to_stdl( (stdl_to_natural(regs(rs)) - stdl_to_natural(regs(rt))) mod 2**16, 16 );

							when funct_and =>
								for index in regs(rd)'range loop
									regs(rd)(index) := regs(rs)(index) and regs(rt)(index);
								end loop;
							
							when funct_or =>
								for index in regs(rd)'range loop
									regs(rd)(index) := regs(rs)(index) or regs(rt)(index);
								end loop;
							
							when funct_xor =>
								for index in regs(rd)'range loop
									regs(rd)(index) := regs(rs)(index) xor regs(rt)(index);
								end loop;
							
							when funct_slt =>
								if (stdl_to_signed(regs(rs))<stdl_to_signed(regs(rt))) then
									regs(rd) := "0000000000000001";
								else
									regs(rd) := "0000000000000000";
								end if;
								
							when funct_sltu =>
								if (stdl_to_natural(regs(rs))<stdl_to_natural(regs(rt))) then
									regs(rd) := "0000000000000001";
								else
									regs(rd) := "0000000000000000";
								end if;
							
							when others =>
								report "EINVFUNC - invalid function received";
								null;
						end case;
					
					when op_addi =>
						regs(rt) := signed(regs(rs)) + signed(imm);
						
					when op_j =>
						-- PC && 0xE000 + target
						PC := MemAddr(stdl_to_natural(natural_to_stdl(natural(PC),16)(15 downto 13) & target));
						
					when op_jal =>
						regs(7) := natural_to_stdl(PC,16);
						PC := MemAddr(stdl_to_natural(natural_to_stdl(natural(PC),16)(15 downto 13) & target));
						
					when op_beq =>
						if (regs(rs)=regs(rt)) then 
							PC := wraparound(PC + stdl_to_signed(imm),0,2**16);
						end if;
					
					when op_bne =>
						if (regs(rs)/=regs(rt)) then 
							PC := wraparound(PC + stdl_to_signed(imm),0,2**16);
						end if;
					
					when op_lw =>
						regs(rt) := DMemory( wraparound(stdl_to_natural(regs(rs)) + stdl_to_signed(imm), 0,2**16) );
					
					when op_sw =>
						DMemory( wraparound(stdl_to_natural(regs(rs))+ stdl_to_signed(imm), 0,2**16) ) := regs(rt);
						
					-- If opcode hasn't matched any other case, it is invalid.
					when others =>
						report "EINVOPCODE - invalid opcode received";
						null;
						
				end case;
			END IF;
			-- Model a hardwired $0.
			regs(0) := "0000000000000000";
			
			-- output registercontent on signal view_reg.
			view_reg <= regs;
			
			WAIT ON clk;
		END LOOP;
	END PROCESS;
END specification;


